/*
 * Copyright (c) 2006 by Kolja Waschk (ixo.de)
 *
 * Derived from sms.c, 
 *   Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

/* WARNING: needs more testing, and more information from Siemens. */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <gigaset/am.h>
#include <gigaset/block.h>
#include <gigaset/gigdata.h>
#include "interf-private.h"


static ssize_t gigaset_amcommand(struct gstatus * gs, const char *cmd, const char * rsp, size_t rsplen, char ** buf)
{
	ssize_t r;

	r = gigaset_sendcommand(gs, cmd);
	if (r >= 0) r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r >= 0 && rsp) r = gigaset_expectresponse(gs, rsp, rsplen, buf, 30);
	//if (r >= 0 && !buf) r = 0; //FIXME?
	return r;
}

static ssize_t gigaset_amnumcommand(struct gstatus * gs, const char *fmt, unsigned num, const char * rsp, size_t rsplen, char ** buf)
{
	int size;
	char cmd[80];

	size = snprintf(cmd, sizeof cmd, fmt, num);
	if (size < 0 || (unsigned)size >= sizeof cmd)
		return -ERR_MEM;
	return gigaset_amcommand(gs, cmd, rsp, rsplen, buf);
}

static int gigaset_amclaim(struct gstatus *gs, unsigned amindex)
{
	ssize_t r;

	r = amindex == 0 ?
	    gigaset_amcommand(gs, "AT^SABSTART", "ZABSTART", 8+1, NULL) :
	    gigaset_amnumcommand(gs, "AB^SABSTART=%u", amindex, "ZABSTART", 8+1, NULL);
	return r >= 0 ? 0 : r;
}

static int gigaset_amrelease(struct gstatus *gs)
{
	ssize_t r;

	r = gigaset_amcommand(gs, "AT^SABEND", NULL, 0, NULL);
	if (r < 0)
		return r;

	if (gs->temp_filter[FLT_ZABEND]->status) {
		gigaset_usleep_unint(800000L);
		return 0;
	}

	r = gigaset_expectresponse(gs, NULL, 0, NULL, 30); /* wait for ZABEND */
	if (r == -FILTER_ZABEND) {
		gigaset_usleep_unint(800000L);
		return 0;
	}

	return r;
}

int gigaset_amstart(struct gstatus *gs, unsigned amindex)
{
	int r = -ERR_MEM;

	gs->temp_filter[FLT_ZABINFO] = gigaset_addfilter(gs, -1,             7+1, "ZABINFO", 0, NULL); // FIXME: ignore this globally
	gs->temp_filter[FLT_ZABCAU]  = gigaset_addfilter(gs, -1,             7,   "ZABCAU=", 0, NULL);
	gs->temp_filter[FLT_ZABEND]  = gigaset_addfilter(gs, FILTER_ZABEND,  6+1, "ZABEND",  0, NULL);

	if (gs->temp_filter[FLT_ZABINFO] &&
	    gs->temp_filter[FLT_ZABCAU] && gs->temp_filter[FLT_ZABEND]) {
		r = gigaset_amclaim(gs, amindex);
		if (r >= 0)
			return 0;
	}

	gigaset_delfilter(gs, gs->temp_filter[FLT_ZABINFO]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZABCAU ]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZABEND ]);

	return r;
}

int gigaset_amcontinue(struct gstatus *gs)
{
	//FIXME
	if (gs->temp_filter[FLT_ZABEND]->status) return -ERR_ABORT;
	return 0;
}

int gigaset_amend(struct gstatus * gs)
{
	int r;

	r = gigaset_amrelease(gs);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZABCAU]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZABINFO]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZABEND]);

	return r;
}

static ssize_t gigaset_amblockrec(struct gstatus * gs, const char *cmd, uint8_t blocktype, char ** buffer)
{
	ssize_t r;

	r = gigaset_sendcommand(gs, cmd);
	if (r >= 0) r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r >= 0) r = gigaset_blockreceive(gs, blocktype, buffer);
	return r;
}

static ssize_t gigaset_amnumblockrec(struct gstatus * gs, const char *fmt, unsigned num, uint8_t blocktype, char ** buffer)
{
	char cmd[80];
	int size;

	size = snprintf(cmd, sizeof cmd, fmt, num);
	if (size < 0 || (unsigned)size >= sizeof cmd)
		return -ERR_MEM;

	return gigaset_amblockrec(gs, cmd, blocktype, buffer);
}

ssize_t gigaset_amgetlist(struct gstatus * gs, char ** buffer)
{
	return gigaset_amblockrec(gs, "AT^SABINFO?", 0x02, buffer);
}

ssize_t gigaset_aminfo(struct gstatus * gs, unsigned hdl, char **buffer)
{
	return gigaset_amnumblockrec(gs, "AT^SABINFO=%u", hdl, 0x02, buffer);
}

ssize_t gigaset_amgetmsg(struct gstatus * gs, unsigned hdl, char **buffer)
{

	return gigaset_amnumblockrec(gs, "AT^SABGET=%u", hdl, 0x02, buffer);
}

int gigaset_amputmsg(struct gstatus * gs, const char * buffer, size_t len)
{
	ssize_t r;

	r = gigaset_sendcommand(gs, "AT^SABPUT");
	if (r >= 0) r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r >= 0) r = gigaset_blocksend(gs, len, 0x08, buffer);

	if (r < 0)
		return r;

	/* Here, the phone sends ZABEND first and we respond(!) with AT^SABEND */
	r = gigaset_expectresponse(gs, NULL, 0, NULL, 30); /* wait for ZABEND */
	if (r != -FILTER_ZABEND)
		return r;

	r = gigaset_amend(gs); /* AT^SABEND + clear filters */

	return r < 0 ? r : 0;
}

int gigaset_amdelmsg(struct gstatus * gs, unsigned hdl)
{
	ssize_t r;

	r = gigaset_amnumcommand(gs, "AT^SABDEL=%u", hdl, "ZABDEL", 6+1, NULL);

	return r < 0 ? r : 0;
}

int gigaset_amdelall(struct gstatus * gs)
{
	ssize_t r;

	r = gigaset_amcommand(gs, "AT^SABDELLST", "ZABDEL", 6+1, NULL);

	return r < 0 ? r : 0;
}

int gigaset_amdeltyped(struct gstatus * gs, unsigned msgtype)
{
	ssize_t r;

	r = gigaset_amnumcommand(gs, "AT^SABDELLST=%u", msgtype, "ZABDEL", 6+1, NULL);

	return r < 0 ? r : 0;
}
