/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#define GIG_WAIT

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <gigaset/interf.h>
#include "interf-private.h"
#include <gigaset/basemisc.h>

int gigaset_msnget (struct gstatus * gs, int id, char ** buf)
{
	ssize_t r;
	int l;
	char cmd[7+2+1+1];
	char rsp[4+1+2+1+1];

	if (id<1||id>10) return -ERR_PARAMINVALID;
	r = snprintf(cmd,sizeof(cmd),"AT+SMSN%d?",id);
	if (r<0||(unsigned)r>=sizeof(cmd)) return -ERR_MEM; /* are there still libraries that return negative values? */
	l = snprintf(rsp,sizeof(rsp),"ZMSN=%d,",id);
	if (l<0||(unsigned)l>=sizeof(rsp)) return -ERR_MEM; /* are there still libraries that return negative values? */
	r=gigaset_sendcommand (gs, cmd);
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r<0) return r;
	r=gigaset_expectresponse(gs, rsp, l, buf, 30);
	if (r<0) {
		free(*buf);
		return r;
	}
	memmove(*buf,*buf+l,strlen(*buf)-l+1);
	return 0;
}

int gigaset_msnset (struct gstatus * gs, int id, const char * buf)
{
	ssize_t r;
	int l;
	size_t s;
	char * cmd;

	if (id<1||id>10) return -ERR_PARAMINVALID;
	//FIXME also check string
	s=strlen(buf)+ 7+2+1+1;
	cmd=malloc(s);
	if (!cmd) return -ERR_MEM;
	l = snprintf(cmd,s,"AT+SMSN%d=%s",id,buf);
	if (l<0||(unsigned)l>=s) /* are there still libraries that return negative values? */
	{
		free (cmd);
		return -ERR_MEM;
	}
	r=gigaset_sendcommand (gs, cmd);
	free (cmd);
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r<0) return r;
#ifdef GIG_WAIT
	gigaset_usleep_unint(1000000);
	r=gigaset_clear (gs);
	if (r<0) return r;
#endif
	return 0;
}

int gigaset_cfxset (struct gstatus * gs, const char * msn, int mode, int proc, int service, char * to_msn)
{
	ssize_t r;
	int l;
	size_t s;
	char * cmd;

	s=255;
	cmd=malloc(s);
	if (!cmd) return -ERR_MEM;
	l = snprintf(cmd,s,"AT^SCFX=%s,%d,%d,%d,%s", msn, mode, proc, service, to_msn);
	if (l<0||(unsigned)l>=s)
	{
		free (cmd);
		return -ERR_MEM;
	}
	r=gigaset_sendcommand (gs, cmd);
	free (cmd);
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r<0) return r;
#ifdef GIG_WAIT
	gigaset_usleep_unint(1000000);
	r=gigaset_clear (gs);
	if (r<0) return r;
#endif
	return 0;
}

static int vendorstring (struct gstatus * gs, char ** buf, const char * cmd)
{
	ssize_t r;
	r=gigaset_sendcommand(gs,cmd);
	if (r<0) return r;
	r=gigaset_getresponse(gs,buf,30);
	if (r<0) return r;
	r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 0);
	if (r == -ERR_TIMEOUT) {
		if (!strcmp(*buf, "OK")) {
			**buf = 0;
			r = 0;
		} else if (!strcmp(*buf, "ERROR")) {
			r = -FILTER_ERROR;
		}
	}
	if (r>=0) return 0;
	free(*buf);
	return r;
}

int gigaset_vendor (struct gstatus * gs, char ** buf)
{
	return vendorstring (gs,buf,"AT+GMI");
}

int gigaset_product (struct gstatus * gs, char ** buf)
{
	return vendorstring (gs,buf,"AT+GMM");
}

int gigaset_version (struct gstatus * gs, char ** buf)
{
	return vendorstring (gs,buf,"AT+GMR");
}

int gigaset_dsversion (struct gstatus * gs, char ** buf)
{
	return vendorstring (gs,buf,"AT^SDGMR");
}

int gigaset_device (struct gstatus * gs, char ** buf)
{
	return vendorstring (gs,buf,"AT^SGMM");
}

int gigaset_dial(struct gstatus *gs, unsigned type, const char *number, const char *id)
{
	int r, r2;
	char *apdcmd, *dialcmd;
	size_t len, size;
	char *buf;
	char *end;
	unsigned long cid;
	char cmdbuf[2+5+ 7 +1];

	if (type > 4)
		return -ERR_PARAMINVALID;

	len = strlen(number);
	size = len + 2 + 5 + 1;
	if (size < len)
		return -ERR_MEM;
	dialcmd = malloc(size);
	if (!dialcmd)
		return -ERR_MEM;

	len = strlen(id);
	size = len + 8 + 1;
	r = -ERR_MEM;
	if (size >= len) {
		apdcmd = malloc(size);
		if (apdcmd) {
			memcpy(apdcmd, "AT^SAPD=", 8);
			memcpy(apdcmd+8, id, len);
			apdcmd[size-1] = 0;
			r = gigaset_sendcommand(gs, apdcmd);
			free (apdcmd);
		}
	}
	if (r < 0) {
		free(dialcmd);
		return r;
	}

	r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r < 0) {
		free(dialcmd);
		return r;
	}
	r = gigaset_expectresponse(gs, "ZAPD=", 5, NULL, 30);
	if (r < 0)
		goto APD0;
	/* FIXME: ZAPD=0 => error */

	/* get cid */

	r = gigaset_sendcommand(gs, "AT^SGACI?");
	if (r < 0)
		goto APD0;
	r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r < 0)
		goto APD0;
	r = gigaset_expectresponse(gs, "ZGACI=", 6, &buf, 30);
	if (r < 0) {
		free(buf);
		goto APD0;
	}

	errno = 0;
	cid = strtoul(buf+6, &end, 10);
	if (errno || end == buf+6 || *end || cid < 1 || cid > 65535) {
		r = -ERR_RESPINVALID;
		free(buf);
		goto APD0;
	}
	free(buf);

	sprintf(cmdbuf, "AT%u^SCTP=%u", (unsigned)cid, type);
	r = gigaset_sendcommand(gs, cmdbuf);
	if (r < 0)
		goto rmcid;
	r = gigaset_expectcidresponse(gs, "OK", 2+1, cid, NULL, 30);
	if (r < 0)
		goto rmcid;
	//r = gigaset_expectcidresponse(gs, "ZCTP=", 5, cid, NULL, 30);
	//if (r < 0)
	//	goto APD0;

	sprintf(cmdbuf, "AT%u+VLS=7", (unsigned)cid);
	r = gigaset_sendcommand(gs, cmdbuf);
	if (r < 0)
		goto rmcid;
	r = gigaset_expectcidresponse(gs, "OK", 2+1, cid, NULL, 30);
	if (r < 0)
		goto rmcid;
	r = gigaset_expectcidresponse(gs, "ZVLS=7", 6+1, cid, NULL, 30);
	if (r < 0)
		goto rmcid;
	r = gigaset_expectcidresponse(gs, "ZSAU=OUTGOING_CALL_PROCEEDING", 29+1, cid, NULL, 30);
	if (r < 0)
		goto rmcid;

	sprintf(dialcmd, "AT%uD%s", (unsigned)cid, number);
	r = gigaset_sendcommand(gs, dialcmd);
	if (r < 0)
		goto rmcid;
	r = gigaset_expectcidresponse(gs, "OK", 2+1, cid, NULL, 30);
	if (r < 0)
		goto rmcid;
	r = gigaset_expectcidresponse(gs, "ZCPN=", 5, cid, NULL, 30);
	if (r < 0)
		goto rmcid;

	gigaset_log(gs->gl, GLOG_STATUS, "dialing %s\n", number);

	//if (gs->sigint)
	//	*gs->sigint = 0; //FIXME
	gs->may_interrupt = 1;
	r = gigaset_expectcidresponse(gs, "ZSAU=", 5, cid, &buf, GIG_INFTY);
	gs->may_interrupt = 0;
	if (r < 0) {
		free(buf);
		goto rmcid;
	}
	if (!strcmp(buf, "ZSAU=CALL_DELIVERED")) {
		free(buf);
		gs->may_interrupt = 1;
		r = gigaset_expectcidresponse(gs, "ZSAU=", 5, cid, &buf, GIG_INFTY);
		gs->may_interrupt = 0;
		if (r < 0) {
			free(buf);
			goto rmcid;
		}
	}
	if (strcmp(buf, "ZSAU=ACTIVE")) {
		//FIXME return codes for local hup, remote hup, success, error?
		// local hup:
		//    ZSAU=DISCONNECT_REQ
		//    ZCAU=xx,yy
		//    ZSAU=NULL
		// remote hup, busy, ...
		//    ZSAU=DISCONNECT_IND
		//    ZCAU=xx,yy
		if (strcmp(buf, "ZSAU=DISCONNECT_REQ")) {
			free(buf);
			goto rmcid;
		}
		r = gigaset_expectcidresponse(gs, "ZSAU=NULL", 9+1, cid, NULL, 3);
		if (r < 0)
			goto rmcid;
		goto APD0;
	}
	free(buf);
	
	/* ignore ZCON=NUMBER;CID */

	//FIXME wait for ZSAU=NULL?

	r = 0;
	goto APD0;

rmcid:
	r2 = gigaset_clear(gs);
	if (r >= 0 && r2 < 0)
		r = r2;
	sprintf(cmdbuf, "AT%u+VLS=0", (unsigned)cid);
	r2 = gigaset_sendcommand(gs, cmdbuf);
	if (r2 >= 0)
		r2 = gigaset_expectcidresponse(gs, "OK", 2+1, cid, NULL, 30);
	if (r2 >= 0)
		r2 = gigaset_expectcidresponse(gs, "ZVLS=0", 6+1, cid, NULL, 30);

	if (r >= 0 && r2 < 0)
		r = r2;

	/* wait for ZSAU/ZCAU messages */
	gigaset_usleep_unint(1000000); //FIXME

APD0:
	//FIXME ignore ZSAU/ZCAU events
	r2 = gigaset_clear(gs);
	if (r >= 0 && r2 < 0)
		r = r2;
	r2 = gigaset_sendcommand(gs, "AT^SAPD=0");
	if (r2 >= 0)
		r2 = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r2 >= 0)
		r2 = gigaset_expectresponse(gs, "ZAPD=", 5, NULL, 30);

	if (r >= 0 && r2 < 0)
		r = r2;

	free(dialcmd);
	return r;
}
