/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <gigaset/interf.h>
#include "interf-private.h"
#include <gigaset/conf.h>
#include <gigaset/gigdata.h>

int gigaset_confwrite(struct gstatus *gs, uint16_t id, uint8_t tid, uint8_t datatype, const char *buf, uint16_t l)
{
	char *cmd;
	char *raw = NULL;
	int headlen;
	ssize_t r;
	struct resp_filter *f;
	int newstyle;

	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	newstyle = gs->ver_fwbase[0] >= 2; /* this is guessed */

	if (newstyle && l > 0xff)
		return -ERR_PARAMINVALID;

	f = gigaset_addfilter(gs, FILTER_ZCFGERR, 7+1, "ZCFGERR", 0, NULL);
	if (!f)
		return -ERR_MEM;

	headlen=newstyle?4:5;
	raw=malloc(l+headlen);
	if (!raw) {
		r = -ERR_MEM;
		goto exit;
	}
	gigaset_to_le16(raw+0,id);
	raw[2]=newstyle?tid:datatype; //FIXME range check
	if (newstyle)
		raw[3]=l;
	else
		gigaset_to_le16(raw+3,l);
	memcpy (raw+headlen, buf, l);

	r=gigaset_bytestuff(&cmd,raw,headlen+l,0x7d,0x7e/*??*/, 9, 4+1);
	if (r < 0)
		goto exit;

	memcpy (cmd+0,"AT^SCFGW=",9);
	gigaset_hexchecksum (cmd+9+r,raw,headlen+l);
	free(raw);
	raw = NULL;
	cmd[9+r+4]=0;

	r=gigaset_sendcommand(gs, cmd);
	free(cmd);
	if (r<0) goto exit;
	if (r >= 0) r=gigaset_expectresponse(gs, "ZCFGW", 5+1, NULL, 30);
	if (r >= 0) r = 0;
exit:
	free(raw);
	if (r == -FILTER_ZCFGERR)
		r = -ERR_ABORT;
	gigaset_delfilter(gs, f);
	return r;
}

int_fast32_t gigaset_confread(struct gstatus *gs, uint16_t id, uint8_t tid, uint8_t datatype, char **buf)
{
	struct gconf conf;
	int r;

	conf.next = NULL;
	conf.id = id;
	conf.tid = tid; //FIXME range check
	conf.datatype = datatype; //FIXME range check
	r = gigaset_confreadarray(gs, &conf);
	if (r < 0)
		return r;
	*buf = conf.buf;
	return conf.len;
}

int gigaset_confreadarray(struct gstatus *gs, struct gconf *list)
{
	char * buf;
	char * pos;
	size_t l;
	char * s;
	ssize_t bs,r;
	int32_t fcs;
	size_t count;
	struct gconf * cur, * end;
	int headlen;
	struct resp_filter *f;
	int newstyle;
	uint8_t bval;

	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	newstyle = gs->ver_fwbase[0] >= 2; /* this is guessed */


	headlen=newstyle?4:5;

	count=0;
	for (cur=list;cur;cur=cur->next) ++count;
	if (!count) return 0;

	f = gigaset_addfilter(gs, FILTER_ZCFGERR, 7+1, "ZCFGERR", 0, NULL);
	if (!f)
		return -ERR_MEM;

	l=9+6*count+1;
	buf=malloc(l);
	if (!buf) {
		r = -ERR_MEM;
		goto exit;
	}

	memcpy (buf, "AT^SCFGR=", 9);

	pos=buf+9;
	for (cur=list;cur;cur=cur->next,pos+=6)
	{
		gigaset_tohex4 (pos, cur->id);
		gigaset_todec2 (pos+4, newstyle?cur->tid:cur->datatype);
	}
	*pos='\0';
	r=gigaset_sendcommand(gs, buf);
	free(buf);
	if (r<0) goto exit;

	//r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30, 16);
	//if (r<0) return r;

	r=gigaset_expectresponse(gs, "ZCFGR=", 6, &s, 30);
	if (r<0) {
		if (r==-FILTER_ZCFGERR) r=-ERR_ABORT;
		free(s);
		goto exit;
	}

	l=strlen(s+6)-4;
	fcs=gigaset_fromhex4(s+6+l);
	if (fcs < 0) {
		free(s);
		r = fcs;
		goto exit;
	}
	bs=gigaset_unstuff(&buf, s+6, l, 0x7d);
	free (s);
	if (bs < 0) {
		r = bs;
		goto exit;
	}
	if (!gigaset_verifychecksum(buf, bs, fcs))
	{
		free (buf);
		r = -ERR_CRC;
		goto exit;
	}

	pos=buf;
	for (cur=list;cur;cur=cur->next)
	{
		if (bs<headlen)
		{
			r=-1;
			goto err;
		}
		bs-=headlen;
		bval = newstyle ? cur->tid : cur->datatype;
		if (cur->id !=((uint8_t)pos[1]<<8|(uint8_t)pos[0]) ||bval!=(uint8_t)pos[2])
		{
			r=-1;
			goto err;
		}

		cur->len=(uint16_t)pos[3];
		if (!newstyle) cur->len|=(uint16_t)pos[4]<<8; // FIXME is pos[4] really the highbyte??
		if (bs<cur->len)
		{
			r=-1;
			goto err;
		}
		cur->buf=malloc(cur->len);
		if (!cur->buf)
		{
			r=-ERR_MEM;
			goto err;
		}
		pos+=headlen;
		memcpy (cur->buf, pos, cur->len);
		pos+=cur->len;
		bs-=cur->len;
	}
	r = 0;
	free (buf);
	goto exit;

err:	end=cur;
	for (cur=list;cur!=end;cur=cur->next) free(cur->buf);
	free (buf);
exit:
	gigaset_delfilter(gs, f);
	return r;
}

int gigaset_confinitlock (struct gstatus * gs)
{
	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	if (gs->ver_fwbase[0] < 2) /* this is guessed */
		return 0;

	gs->temp_filter[FLT_ZLOCKERR] = gigaset_addfilter(gs, FILTER_ZLOCKERR, 8+1, "ZLOCKERR", 0, NULL);
	gs->temp_filter[FLT_ZLOCKTIM] = gigaset_addfilter(gs, FILTER_ZLOCKTIM, 8+1, "ZLOCKTIM", 0, NULL);

	if (gs->temp_filter[FLT_ZLOCKERR] && gs->temp_filter[FLT_ZLOCKTIM])
		return 0;

	gigaset_delfilter(gs, gs->temp_filter[FLT_ZLOCKERR]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZLOCKTIM]);
	return -ERR_MEM;
}

int gigaset_confkeeplock(struct gstatus *gs)
{
	ssize_t r;

	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	if (gs->ver_fwbase[0] < 2) /* this is guessed */
		return 0;

	r = gigaset_sendcommand(gs, "AT^SLOCK");
	if (r >= 0) r = gigaset_expectresponse(gs, "ZLOCKOK", 7+1, NULL, 30);
	if (r >= 0) r = 0;
	return r;
}

int gigaset_confunlock (struct gstatus * gs)
{
	ssize_t r;

	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	if (gs->ver_fwbase[0] < 2) /* this is guessed */
		return 0;

	r = gigaset_sendcommand(gs, "AT^SUNLOCK");
	if (r >= 0) r = gigaset_expectresponse(gs, "ZLOCKOK", 7+1, NULL, 30);
	if (r >= 0) r = 0;

	gigaset_delfilter(gs, gs->temp_filter[FLT_ZLOCKERR]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZLOCKTIM]);
	return r;
}
