/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdlib.h>
#include <stdint.h>
#include <gigaset/gigdata.h>
#include <gigaset/interf.h>

static uint16_t gigaset_fcstab[256] =
{
	0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
	0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
	0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
	0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
	0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
	0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
	0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
	0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
	0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
	0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
	0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
	0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
	0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
	0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
	0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
	0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
	0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
	0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
	0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
	0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
	0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
	0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
	0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
	0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
	0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
	0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
	0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
	0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
	0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
	0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
	0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
	0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

uint16_t gigaset_checksum(const char * buf, size_t l)
{
	uint16_t fcs;
	size_t i;
	const char * c;

	fcs=0xffff;

	for (i=0,c=buf;i<l;++i)
		fcs = (fcs >> 8) ^ gigaset_fcstab[(fcs ^ *c++) & 0xff];
	return fcs;
}

static char const hx[16]="0123456789ABCDEF";

void gigaset_hexchecksum (char * dest, const char * buf, size_t l)
{
	uint16_t fcs;
	char * w;
	int i;
	fcs=gigaset_checksum(buf, l);
	for (w=dest+3,i=0;i<4;++i,--w,fcs>>=4) *w=hx[fcs&0xf];
}

int gigaset_verifychecksum(const char * buf, size_t l, uint16_t CRC)
{
	return gigaset_checksum(buf, l)==CRC;
}

void gigaset_tohex4 (char * dest, uint16_t val)
{
	char * w;
	int i;
	for (w=dest+3,i=0;i<4;++i,--w,val>>=4) *w=hx[val&0xf];
}

void gigaset_todec2 (char * dest, unsigned char val)
{
	val%=100;
	*dest=val/10+'0';
	*++dest=val%10+'0';
}

int8_t gigaset_fromhex1(char c)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	if (c >= 'a' && c <= 'f')
		return c - 'a' + 10; // ASCII charset required!!111
	if (c >= 'A' && c <= 'F')
		return c - 'A' + 10; // ASCII charset required!!111
	return -1;
}

int16_t gigaset_fromhex2 (const char *src)
{
	uint8_t r=0;
	int i;

	for (i=0;i<2;++i,++src)
	{
		r<<=4;
		if (*src>='0'&&*src<='9')
			r|=*src-'0';
		else if (*src>='A'&&*src<='F')
			r|=*src-'A'+10;
		else if (*src>='a'&&*src<='f')
			r|=*src-'a'+10;
		else
			return -1;
	}
	return r;
}

int32_t gigaset_fromhex4 (const char * src)
{
	uint16_t r=0;
	int i;

	for (i=0;i<4;++i,++src)
	{
		r<<=4;
		if (*src>='0'&&*src<='9')
			r|=*src-'0';
		else if (*src>='A'&&*src<='F')
			r|=*src-'A'+10;
		else if (*src>='a'&&*src<='f')
			r|=*src-'a'+10;
		else
			return -1;
	}
	return r;
}

int gigaset_32fromdec (uint32_t * dest, const char * src)
{
	uint32_t r=0;
	char c;
	int count=0;
	for (;;)
	{
		c=*src++;
		if (!c)
		{
			if (!count) return -1;
			*dest=r;
			return 0;
		}
		if (count==10) return -1/*-11*/;
		if (c<'0'||c>'9') return -1/*-count-1*/;
		if (count==9&&(r>429496729||(r==429496729&&c>'5')))
			return -1/*-10*/;
		r=r*10+(c-'0');
		++count;
	}
}

void gigaset_32todec (char*dest, uint32_t val)
{
	char buf[11];
	char *pos=buf+10;

	*pos=0;
	do
		*--pos=(val%10)+'0';
	while (val/=10);
	while ((*dest++=*pos++)) ;
}

size_t gigaset_unpackseptetbuf(char * dest, const char *src, size_t septets, size_t skipoctets)
/* should we return an errorcode if not padded with 0 bits? */
{
	unsigned shift;
	unsigned shift2;
	uint8_t r, cur;
	size_t skip;
	size_t i;

	skip=skipoctets+(skipoctets+6)/7;
	if (skip>=septets) return 0;
	septets-=skip;

	shift2=skip&7;
	shift=(8-shift2)&7;
	/* skipoctets=skip-(skip+7)/8; */
	src+=skipoctets;
	cur=*src;

	i=septets;
	while (i--) {
		if (!shift)
		{
			shift2=1;
			shift=7;
			cur=*src;
			r=cur&0x7f;
		}
		else
		{
			r=cur>>shift;
			++src;
			if (--shift)
			{
				cur=*src;
				r|=(cur<<shift2++)&0x7f;
			}
		}
		*dest++=r;
	}

	return septets;
}

size_t gigaset_packseptetbuf(char * dest, const char *src, size_t octets, size_t skipoctets)
/* should we return an errorcode if an octet is > 0x7f ? */
{
	unsigned shift=0;
	unsigned shift2;
	uint8_t r, cur;
	size_t skip;
	size_t size;

	size=skipoctets;
	if (!octets) return size;
	skip=skipoctets+(skipoctets+6)/7;
	dest+=skipoctets;
	cur=0;
	shift2=skip&7;
	shift=(8-shift2)&7;

	while (octets--)
	{
		r=*src++;
		// if (r&0x80)...
		if (!shift)
		{
			shift2=1;
			shift=7;
			cur=r&0x7f;
		}
		else
		{
			++size;
			*dest++=cur|(r<<shift--);
			cur=(r&0x7f)>>shift2++;
		}
	}
	if (shift) {
		++size;
		*dest=cur;
	}
	return size;
}

size_t gigaset_stuffedsize(const char *src, size_t l, char esc, char add)
{
	size_t i,s;

	s=l;
	for (i=0;i<l;++src,++i)
		if ((unsigned char)*src<0x20||*src==esc||*src==add)
			++s;
	return s;
}

ssize_t gigaset_unstuffedsize(const char *src, size_t l, char esc)
{
	size_t s,i;

	s=l;
	for (i=0;i<l;++i)
		if (*src++==esc)
		{
			--s;
			if (++i==l) return -ERR_DATA;
			src++;
		}

	return s;
}

ssize_t gigaset_stuffbuf(char * dest, const char *src,  size_t l, char esc, char add, size_t bufsize)
{
	size_t i,s;

	s=0;
	for (i=0;i<l;++i)
		if (++s > bufsize)
			return -1;
		else if ((unsigned char)*src<0x20||*src==esc||*src==add)
		{
			*dest++=esc;
			if (++s > bufsize) return -1;
			*dest++=*src++ ^ 0x20;
		}
		else
			*dest++=*src++;
	return s;
}

/* gigaset_unstuffbuf
 * decode bytestuffed data to fixed size buffer
 * Parameters:
 *	dest	destination buffer for unstuffed data
 *	src	bytestuffed source data (set src==dest for in-place unstuffing)
 *	l	length of source data
 *	esc	stuffing character
 *	bufsize	size of destination buffer
 * Return value:
 *	number of bytes written to destination buffer
 *	< 0: error (-1: end of dest reached, -ERR_DATA: trailing esc)
 */
ssize_t gigaset_unstuffbuf(char * dest, const char *src,  size_t l, char esc, size_t bufsize)
{
	size_t s,i;

	s=0;
	for (i=0;i<l;++i)
		if (++s > bufsize)
			return -1;
		else if (*src==esc)
		{
			++src;
			if (++i==l) return -ERR_DATA;
			*dest++=*src++ ^ 0x20;
		}
		else
			*dest++=*src++;
	return s;
}

ssize_t gigaset_bytestuff(char ** dest, const char * src, size_t l, char esc, char add, size_t headersize, size_t tail)
{
	size_t s;
	ssize_t r;

	s=gigaset_stuffedsize(src,l,esc,add);

	*dest=malloc(s+headersize+tail);
	if (!*dest) return -ERR_MEM;
	r=gigaset_stuffbuf(*dest+headersize,src,l,esc,add,s);
	if (r>=0&&(size_t)r==s) return r;

	if (r>=0) r=-1;
	free (*dest);
	return r;
}

/* gigaset_unstuff
 * decode bytestuffed data to dynamically allocated buffer
 * Parameters:
 *	dest	receives pointer to destination buffer with unstuffed data
 *	src	bytestuffed source data
 *	l	length of source data
 *	esc	stuffing character
 * Return value:
 *	number of bytes written to destination buffer
 *	< 0: error code
 */
ssize_t gigaset_unstuff(char ** dest, const char * src, size_t l, char esc)
{
	ssize_t s,r;

	s=gigaset_unstuffedsize(src,l,esc);
	if (s<0) return s;

	*dest=malloc(s);
	if (!*dest) return -ERR_MEM;
	r=gigaset_unstuffbuf(*dest,src,l,esc,s);
	if (r==s) return r;

	if (r >= -1)	/* gigaset_unstuffedsize() counted wrong */
	    r = -ERR_DATA;
	free (*dest);
	*dest = NULL;
	return r;
}

#define GIG_SMS_ESC 0x1b
typedef uint16_t gig_unicode;
typedef uint16_t septet_table[0x100];

static const septet_table table_00 = {
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0000 */
	0xffff,0xffff,0x000a,0xffff,0x010a,0x000d,0xffff,0xffff,  /* 0x0008 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0010 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0018 */
	0x0020,0x0021,0x0022,0x0023,0x0002,0x0025,0x0026,0x0027,  /* 0x0020 */
	0x0028,0x0029,0x002a,0x002b,0x002c,0x002d,0x002e,0x002f,  /* 0x0028 */
	0x0030,0x0031,0x0032,0x0033,0x0034,0x0035,0x0036,0x0037,  /* 0x0030 */
	0x0038,0x0039,0x003a,0x003b,0x003c,0x003d,0x003e,0x003f,  /* 0x0038 */
	0x0000,0x0041,0x0042,0x0043,0x0044,0x0045,0x0046,0x0047,  /* 0x0040 */
	0x0048,0x0049,0x004a,0x004b,0x004c,0x004d,0x004e,0x004f,  /* 0x0048 */
	0x0050,0x0051,0x0052,0x0053,0x0054,0x0055,0x0056,0x0057,  /* 0x0050 */
	0x0058,0x0059,0x005a,0x013c,0x012f,0x013e,0x0114,0x0011,  /* 0x0058 */
	0xffff,0x0061,0x0062,0x0063,0x0064,0x0065,0x0066,0x0067,  /* 0x0060 */
	0x0068,0x0069,0x006a,0x006b,0x006c,0x006d,0x006e,0x006f,  /* 0x0068 */
	0x0070,0x0071,0x0072,0x0073,0x0074,0x0075,0x0076,0x0077,  /* 0x0070 */
	0x0078,0x0079,0x007a,0x0128,0x0140,0x0129,0x013d,0xffff,  /* 0x0078 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0080 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0088 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0090 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0098 */
	0xffff,0x0040,0xffff,0x0001,0x0024,0x0003,0xffff,0x005f,  /* 0x00a0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x00a8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x00b0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0x0060,  /* 0x00b8 */
	0xffff,0xffff,0xffff,0xffff,0x005b,0x000e,0x001c,0xffff,  /* 0x00c0 */
	0xffff,0x001f,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x00c8 */
	0xffff,0x005d,0xffff,0xffff,0xffff,0xffff,0x005c,0xffff,  /* 0x00d0 */
	0x000b,0xffff,0xffff,0xffff,0x005e,0xffff,0xffff,0x001e,  /* 0x00d8 */
	0x007f,0xffff,0xffff,0xffff,0x007b,0x000f,0x001d,0x0009,  /* 0x00e0 */
	0x0004,0x0005,0xffff,0xffff,0x0007,0xffff,0xffff,0xffff,  /* 0x00e8 */
	0xffff,0x007d,0x0008,0xffff,0xffff,0xffff,0x007c,0xffff,  /* 0x00f0 */
	0x000c,0x0006,0xffff,0xffff,0x007e,0xffff,0xffff,0xffff,  /* 0x00f8 */
};

static const septet_table table_03 = {
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0300 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0308 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0310 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0318 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0320 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0328 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0330 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0338 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0340 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0348 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0350 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0358 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0360 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0368 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0370 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0378 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0380 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x0388 */
	0xffff,0xffff,0xffff,0x0013,0x0010,0xffff,0xffff,0xffff,  /* 0x0390 */
	0x0019,0xffff,0xffff,0x0014,0xffff,0xffff,0x001a,0xffff,  /* 0x0398 */
	0x0016,0xffff,0xffff,0x0018,0xffff,0xffff,0x0012,0xffff,  /* 0x03a0 */
	0x0017,0x0015,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03a8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03b0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03b8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03c0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03c8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03d0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03d8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03e0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03e8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03f0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x03f8 */
};

static const septet_table table_20 = {
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2000 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2008 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2010 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2018 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2020 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2028 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2030 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2038 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2040 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2048 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2050 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2058 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2060 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2068 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2070 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2078 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2080 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2088 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2090 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x2098 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20a0 */
	0xffff,0xffff,0xffff,0xffff,0x0165,0xffff,0xffff,0xffff,  /* 0x20a8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20b0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20b8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20c0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20c8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20d0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20d8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20e0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20e8 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20f0 */
	0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,0xffff,  /* 0x20f8 */
};

static const uint16_t * const unicode_to_septet[0x100]={
	&table_00[0],0,0,&table_03[0],0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	&table_20[0]};

static const gig_unicode septet_to_unicode[GIG_SMS_ESC_DEPTH][0x80] = {
	{
		0x0040,0x00a3,0x0024,0x00a5,0x00e8,0x00e9,0x00f9,0x00ec,  /* 0x0000 */
		0x00f2,0x00e7,0x000a,0x00d8,0x00f8,0x000d,0x00c5,0x00e5,  /* 0x0008 */
		0x0394,0x005f,0x03a6,0x0393,0x039b,0x03a9,0x03a0,0x03a8,  /* 0x0010 */
		0x03a3,0x0398,0x039e,0,     0x00c6,0x00e6,0x00df,0x00c9,  /* 0x0018 */
		0x0020,0x0021,0x0022,0x0023,0x00a4,0x0025,0x0026,0x0027,  /* 0x0020 */
		0x0028,0x0029,0x002a,0x002b,0x002c,0x002d,0x002e,0x002f,  /* 0x0028 */
		0x0030,0x0031,0x0032,0x0033,0x0034,0x0035,0x0036,0x0037,  /* 0x0030 */
		0x0038,0x0039,0x003a,0x003b,0x003c,0x003d,0x003e,0x003f,  /* 0x0038 */
		0x00a1,0x0041,0x0042,0x0043,0x0044,0x0045,0x0046,0x0047,  /* 0x0040 */
		0x0048,0x0049,0x004a,0x004b,0x004c,0x004d,0x004e,0x004f,  /* 0x0048 */
		0x0050,0x0051,0x0052,0x0053,0x0054,0x0055,0x0056,0x0057,  /* 0x0050 */
		0x0058,0x0059,0x005a,0x00c4,0x00d6,0x00d1,0x00dc,0x00a7,  /* 0x0058 */
		0x00bf,0x0061,0x0062,0x0063,0x0064,0x0065,0x0066,0x0067,  /* 0x0060 */
		0x0068,0x0069,0x006a,0x006b,0x006c,0x006d,0x006e,0x006f,  /* 0x0068 */
		0x0070,0x0071,0x0072,0x0073,0x0074,0x0075,0x0076,0x0077,  /* 0x0070 */
		0x0078,0x0079,0x007a,0x00e4,0x00f6,0x00f1,0x00fc,0x00e0,  /* 0x0078 */
	},{
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0100 */
		0,     0,0x000c,     0,     0,     0,     0,     0,       /* 0x0108 */
		0,     0,     0,     0,0x005e,     0,     0,     0,       /* 0x0110 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0118 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0120 */
		0x007b,0x007d,0,     0,     0,     0,     0,0x005c,       /* 0x0128 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0130 */
		0,     0,     0,     0,0x005b,0x007e,0x005d,     0,       /* 0x0138 */
		0x007c,0,     0,     0,     0,     0,     0,     0,       /* 0x0140 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0148 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0150 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0158 */
		0,     0,     0,     0,     0,0x20ac,     0,     0,       /* 0x0160 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0168 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0170 */
		0,     0,     0,     0,     0,     0,     0,     0,       /* 0x0178 */
	}
};

int gigaset_valid_sms_char(uint16_t c_ucs2)
{
	const uint16_t *table = unicode_to_septet[c_ucs2 >> 8];
	return table && table[c_ucs2 & 0xff];
}

static size_t UCS2_to_utf8(uint16_t cu, unsigned char *dest, size_t destspace)
{
	if (cu < 0x80) {
		if (!destspace)
			return 0;
		*dest++ = cu;
		return 1;
	} else if (cu < 0x800) {
		if (destspace < 2)
			return 0;
		*dest++ = 0xc0 | (cu>>6);
		cu &= 0x3f;
		*dest++ = 0x80 | cu;
		return 2;
	} else if (1) /* (cu < 0x10000) */ {
		if (destspace < 3)
			return 0;
		*dest++ = 0xe0 | (cu>>12);
		cu&=0xfff;
		*dest++ = 0x80 | (cu>>6);
		cu &= 0x3f;
		*dest++ = 0x80 | cu;
		return 3;
	} else return 0;
}

/* returns 0 on failure */
size_t gigaset_UCS2_to_utf8buf(const unsigned char *src, size_t srclen, unsigned char *dest, size_t destlen, unsigned *incomplete, unsigned char *prev)
{
	uint16_t cu;
	unsigned char high, low;
	size_t destspace=destlen;
	size_t r;

	if (!srclen)
		return 0;

	if (*incomplete)
		high = *prev;
	else {
		--srclen;
		high = *src++;
	}

	while (srclen) {
		--srclen;
		low = *src++;
		cu = (uint16_t)high << 8 | low;
		if (!cu)
			cu = 0x20; //FIXME
		r = UCS2_to_utf8(cu, dest, destspace);
		if (!r)
			return 0;
		dest += r;
		destspace -= r;
		if (!srclen) {
			*incomplete = 0;
			return destlen - destspace;
		}
		--srclen;
		high = *src++;
	}

	*incomplete = 1;
	*prev = high;
	return destlen-destspace;
}

/* returns 0 on failure */
size_t gigaset_septets_to_utf8buf (const unsigned char * src, size_t srclen, unsigned char * dest, size_t destlen, unsigned * esc)
{
	unsigned esccount;
	uint16_t cu;
	unsigned char c;
	size_t destspace=destlen;

	if (esc)
		esccount=*esc;
	else
		esccount=0;

	for (;srclen;++src,--srclen) {
		c=*src;
		if (c==GIG_SMS_ESC) {
			++esccount;
			continue;
		}
		//if (c>0x7f) return 0; FIXME
		c&=0x7f;
		if (esccount>=GIG_SMS_ESC_DEPTH
		    ||(esccount&&!septet_to_unicode[esccount][c])) {
			do {
				if (!destspace--) return 0;
				*dest++=0x20;
			} while (--esccount);
		}
		cu=septet_to_unicode[esccount][c];
		esccount=0;
		if (!cu) cu=0x20;
		if (cu<0x80) {
			if (!destspace) return 0;
			destspace-=1;
			*dest++=cu;
		} else if (cu < 0x800) {
			if (destspace<2) return 0;
			destspace-=2;
			*dest++=0xc0|(cu>>6);
			cu&=0x3f;
			*dest++=0x80|cu;
		} else if (1) /* (cu < 0x10000) */ {
			if (destspace<3) return 0;
			destspace-=3;
			*dest++=0xe0|(cu>>12);
			cu&=0xfff;
			*dest++=0x80|(cu>>6);
			cu&=0x3f;
			*dest++=0x80|cu;
		} else return 0;
	}
	//if (esccount) return 0; FIXME
	if (esc) *esc=esccount;
	return destlen-destspace;
}

/* result is terminated with 0 (ressize may be NULL) */
unsigned char * gigaset_septets_to_utf8 (const unsigned char * src, size_t srclen, size_t * ressize)
{
	size_t bufsize;
	size_t realsize;
	unsigned char * buf;
	unsigned char * newbuf;

	if (ressize) *ressize=0;
	bufsize=srclen*GIG_UTF8_MAX_BYTES+1; /* worst case: src[] contains no escape characters
	                                      * and every character needs 3 bytes (+NUL).
	                                      */
	buf=malloc(bufsize);
	if (!buf) return 0;
	realsize=gigaset_septets_to_utf8buf(src,srclen,buf,bufsize,NULL);
	if (!realsize) {
		free(buf);
		return 0;
	}
	bufsize=realsize+1;
	if (bufsize<realsize) {
		free(buf);
		return 0;
	}
	newbuf=realloc(buf,bufsize);
	if (newbuf) buf=newbuf;
	buf[realsize]=0;
	if (ressize) *ressize=realsize;
	return buf;
}

ssize_t gigaset_utf8_to_septetbuf (const unsigned char * src, size_t srclen, unsigned char * dest, size_t destlen,
                                   unsigned * esccount, char * escapedchar, size_t * septets)
{
	uint16_t cu;
	uint16_t val;
	uint8_t table;
	uint8_t index;
	uint8_t esc;
	unsigned char c;
	size_t destspace=destlen;
	size_t srcleft=srclen;

	if (esccount) {
		while (*esccount) {
			if (!destspace) {
				*septets=destlen;
				return 0;
			}
			--destspace;
			if (--*esccount)
				*dest++=GIG_SMS_ESC;
			else
				*dest++=*escapedchar;
		}
	}

	for (;srcleft;++src,--srcleft) {
		c=*src;
		if (c<0x80) {
			cu=c;
		} else if (c < 0xe0) {
			if (!srcleft||(src[1]&0xc0)!=0x80) return -ERR_DATA;
			cu=((uint16_t)c&0x1fU)<<6|(src[1]&0x3fU);
			if (cu<0x80) return -ERR_DATA;
			srcleft-=1;
			src+=1;
		} else if (c < 0xf0) {
			if (srcleft<2||(src[1]&0xc0)!=0x80||(src[2]&0xc0)!=0x80) return -ERR_DATA;
			cu=((uint16_t)c&0x0fU)<<12|((uint16_t)src[1]&0x3fU)<<6|(src[2]&0x3fU);
			if (cu<0x800) return -ERR_DATA;
			srcleft-=2;
			src+=2;
		} else return -ERR_DATA;
		table=cu>>8;
		index=cu&0xff;
		if (!unicode_to_septet[table]) return -ERR_DATA;
		val=unicode_to_septet[table][index];
		if(val==0xffff) return -ERR_DATA;
		esc=val>>8;
		for (;esc;--esc) {
			if (!destspace) break;
			--destspace;
			*dest++=GIG_SMS_ESC;
		}
		if (!destspace) {
			if (esccount) {
				*escapedchar=val&0xff;
				*esccount=esc+1;
			}
			*septets=destlen;
			return srclen-srcleft+1;
		}
		--destspace;
		*dest++=val&0xff;
	}
	if (esccount) *esccount=0;
	*septets=destlen-destspace;
	return srclen;
}

unsigned char * gigaset_utf8_to_septets (const unsigned char * src, size_t srclen, size_t * ressize)
{
	size_t bufsize;
	size_t realsize;
	unsigned char * buf;
	unsigned char * newbuf;
	ssize_t num;

	if (!ressize) return 0;
	*ressize=0;
	bufsize=srclen*GIG_SMS_ESC_DEPTH;    /* worst case: src[] contains unly characters < 0x80
	                                      * and every character needs 2 bytes.
	                                      */
	buf=malloc(bufsize);
	if (!buf) return 0;
	num=gigaset_utf8_to_septetbuf(src,srclen,buf,bufsize,NULL,NULL,&realsize);
	if (num<0) {
		free(buf);
		return 0;
	}
	newbuf=realloc(buf,realsize);
	if (newbuf) buf=newbuf;
	*ressize=realsize;
	return buf;
}


#if 0
#include <stdio.h>

static unsigned char testtext[]={'W','i','l', /*EUR*/ 0xe2, 0x82, 0xac, /*ouml*/  0xc3, 0xb6, 0};

static void test_tables(void)
{
	unsigned i,j;
	uint16_t val;
	uint8_t esc;
	uint8_t table;
	uint8_t index;
	puts("scanning unicode_to_septet");
	for (i=0;i<0x100;++i) {
		if (!unicode_to_septet[i]) continue;
		for (j=0;j<0x100;++j) {
			val=unicode_to_septet[i][j];
			if (val==0xffff) continue;
			esc=val>>8;
			index=val&0xff;
			if (esc>=GIG_SMS_ESC_DEPTH||index&0x80) {
				printf ("error: unicode_to_septet[0x%02x][0x%02x]==0x%04x\n",
				        (unsigned) i, (unsigned) j, (unsigned) val);
				continue;
			}
			val=septet_to_unicode[esc][index];
			if (val!=((uint16_t)i<<8|j)) {
				printf ("error: unicode %02x %02x => %02x %02x => unicode %04x\n",
				        (unsigned) i, (unsigned) j, (unsigned) esc, (unsigned) index, (unsigned) val);
			}
		}
	}
	puts("scanning septet_to_unicode");
	for (i=0;i<GIG_SMS_ESC_DEPTH;++i) {
		for (j=0;j<0x80;++j) {
			val=septet_to_unicode[i][j];
			if (!val) continue;
			table=val>>8;
			index=val&0xff;
			if (!unicode_to_septet[table]) {
				printf ("error: septet_to_unicode[0x%02x][0x%02x]==0x%04x\n",
				        (unsigned) i, (unsigned) j, (unsigned) val);
				continue;
			}
			val=unicode_to_septet[table][index];
			if (val!=((uint16_t)i<<8|j)) {
				printf ("error: %02x %02x => unicode %02x %02x => %04x\n",
				        (unsigned) i, (unsigned) j, (unsigned) esc, (unsigned) index, (unsigned) val);
			}
		}
	}
}

static void dump (const unsigned char * buf, size_t len)
{
	unsigned count=0;
	for(;len;--len) {
		printf (" %02x",(unsigned)*buf++);
		count=(count+1)&0xf;
		if (!count) putchar ('\n');
	}
	if (count) putchar ('\n');
}

int main (void)
{
	unsigned char * buf1=NULL, * buf2=NULL, * buf3=NULL;
	size_t l1,l2,l3;

	test_tables();

	dump(testtext,sizeof testtext);
	buf1=gigaset_utf8_to_septets (testtext,sizeof testtext-1, &l1);
	if (!buf1) {
		puts ("buf1==NULL");
		goto exit;
	}
	dump(buf1,l1);

	//l2=(l1*7+7)/8;
	l2=l1-l1/8;
	buf2=malloc(l2);
	if (!buf2) {
		puts ("buf2==NULL");
		goto exit;
	}
	gigaset_packseptetbuf(buf2, buf1, l1);
	dump(buf2,l2);

	gigaset_unpackseptetbuf(buf1, buf2, l1);
	dump(buf1,l1);

	buf3=gigaset_septets_to_utf8 (buf1,l1,&l3);
	if (!buf3) {
		puts ("buf3==NULL");
		goto exit;
	}
	dump(buf3,l3+1);

exit:	free (buf1);
	free (buf2);
	free (buf3);

	return EXIT_SUCCESS;
}
#endif
