/*
 * This part of the configuration program handles the "low level"
 * communication with the device driver.
 *
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#ifndef GIGASET_INTERF_PRIVATE_H
#define GIGASET_INTERF_PRIVATE_H

#include <gigaset/interf.h>

#define FLT_ZSMEND    0
#define FLT_ZSMCANCEL 1
#define FLT_ZSMERR    2
#define FLT_ZLOCKERR  3
#define FLT_ZLOCKTIM  4
#define FLT_ZABEND    5
#define FLT_ZABCAU    6
#define FLT_ZABINFO   7
#define FLT_NUM       8

#
/* these should not used as normal error codes */
#define FILTER_ERROR            257

/* these should only be defined in the corresponding modules */
#define FILTER_CURRENT          256
#define FILTER_ZSMEND           258
#define FILTER_ZSMCANCEL        259
#define FILTER_ZSMERR           260
#define FILTER_ZBABORT          261
#define FILTER_ZPBEND           262
#define FILTER_ZPBCAU           263
#define FILTER_ZPBEOT           264
#define FILTER_ZLOCKERR         265
#define FILTER_ZLOCKTIM         266
#define FILTER_ZCFGERR          267
#define FILTER_FAILURE          268
#define FILTER_ZABEND           269
//#define FILTER_ZABCAU           270
//#define FILTER_ZABINFO          271

#define FILTER_LIBMAX           511

struct resp_filter {
	struct resp_filter *next, *prev;
	unsigned status;
	int retval;
	size_t len;
	char str[1];
};

struct resp_buf {
	struct resp_buf *next, *prev;
	size_t len;
	char *str;
};

struct hwstate;

struct gstatus {
	struct glog *gl;
	struct hwstate *hws;
	struct resp_buf *response;
	//int connected;
	volatile sig_atomic_t *sigint;
	int may_interrupt;
	int locked;
	int config;
	int cidmode;
	int unimodem;
	int have_ver_fwbase;
	unsigned ver_fwbase[4];
	int error; /* errno */
	char *prefix;
	size_t preflen;
	struct resp_filter *filter, *lastfilter;
	struct resp_buf *respqueue, *resplast;
	int last_timeout;
	struct resp_filter *temp_filter[FLT_NUM];
};

int gigaset_read(struct gstatus *gs, int timeout);
int gigaset_write(struct gstatus *gs, const char *s);
int gigaset_config(struct gstatus *gs, int l);

struct resp_filter *gigaset_addfilter(struct gstatus *gs, int retval,
                                      size_t len, const char *str, size_t preflen, const char *prefix);
void gigaset_delfilter(struct gstatus *gs, struct resp_filter *f);

int gigaset_configcommand (struct gstatus * gs, const char * s);
int gigaset_configsequence(struct gstatus *gs, const char *cmd, int strings,
			   int testok, char **p, int timeout);

int gigaset_sendcommand(struct gstatus *gs, const char *s);
ssize_t gigaset_getresponse(struct gstatus *gs, char **p, int timeout);
ssize_t gigaset_expectcidresponse(struct gstatus *gs, const char *s, size_t l, unsigned cid, char **p, int timeout);
ssize_t gigaset_expectresponse(struct gstatus *gs, const char *s, size_t l, char **p, int timeout);
ssize_t gigaset_expectconfigresponse(struct gstatus *gs, const char *s, size_t l, char **p, int timeout);

#endif // GIGASET_INTERF_PRIVATE_H
