/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <gigaset/interf.h>
#include "interf-private.h"
#include <gigaset/m10x.h>

static int m10x_getinfo (struct gstatus * gs, int id, int adapter, char ** buf)
{
	char cmd[5]="I0,0";
//	int r;

	if (id<1||id>6||adapter<0||adapter>1) return -1;

	cmd[1]+=id;
	cmd[3]+=adapter;

	return gigaset_configsequence(gs, cmd, 1, 0, buf, 30);
}

#if 0
static int m10x_setvalue (struct gstatus * gs, int index, int max, const char * cmd)
{
	char * buf;
	size_t l;
	int r;

	if (index<0||index>max) return -1;
	if (index>9999) return -1;
	l=strlen (cmd);
	buf=malloc(l+4+1);
	if (!buf) return -ERR_MEM;
	r=snprintf (buf, l+4+1, "%s%d", cmd, index);
	if (r<0||(unsigned)r>=l+4+1)
		r=-1;
	else
		r = gigaset_configsequence(gs, buf, 0, 0, NULL, 30);
	free (buf);
	return r;
}
#endif

int m10x_vendor (struct gstatus * gs, int adapter, char ** buf)
{
	return m10x_getinfo (gs, 3, adapter, buf);
}

int m10x_getname (struct gstatus * gs, int adapter, char ** buf)
{
	return m10x_getinfo (gs, 4, adapter, buf);
}

int m10x_devid (struct gstatus * gs, int adapter, char ** buf)
{
	return m10x_getinfo (gs, 2, adapter, buf);
}

int m10x_version (struct gstatus * gs, int adapter, char ** buf)
{
	return m10x_getinfo (gs, 1, adapter, buf);
}

int m10x_versionhw (struct gstatus * gs, int adapter, char ** buf)
{
	return m10x_getinfo (gs, 6, adapter, buf);
}

int m10x_connstate(struct gstatus *gs, char **buf)
{
	return gigaset_configsequence(gs, "#C", 1, 0, buf, 30);
}

int m10x_field (struct gstatus * gs, char ** buf)
{
	return gigaset_configsequence(gs, "#I0", 1, 0, buf, 30);
}

int m10x_quality (struct gstatus * gs, char ** buf)
{
	return gigaset_configsequence(gs, "#I1", 1, 0, buf, 30);
}

int m10x_setpartner(struct gstatus *gs, unsigned partner)
{
	char cmd[20];
	int r;

	r = snprintf(cmd, sizeof cmd, "#D%u", partner);
	if (r >= sizeof cmd)
		return -ERR_MEM;

	return gigaset_configsequence(gs, cmd, 0, 0, NULL, 30);
}

static int get_list(struct gstatus *gs, char command, int adapter, void *buf, int to_int, unsigned max)
{
	char cmd[4];
	int i;
	int r, count;
	unsigned long val;
	char *endp;
	char *startp;
	char **list;
	unsigned int *intlist = NULL;

	if (adapter < 0 || adapter > 1)
		return -ERR_PARAMINVALID;

	cmd[0] = '#';
	cmd[1] = command;
	cmd[2] = adapter + '0';
	cmd[3] = 0;

	list = malloc(10 * sizeof *list);
	if (!list)
		return -ERR_MEM;

	if (to_int) {
		intlist = malloc(10 * sizeof *intlist);
		if (!list) {
			free(intlist);
			return -ERR_MEM;
		}
	}

	r = gigaset_configsequence(gs, cmd, 10, 1, list, 30);
	if (r < 0) {
		free(intlist);
		return r;
	}

	count = r;

	for (i = 0; i < count; ++i) {
		if (list[i][0] != '0' + i || list[i][1] != '=') {
			r = -ERR_DATA;
			break;
		}
		if (to_int) {
			errno = 0;
			startp = list[i] + 2;
			val = strtoul(startp, &endp, 10);
			if (errno || endp == startp || *endp != 0 || val > max) {
				r = -ERR_DATA;
				break;
			}
			intlist[i] = val;
		} else {
			startp = list[i];
			endp = startp + 2;
			while ((*startp++ = *endp++) != 0)
				;
		}
	}

	if (r < 0 || to_int) {
		for (i = 0; i < count; ++i)
			free(list[i]);
		free(list);
	}

	if (r < 0)
		free(intlist);
	else if (to_int)
		*(unsigned int **)buf = intlist;
	else
		*(char ***)buf = list;

	return r;
}

int m10x_adapterlist(struct gstatus *gs, int adapter, char ***list)
{
	return get_list(gs, 'R', adapter, list, 0, 0);
}

int m10x_listautorel(struct gstatus *gs, int adapter, unsigned **list)
{
	return get_list(gs, 'P', adapter, list, 1, UINT_MAX);
}

int m10x_listconns(struct gstatus *gs, int adapter, unsigned **list)
{
	return get_list(gs, 'W', adapter, list, 1, 1);
}

int m10x_occupyconn(struct gstatus *gs)
{
	return gigaset_configsequence(gs, "#V1", 0, 0, NULL, 30);
}

int m10x_releaseconn(struct gstatus * gs)
{
	return gigaset_configsequence(gs, "#V0", 0, 0, NULL, 30);
}

int m10x_setautorel(struct gstatus *gs, int secs)
{
	char cmd[6];
	int r;
	if (secs < 0 || secs > 655) return -1;
	r=snprintf(cmd, sizeof cmd, "#Q%d", secs);
	if (r < 0 || r >= sizeof cmd) return -1;

	return gigaset_configsequence(gs, cmd, 0, 0, NULL, 30);
}

//int m10x_status (struct gstatus * gs, char ** buf) //FIXME only for testing
//{
//	return -1;
//}

int m10x_reset (struct gstatus * gs)
{
	return gigaset_configsequence(gs, "&F", 0, 0, NULL, 30);
}

/* The commands are from ftp://ftp.linux-magazin.de/pub/devel/gigaset-0.8.tgz
 * are not working (and not mentioned in the specs). We use
 *   ^SBASCx          instead of ^SETTYPEx
 * and
 *   ^SOACxxxx,y      instead of ^REGISTERxxxx,y (which y too choose?)
 * and
 *   ^SENB            instead of ^ENABREG
 */

int m10x_listen (struct gstatus * gs)
{
	return gigaset_configsequence(gs, "^SENB", 0, 0, NULL, 30);
}

int m10x_settype (struct gstatus * gs, int type)
{
	char cmd[]="^SBASCx";
	if (type<0||type>1) return -1;
	cmd[6]='0'+type;
	return gigaset_configsequence(gs, cmd, 0, 0, NULL, 30);
}

int m10x_register (struct gstatus * gs, const char * pin, int index)
{
	char cmd[80];
	size_t l;
	const char * pos;
	struct resp_filter *f;
	int r;

	if (index<0)
		return -1;

	l=strlen(pin);
	if (!l||l>8)
		return -1;

	for (pos=pin;*pos;++pos)
		if (!isdigit((unsigned char)*pos))
			return -1;

	f = gigaset_addfilter(gs, FILTER_FAILURE, 7+1, "FAILURE", 0, NULL);
	if (!f)
		return -ERR_MEM;

	r=snprintf(cmd, sizeof cmd, "^SOAC%s,%d", pin, index);
	if (r<0||r>=sizeof cmd)
		return -1;

	r = gigaset_configsequence(gs, cmd, 0, 0, NULL, 1000); //FIXME timeout

	gigaset_delfilter(gs, f);
	return r;
}

#if 0
int m10x_setmode (struct gstatus * gs, int index)
{
	return m10x_setvalue (gs, index, 5, "B");
}

int m10x_setrate (struct gstatus * gs, int index)
{
	return m10x_setvalue (gs, index, 10, "\\R");
}

int m10x_setformat (struct gstatus * gs, int index)
{
	return m10x_setvalue (gs, index, 10, "\\F");
}

int m10x_sethandshake (struct gstatus * gs, int index)
{
	return m10x_setvalue (gs, index, 6, "\\Q");
}
#endif

static inline int addcmd (char*s, const char * fmt, int val)
{
	return (val>=0)?sprintf(s, fmt, val):0;
}

int m10x_setstate (struct gstatus * gs, const struct m10x_state * mstate)
{
	char cmd[80];
	char * s;

	if (mstate->type>=0||mstate->partner>=0)
		return -1;
	if (mstate->mode>5||mstate->baudrate>10||mstate->format>10
	    ||mstate->flow>6||mstate->handshake>1)
		return -1;

	s=cmd;
	s += addcmd(s, "B%d",   mstate->mode);
	s += addcmd(s, "\\R%d", mstate->baudrate);
	s += addcmd(s, "\\F%d", mstate->format);
	s += addcmd(s, "\\P%d", mstate->handshake);
	s += addcmd(s, "\\Q%d", mstate->flow);

	if (s==cmd) return -1;
	return gigaset_configsequence(gs, cmd, 0, 0, NULL, 30);
}

int m10x_connected (struct gstatus * gs)
{
	int r;
	char * buf;

	r = m10x_connstate(gs, &buf);
	if (r<0) return r;
	if (!strcmp(buf,"0"))
		r=0;
	else if (!strcmp(buf,"1"))
		r=1;
	else
		r=-1;
	free(buf);
	return r;
}

int m10x_getstate (struct gstatus * gs, struct m10x_state * mstate)
{
	char * buf, * s, * e;
	int * val;
	int r;
	int pref=0;
	unsigned max;
	unsigned long l;

	r = gigaset_configsequence(gs, "&V", 1, 0, &buf, 30);
	if (r<0) return r;
	mstate->mode=-1;
	mstate->baudrate=-1;
	mstate->format=-1;
	mstate->handshake=-1;
	mstate->flow=-1;
	mstate->type=-1;
	mstate->partner=-1;

	r=-1;
	s=buf;
	while (*s)
	{
		val=NULL;
		switch (*s++)
		{
		case '\\':
			if (pref) goto exit;
			pref=1;
			break;
		case '#':
			if (pref) goto exit;
			pref=2;
			break;
		case 'B':
			pref=pref==0;
			max=5;
			val=&mstate->mode;
			break;
		case 'R':
			max=10;
			val=&mstate->baudrate;
			break;
		case 'F':
			max=10;
			val=&mstate->format;
			break;
		case 'P':
			max=1;
			val=&mstate->handshake;
			break;
		case 'Q':
			max=6;
			val=&mstate->flow;
			break;
		case 'T':
			pref=pref==2;
			max=1;
			val=&mstate->type;
			break;
		case 'D':
			pref=pref==2;
			max=INT_MAX;
			val=&mstate->partner;
			break;
		default:
			goto exit;
		}
		if (!val) continue;
		if(pref!=1||*val!=-1) goto exit;
		l=strtoul(s,&e,10);
		if (s==e||l>max) goto exit;
		*val=l;
		s=e;
		pref=0;
	}
	if(pref) goto exit;
	if (mstate->mode==-1||
	    mstate->baudrate==-1||
	    mstate->format==-1||
	    mstate->handshake==-1||
	    mstate->flow==-1||
	    mstate->type==-1||
	    mstate->partner==-1) goto exit;

	r=0;

exit:	free (buf);
	return r;
}
