/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <gigaset/interf.h>
#include "interf-private.h"
#include <gigaset/phonebook.h>
#include <gigaset/gigdata.h>
#include <gigaset/conf.h>

#if 0
ssize_t gigaset_createpbent (char ** result, const char * name, const char * number)
{
	//FIXME: 0x7c allowed in name??? (->byte stuffing!)
	size_t lnm,lno,i;
	ssize_t lbs;
	const char * c;
	char * buf, * bs;

	*result=NULL;
	lnm=strlen(name);
	lno=strlen(number);
	if (lnm>32||lno>16) return -ERR_LENGTH;
	for (c=name,i=0;i<32&&*c;++i,++c)
		if(*c==0x7c) return -ERR_DATA; // 0x7c makes problems...
	for (c=number;*c;++c)
		if((*c<'0'||*c>'9')&&*c!='#'&&*c!='*'&&*c!='P'&&*c!='R')
			return -ERR_DATA;
	//buf=malloc(lnm+1+lno);
	//if (!buf) return -ERR_MEM;

	lbs=gigaset_bytestuff(&bs, name, lnm, 0x7d, 0x7e /*0x7c*/, 0, 0); //FIXME byte stuffing of 0x7c=='|' not possible!? Must we stuff 0x7e??
	if (lbs<0) return lbs;

	buf=malloc(lbs+1+lno+4+1);
	if (!buf)
	{
		free (bs);
		return -ERR_MEM;
	}
	memcpy (buf, bs, lbs);
	free(bs);
	buf[lbs]='|';
	memcpy (buf+lbs+1, number, lno);

	gigaset_hexchecksum (buf+lbs+1+lno, buf,lbs+1+lno);//FIXME: checksum of unstuffed  data!?
	buf[lbs+1+lno+4]='\0';
	*result=buf;
	return lbs+1+lno+4;
}
#endif

static const struct pbcau_desc {
    unsigned type, value, level;
    const char *text;
} pbcau_table[] = {
    { 0, 0x10, GLOG_DEBUG, "normal call clearing" }
};

#define pbcau_table_END (pbcau_table + sizeof(pbcau_table)/sizeof((pbcau_table)[0]))


static void gigaset_logpbcau(struct glog *gl, const char *s)
{
	unsigned type, value;
	char * vptr;
	const struct pbcau_desc * ppd;

	if (memcmp(s, "ZPBCAU=", 7)) {
		gigaset_log(gl, GLOG_ERR, "invalid cause code: %s\n", s);
		return;
	}
	type = strtoul(s+7, &vptr, 16);
	if (*vptr != ',') {
		gigaset_log(gl, GLOG_ERR, "invalid cause code: %s\n", s);
		return;
	}
	value = strtoul(vptr+1, &vptr, 16);
	if (*vptr) {
		gigaset_log(gl, GLOG_ERR, "invalid cause code: %s\n", s);
		return;
	}
	for (ppd = pbcau_table; ppd < pbcau_table_END; ppd++)
		if (type == ppd->type && value == ppd->value) {
			gigaset_log(gl, ppd->level, "cause code: %s\n", ppd->text);
			return;
		}
	gigaset_log(gl, GLOG_WARN, "cause code: %x,%x\n", type, value);
}

static int pbinit(struct gstatus *gs, struct resp_filter *filter[3])
{
	filter[0] = gigaset_addfilter(gs, FILTER_ZPBEND, 6+1, "ZPBEND", 0, NULL);
	filter[1] = gigaset_addfilter(gs, FILTER_ZPBCAU, 7, "ZPBCAU=", 0, NULL);
	filter[2] = gigaset_addfilter(gs, FILTER_ZPBEOT, 6+1, "ZPBEOT", 0, NULL);

	if (filter[0] && filter[1] && filter[2])
		return 0;

	gigaset_delfilter(gs, filter[0]);
	gigaset_delfilter(gs, filter[1]);
	gigaset_delfilter(gs, filter[2]);

	return -ERR_MEM;
}

static void pbfinish(struct gstatus *gs, struct resp_filter *filter[3])
{
	gigaset_delfilter(gs, filter[0]);
	gigaset_delfilter(gs, filter[1]);
	gigaset_delfilter(gs, filter[2]);
}

static int gigaset_pbend (struct gstatus * gs)
{
	int r, needok, needend;
	char * resp;

	r = gigaset_sendcommand(gs, "AT^SPBEND");

	for (needok = needend = 1; r >= 0 && (needok || needend); )
	{
		r = gigaset_expectresponse(gs, "OK", 2+1, &resp, 30);
		if (r > 0)
			needok = 0;
		else if (r == -FILTER_ZPBEND) {
			needend = 0;
			r = 0;
		} else if (r == -FILTER_ZPBCAU) {
			gigaset_logpbcau(gs->gl, resp);
			r = 0;
		}
		free(resp);
	}

	return r;
}

static int gigaset_pbstart (struct gstatus * gs, const char * nmbr, char mode)
{
	size_t l;
	ssize_t r;
	char * cmd, * resp;

	l=strlen(nmbr);
	cmd=malloc(12+l+2+1);
	if (!cmd) return -ERR_MEM;
	memcpy(cmd,"AT^SPBSTART=",12);
	memcpy(cmd+12,nmbr, l);
	cmd[12+l+0]=',';
	cmd[12+l+1]=mode;
	cmd[12+l+2]='\0';
	r=gigaset_sendcommand(gs, cmd);
	free(cmd);
	if (r < 0) return r;

	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r < 0) return r;

	gs->may_interrupt = 1;
	r = gigaset_expectresponse(gs, "ZPBCONN", 7+1, &resp, 150);
	if (r >= 0) {
		free(resp);
		r = gigaset_expectresponse(gs, "ZPBREADY", 8+1, &resp, 600);
	}
	gs->may_interrupt = 0;

	if (r == -FILTER_ZPBCAU) {
		gigaset_logpbcau(gs->gl, resp);
		r = gigaset_expectresponse(gs, NULL, 0, NULL, 0);
		if (r != -FILTER_ZPBEND)
			r = -ERR_ABORT;
	}

	free(resp);
	if (r == -FILTER_ZPBEND)
		return -ERR_ABORT;
	if (r < 0)
		gigaset_pbend(gs);
	return r;
}

int gigaset_delpb (struct gstatus * gs, const char * nmbr)
{
	int r;
	int error = 0;
	struct resp_filter *filter[3];

	r = pbinit(gs, filter);
	if (r < 0)
		return r;

	r = gigaset_pbstart(gs, nmbr, 'D');
	if (r < 0)
		error = r;

	r = gigaset_expectresponse(gs, NULL, 0, NULL, 30);
	if (r != -FILTER_ZPBEOT)
		error = r;

	r = gigaset_pbend(gs);
	if (r < 0)
		error = r;

	pbfinish(gs, filter);
	return error;
}

int gigaset_sendpbarray (struct gstatus * gs, struct gphonebook * list)
{
	size_t l;
	ssize_t num,r;
	const char * src;
	char * pos, * s;
	struct gphonebook * cur;
	const char * prev=NULL;
	int error=0;
	char cmdstuffed[11+2*16+1+32+1+2*7+4+1]="AT^SPBDATA=";
	char cmd[16+1+32+1+7+4+1];
	int start=1;
	struct resp_filter *filter[3];

	r = pbinit(gs, filter);
	if (r < 0)
		return r;

	for (cur=list;;cur=cur->next)//FIXME:what happens, if user aborts transfer?
	{
		if (cur)
			gigaset_log(gs->gl, GLOG_DEBUG, "cur: %s %s %s\n", cur->name, cur->number, cur->pnr);
		else
			gigaset_log(gs->gl, GLOG_DEBUG, "cur: <NULL>\n");
		if (prev&&(!cur||error<0||strcmp(prev,cur->pnr)))
		{
			start=1;
			r=gigaset_sendcommand(gs, "AT^SPBEOT");
			if (r < 0)
				error=r;
			else
			{
				r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
				if (r < 0) error=r;
			}

			r=gigaset_pbend(gs);
			if (r < 0) error=r;
		}

		if (!cur||error<0) break;

		if (start)
		{
			start=0;
			prev=cur->pnr;

			r=gigaset_pbstart (gs, cur->pnr, 'W');
			if (r<0)
			{
				error=r;
				continue;
			}
		}

		l=strlen(cur->name);
		if (l>16)	/* Name seems max 16 byte long; Siemens says 32 byte */
			l=16;
		pos=cmd;
		for (num=0;num<l;++num)
			*pos++=*cur->name++;
		
		*pos++='|';
		
		for (src=cur->number,l=0;*src;++src)	/* Number seems max 32 byte long; Siemens says 16 byte */
			if (l<32&&((*src>='0'&&*src<='9')||*src=='*'||*src=='#'||*src=='R'||*src=='P')){
				*pos=*src;
				l++;
				pos++;
			}
		
		/* supplementary data */
		if (cur->melody >= 0) {
			gigaset_log(gs->gl, GLOG_DEBUG, "cur: %.2i\t%.2i\t%.2i:%.2i\t%.2i-%.2i-%.2i\n",
				cur->melody, cur->alert,
				cur->hour, cur->minute,
				cur->year, cur->month, cur->day);
			*pos++='|';
			*pos++=gigaset_itobcd8(cur->melody);
			*pos++=gigaset_itobcd8(cur->alert);	/* it seems to be ignored */
			*pos++=gigaset_itobcd8(cur->day);
			*pos++=gigaset_itobcd8(cur->month);
			*pos++=gigaset_itobcd8(cur->year);	/* it seems to be ignored */
			*pos++=gigaset_itobcd8(cur->hour);
			*pos++=gigaset_itobcd8(cur->minute);
			l+=8;
		}
		
		gigaset_hexchecksum (pos, cmd, num+1+l);
		pos+=4;
		*pos++='\0';

		pos=cmdstuffed+11;
		num=gigaset_stuffbuf(pos, cmd, num+1+l+4, 0x7d,  0x7e, 85); 
		if (num < 0) {
			error=-1;
			continue;
		}
		pos[num] = '\0';
		
		error=gigaset_sendcommand(gs, cmdstuffed);
		if (error < 0) continue;
		error=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
		if (error < 0) continue;
		//ZPBREPLY may take 7-8 sec to arrive, set ample timeout
		error=gigaset_expectresponse(gs, "ZPBRPLY=", 8, &s, 150);
		//FIXME test ZPBRPLY value for 0000
		free(s);
	}

	pbfinish(gs, filter);
	return error;
}

int gigaset_pbent (struct gstatus * gs, const char * nmbr, const char * name, const char * pnmbr)
{
	struct gphonebook pbent;
	pbent.next=NULL;
	pbent.name=name;
	pbent.number=pnmbr;
	pbent.pnr=nmbr;
	return gigaset_sendpbarray (gs, &pbent);
}

/* decode a phone book entry received from the base into a struct gphonebook
 * elements not present are set to NULL (char*) or -1 (int/char), resp.
 */
static int decode_pbdata(char *buf, size_t len, struct gphonebook **entry)
{
	ssize_t reallen;
	int32_t crc;
	char *npos;

	if (len < 4)
		return -ERR_DATA;

	crc = gigaset_fromhex4(buf + len - 4);
	if (crc < 0)
		return -ERR_DATA;

	reallen = gigaset_unstuffbuf(buf, buf, len-4, 0x7d, len-4);
	if (reallen < 0)
		return -ERR_DATA;

	if (!gigaset_verifychecksum(buf, reallen, crc))
		return -ERR_CRC;

	*entry = calloc(1, sizeof **entry);
	if (!*entry)
		return -ERR_MEM;

	buf[reallen] = 0;
	(*entry)->name = buf;
	(*entry)->melody = -1;
	(*entry)->alert = -1;
	(*entry)->day = -1;
	(*entry)->month = -1;
	(*entry)->year = -1;
	(*entry)->hour = -1;
	(*entry)->minute = -1;

	npos = memchr(buf, 0x7c, reallen);
	if (npos) {
		*npos = 0;
		(*entry)->number = npos + 1;

		npos = memchr(buf, 0x7c, reallen);
		if (npos) {
			*npos = 0;
			switch (reallen-(buf-npos)) {
			/* fall through cascade */
			default:
				/* extraneous bytes - ignore */
			case 7:
				(*entry)->minute = gigaset_bcd8toi(npos[7]);
			case 6:
				(*entry)->hour = gigaset_bcd8toi(npos[6]);
			case 5:
				(*entry)->year = gigaset_bcd8toi(npos[5]);
			case 4:
				(*entry)->month = gigaset_bcd8toi(npos[4]);
			case 3:
				(*entry)->day = gigaset_bcd8toi(npos[3]);
			case 2:
				(*entry)->alert = gigaset_bcd8toi(npos[2]);
			case 1:
				(*entry)->melody = gigaset_bcd8toi(npos[1]);
			case 0:
				;
			}
		}
	}
	return 0;
}

#define PTCAP_GAP	0x0000
#define PTCAP_4000C	0x7d4f
#define PTCAP_4000CwoSMS	0x7f4f	// w/o SMS menu
#define PTCAP_3000M	0x417f
#define PTCAP_3000S	0x4080

int gigaset_readpbarray(struct gstatus *gs, struct gphonebook **list, const char *pnmbr, int method)
{
	struct resp_filter *filter[3];
	int r;
	int error = 0;
	char *s;
	int aborted = 0;
	int send_pbend = 0;
	struct gphonebook *entry, *next;
	struct gphonebook start;
	size_t num = 0;
	ssize_t len;
	int timeout;
	int may_interrupt = 0;
	int oldfirmware;
	uint8_t tid;
	uint32_t ptcap;

	*list = NULL;
	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	oldfirmware = gs->ver_fwbase[0] < 2; /* this is guessed */
	if (oldfirmware)
		return -ERR_VERSION;

	/* select phonebook transfer method */
	if (method==PB_TRANSFER_AUTO){
		tid=atoi(pnmbr);
		r=gigaset_confread(gs, /*FIXME: id */ 483, tid, /* FIXME: datatype */ 2 & 0xf, &s);
		if (r<0) {
			return r;
		}
		if (r==4) {
			ptcap = gigaset_from_le32(s);
			free(s);
		}else{
			free(s);
			return -ERR_DATA;
		}
		gigaset_log(gs->gl, GLOG_DEBUG, "Party Capability (PtCap): %x\n", (unsigned)ptcap);
		switch (ptcap & 0xffff) {
			case PTCAP_GAP:
			case PTCAP_3000S:
				gigaset_log(gs->gl, GLOG_WARN, "Phonebook transfer not supported.\n");
				return -ERR_UNIMPLEMENTED;
				break;
			case PTCAP_4000C:
			case PTCAP_4000CwoSMS:
				method = PB_TRANSFER_NEW;
				gigaset_log(gs->gl, GLOG_DEBUG, "New transfer method invoked.\n");
				break;
			case PTCAP_3000M:
				method = PB_TRANSFER_OLD;
				gigaset_log(gs->gl, GLOG_DEBUG, "Old transfer method invoked.\n");
				gigaset_log(gs->gl, GLOG_WARN, "Start Phonebook transfer at your handset!\n");
				break;
			default:
				gigaset_log(gs->gl, GLOG_WARN, "PtCap %x unknown.\n", (unsigned)ptcap);
				gigaset_log(gs->gl, GLOG_WARN, "Please send your PtCap and handset type to gigaset307x-common@lists.sourceforge.net\n");
				return -ERR_UNIMPLEMENTED;
		}
	}

	r = pbinit(gs, filter);
	if (r < 0)
		return r;

	if (method==PB_TRANSFER_OLD) {
		
		may_interrupt = 1;
		timeout = GIG_INFTY;	/* wait for user action */
		//gs->sigint = 0;
		gs->may_interrupt = may_interrupt;
		error = gigaset_expectresponse(gs, "ZPBSTART", 8+1, NULL, timeout); /* wait of pb transfer from handset */
		gs->may_interrupt = 0;
		if (error >= 0)
			error = gigaset_sendcommand(gs, "AT^SPBCONN");
		if (error >= 0)
			error = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	} else if (method==PB_TRANSFER_NEW) {
		timeout = 150;		/* no user action necessary */
		error = gigaset_pbstart(gs, pnmbr, 'R');
	} else 
		return -ERR_PARAMINVALID;

	if (error >= 0)
		error = gigaset_sendcommand(gs, "AT^SPBREADY");
	if (error >= 0)
		error = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (error >= 0) {
		start.next = NULL;
		entry = &start;
		for (;;) {
			//gs->sigint = 0;
			gs->may_interrupt = may_interrupt;
			len = gigaset_expectresponse(gs, "ZPBDATA=", 8, &s, timeout);
			gs->may_interrupt = 0;
			if (len < 0) {
				if (len == -FILTER_ZPBCAU) {
					gigaset_logpbcau(gs->gl, s);
					send_pbend = 1;
					len = gigaset_expectresponse(gs, NULL, 0, NULL, 30);
				}
				free(s);
				if (len ==  -FILTER_ZPBEND) {
					send_pbend = 0;
				} else if (len ==  -FILTER_ZPBEOT) {
					send_pbend = 1;
					len = gigaset_expectresponse(gs, NULL, 0, &s, 5);
					if (len == -FILTER_ZPBCAU) {
						gigaset_logpbcau(gs->gl, s);
						len = gigaset_expectresponse(gs, NULL, 0, NULL, 5);
						if (len == -FILTER_ZPBEND) /* hangup and zpbend? */
							send_pbend = 0;
					}
					free(s);
				} else if (error >= 0)
					error = len;

				break;
			}
			if (aborted) {
				free(s);
			} else {
				r = decode_pbdata(s+8, len-8, &next);
				if (r >= 0) {
					entry->next = next;
					entry = next;
					entry->next = NULL;
					entry->buffer = s;
				} else {
					aborted = 1;
					if (error >= 0)
						error = r;
					free(s);
				}
			}
			++num;
			r = gigaset_sendcommand(gs, "AT^SPBRPLY=0000"); //FIXME
			if (r >= 0)
				r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
			if (r < 0) {
				if (error >= 0)
					error = r;
				break;
			}
		}
		*list = start.next;

		if (send_pbend) {
			r = gigaset_pbend(gs);
			if (r < 0 && error >= 0)
				error = r;
		}
	}

	pbfinish(gs, filter);
	if (error >= 0)
		return num;

	entry = *list;
	while (entry) {
		next = entry->next;
		free(entry->buffer);
		free(entry);
		entry = next;
	}
	return error;
}
