/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

/* WARNING: needs more testing, and more information from Siemens. */

#include <stdlib.h>
#include <string.h>
#include <gigaset/sms.h>
#include <gigaset/block.h>
#include <gigaset/log.h>
#include <gigaset/gigdata.h>
#include "interf-private.h"

/* gigaset only allows up to 157 characters (138 octets) */
#define SMS_MAX_OCTETS 138
#define SMS_MAX_CHARS  157

#define SMLIST_WORKAROUND

//#define GIG_TEST

/* If GIG_TEST is defined, the sms operations are only simulated: Every file
 * called ID.sm (with 1<=ID<=INT_MAX, no leading 0s) in the current working
 * directory is treated as a short message with id ID (note, that no file ID.tm
 * must exist).
 *
 * The first line of such a file contains a hexadecimal number (4
 * characters) interpreted as message status (AT^SSMLIST) and 15 to 34
 * characters message information (AT^SSMINFO). White space characters are
 * ignored.
 *
 * The other lines are interpreted as hex dump (2 characters for one octet)
 * of the message data (AT^SSMGETTXT/PUTTXT).
 */

#ifndef GIG_TEST
int gigaset_smsend (struct gstatus * gs)
{
	ssize_t r = 0;
	unsigned zsmend;

	zsmend = gs->temp_filter[FLT_ZSMEND]->status;

	gigaset_delfilter(gs, gs->temp_filter[FLT_ZSMCANCEL]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZSMEND   ]);

	if (!zsmend) {
		r = gigaset_sendcommand (gs, "AT^SSMEND");
		if (r >= 0) r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
		if (r >= 0) r = gigaset_expectresponse(gs, "ZSMEND", 6+1, NULL, 30);
		if (r >  0) r = 0;
	}

	gigaset_delfilter(gs, gs->temp_filter[FLT_ZSMERR   ]);

	return r;
}

int gigaset_smsstart(struct gstatus *gs)
{
	gs->temp_filter[FLT_ZSMERR   ] = gigaset_addfilter(gs, FILTER_ZSMERR,    6+1, "ZSMERR",    0, NULL);
	gs->temp_filter[FLT_ZSMCANCEL] = gigaset_addfilter(gs, FILTER_ZSMCANCEL, 9+1, "ZSMCANCEL", 0, NULL);
	gs->temp_filter[FLT_ZSMEND   ] = gigaset_addfilter(gs, FILTER_ZSMEND,    6+1, "ZSMEND",    0, NULL);

	if (gs->temp_filter[FLT_ZSMERR] && gs->temp_filter[FLT_ZSMCANCEL] && gs->temp_filter[FLT_ZSMEND])
		return 0;

	gigaset_delfilter(gs, gs->temp_filter[FLT_ZSMERR   ]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZSMCANCEL]);
	gigaset_delfilter(gs, gs->temp_filter[FLT_ZSMEND   ]);
	return -ERR_MEM;
}

int gigaset_smscontinue(struct gstatus *gs)
{
	ssize_t r;

	if (gs->temp_filter[FLT_ZSMCANCEL]->status || gs->temp_filter[FLT_ZSMEND]->status)
		return -ERR_ABORT;

	r = gigaset_sendcommand(gs, "AT^SSMSTART");
	if (r < 0)
		return r;
	r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r >= 0) r = gigaset_expectresponse(gs, "ZSMSTART", 8+1, NULL, 30);
	if (r >= 0) return 0;

	if (r == -FILTER_ZSMERR || r == -FILTER_ZSMCANCEL || r == -FILTER_ZSMEND)
		return -ERR_ABORT; //FIXME

	return r;
}

#ifdef SMLIST_WORKAROUND
/* try to detect and work around the bug in firmware release 4.204
 * where the result of the AT^SSMLIST command is garbled
 */
static int add_to_list(char *list, int l, uint32_t handle)
{
	int pos;

	for (pos = 0; pos < l; pos += 6) {
		if (handle == gigaset_from_be32(list+pos))
			return l;	/* already in */
	}
	gigaset_to_be32(list+l, handle);	/* append */
	gigaset_to_be16(list+l+4, 0);		/* no flags */
	return l+6;
}

static int fix_buggy_smlist(struct gstatus *gs, char *list, int listlen)
{
	int pos, l, r, iel;
	uint32_t handle;
	uint16_t flags;
	unsigned logid, lognum, opcode;
	struct glogrec *logrec;

	/* detect buggy response: it doesn't end at the end marker */
	for (pos = 0; pos+6 <= listlen; pos += 6) {
		handle = gigaset_from_be32(list+pos);
		flags  = gigaset_from_be16(list+pos+4);
		if (!handle && !flags) break;	/* end marker */
	}
	if (pos+6 == listlen) return listlen;	/* at end: ok */
	if (pos == 0) return 6;			/* empty: do nothing */

	/* buggy response detected: first handle is ok but flags are bogus */
	gigaset_to_be16(list+4, 0);
	/* remaining entries are completely bogus */
	l = 6;

	/* scan SMS log entries in basestation journal for missing handles */
	if (pos == 6) {
		lognum = 0;	/* only one handle: needn't scan */
	} else {
		r = gigaset_logstatus(gs, &logid, &lognum);
		if (r < 0) lognum = 0;	/* error: can't scan */
	}
	while (lognum-- > 0) {
		r = gigaset_logread(gs, logid--, &logrec);
		if (r < 0) break;

		switch (logrec->type) {
		case 12:	/* SMS state change */
			/* scan data elements for opcode */
			for (iel = 0; iel < logrec->nels; iel++) {
				if (logrec->el[iel].type == (ELTYPE_NUMBER|9)) {
					opcode = logrec->el[iel].num;
					break;
				}
			}
			/* ignore entry if "send" or "deleted" */
			if (opcode == 2 || opcode == 4)
				break;
			/* fall through */
		case 10:	/* incoming SM */
			/* scan data elements for handle */
			for (iel = 0; iel < logrec->nels; iel++) {
				if (logrec->el[iel].type == (ELTYPE_NUMBER|8)) {
					/* found handle: add to list */
					l = add_to_list(list, l, logrec->el[iel].num);
					if (l >= pos)
						/* all found: stop scan */
						lognum = 0;
					break;
				}
			}
			break;
		/* case 11:	-- outgoing SM ?? */
		}
		free(logrec);
	}

	gigaset_to_be32(list+l, 0);		/* set new end marker */
	gigaset_to_be16(list+l+4, 0);

	/* SM mode may have timed out */
	gigaset_smscontinue(gs);

	return l+6;
}
#else
static inline int fix_buggy_smlist(struct gstatus *gs, char *list, int listlen)
{
	return listlen;
}
#endif

ssize_t gigaset_smsgetlist (struct gstatus * gs, char ** buffer)
{
	int r;
	r=gigaset_sendcommand (gs, "AT^SSMLIST");
	if (r>=0) r=gigaset_expectresponse(gs,"OK",2+1,NULL,30);
	if (r>=0) r=gigaset_blockreceive (gs, 0x10, buffer);
	if (r>=0) r=fix_buggy_smlist(gs, *buffer, r);
	return r;
}

ssize_t gigaset_smsgettext  (struct gstatus * gs, char ** buffer)
{
	ssize_t r;
	r=gigaset_sendcommand (gs, "AT^SSMGETTXT");
	if (r>=0) r=gigaset_expectresponse(gs,"OK",2+1,NULL,30);
	if (r>=0) r=gigaset_blockreceive (gs, 0x10, buffer);
	return r;
}

int gigaset_smsputtext  (struct gstatus * gs, const char * buffer, size_t len)
{
	int r;
	r=gigaset_sendcommand (gs, "AT^SSMPUTTXT");
	if (r>=0) r=gigaset_expectresponse(gs,"OK",2+1,NULL,30);
	if (r>=0) r=gigaset_blocksend (gs, len, 0x20, buffer);
	return r;
}

static inline ssize_t gigaset_smscommand (struct gstatus * gs, const char * cmd, const char * rsp, size_t rsplen, char ** buf)
{
	ssize_t r;
	r=gigaset_sendcommand (gs, cmd);
	if (r>=0) r=gigaset_expectresponse(gs,"OK",2+1,NULL,30);
	if (r>=0) r=gigaset_expectresponse(gs,rsp,rsplen,buf,30);
	if (r>=0&&!buf) r=0;
	return r;
}

int gigaset_smsdel (struct gstatus * gs)
{
	return gigaset_smscommand (gs, "AT^SSMDEL", "ZSMDEL", 6+1, NULL);
}

int gigaset_smsold (struct gstatus * gs)
{
	return gigaset_smscommand (gs, "AT^SSMMAO", "ZSMMAO", 6+1, NULL);
}

int gigaset_smssend (struct gstatus * gs)
{
	return gigaset_smscommand (gs, "AT^SSMSEND", "ZSMSEND", 7+1, NULL);
}

int gigaset_smsstore (struct gstatus * gs)
{
	return gigaset_smscommand (gs, "AT^SSMSTORE", "ZSMSTORE", 8+1, NULL);
}

int gigaset_smsreset (struct gstatus * gs)
{
	return gigaset_smscommand (gs, "AT^SSMRESET", "ZSMRESET", 8+1, NULL);
}

static inline ssize_t gigaset_smsgetstring (struct gstatus * gs, const char * cmd, const char * rsp, size_t rsplen, char ** buf)
{
	ssize_t r;
	r=gigaset_smscommand(gs,cmd,rsp,rsplen,buf);
	if (r < rsplen) {
		free(*buf);
		return r<0 ? r : -ERR_RESPINVALID;
	}
	r -= rsplen;
	memmove(*buf, *buf+rsplen, r+1);
	return r;
}

static inline int gigaset_smsgetnumber (struct gstatus * gs, const char * cmd, const char * rsp, size_t rsplen, uint32_t * num)
{
	char * buf;
	ssize_t r;
	r=gigaset_smscommand(gs,cmd,rsp,rsplen,&buf);
	if (r < 0) {
		free(buf);
		return r;
	}
	r=gigaset_32fromdec (num,buf+rsplen);
	//if (r<0) r=-1; //FIXME
	return r;
}

int gigaset_smsnew  (struct gstatus * gs, uint32_t * num)
{
	return gigaset_smsgetnumber (gs, "AT^SSMNEW", "ZSMNEW=", 7, num);
}

int gigaset_smslevel  (struct gstatus * gs)
{
	uint32_t num;
	int r;
	r=gigaset_smsgetnumber (gs, "AT^SSMLEVEL", "ZSMLEVEL=", 9, &num);
	if (r<0) return r;
	if (num>100) return -ERR_RESPINVALID;
	return num;
}

/* WARNING: This only works for incomming messages (which is not mentioned
 * in the specs).
 */
int gigaset_smsselect (struct gstatus * gs, uint32_t num)
{
	char cmd[21]="AT^SSMSET=";
	gigaset_32todec(cmd+10,num);
	return gigaset_smscommand (gs, cmd, "ZSMSET", 6+1, NULL);
}

#if 0
/* does not work as expected */

int gigaset_smsdelhandle(struct gstatus * gs, uint32_t num)
{
	char cmd[21]="AT^SSMDEL=";
	gigaset_32todec(cmd+10,num);
	return gigaset_smscommand (gs, cmd, "ZSMDEL", 6+1, NULL);
}

int gigaset_smsoldhandle(struct gstatus * gs, uint32_t num)
{
	char cmd[21]="AT^SSMMAO=";
	gigaset_32todec(cmd+10,num);
	return gigaset_smscommand (gs, cmd, "ZSMMAO", 6+1, NULL);
}
#endif

int gigaset_smsnumber (struct gstatus * gs, const char * src)
{
	char * cmd;
	size_t l;
	int r;
	l=strlen (src);
	cmd=malloc(30+l+1);
	if (!cmd) return -ERR_MEM;
	memcpy(cmd,"AT^SSMNUM=",10);
	memcpy(cmd+10,src,l+1);
	r=gigaset_smscommand (gs, cmd, "ZSMNUM", 6+1, NULL);
	free(cmd);
	return r;
}

ssize_t gigaset_smsinfo (struct gstatus * gs, char ** buf)
{
	return gigaset_smsgetstring (gs, "AT^SSMINFO", "ZSMINFO=", 8, buf);
}

#else	/* def GIG_TEST */

#include <glob.h>
#include <stdio.h>
#include <limits.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

//#define MAXMSG 100
#define SMS_STARTFLAGS 0x1100

static int handle=-2;

int gigaset_smsend (struct gstatus * gs)
{
	/* FIXME find deleted msgs or msgs to send */
	handle=-2;
	return 0;
}

int gigaset_smsstart (struct gstatus * gs)
{
	if (handle==-2) handle=-1;
	return 0;
}

int gigaset_smscontinue(struct gstatus *gs)
{
	//FIXME
	return 0;
}

static inline uint8_t hexdig (char c)
{
	if (c>='0'&&c<='9')
		return c-'0';
	if (c>='a'&&c<='f')
		return c-'a'+10;
	return c-'A'+10;
}

ssize_t gigaset_smsgetlist (struct gstatus * gs, char ** buffer)
{
	glob_t globbuf;
	size_t i;
	char * pos;
	ssize_t r;
	unsigned long val;
	char * end;
	FILE * f;
	char lbuf[256];
	unsigned digits;
	uint16_t flags;
	char * lpos;

	if (handle==-2) return -ERR_RESPINVALID;

	//globbuf.gl_offs = 0;
	if(glob("[1-9]*.sm", 0, NULL, &globbuf))
		return -ERR_RESPINVALID;

	r=6*globbuf.gl_pathc+6;
	*buffer=malloc(r);
	if (!*buffer) {
		globfree (&globbuf);
		return -ERR_MEM;
	}

	for (i=0,pos=*buffer;i<globbuf.gl_pathc;++i,pos+=6) {
		val=strtoul(globbuf.gl_pathv[i], &end, 0);
		if (end==globbuf.gl_pathv[i]||strcmp(end,".sm")||val>INT_MAX) {
			r=-ERR_RESPINVALID;
			break;
		}
		f=fopen(globbuf.gl_pathv[i], "r");
		if (!f) {
			r=-ERR_RESPINVALID;
			break;
		}
		if (!fgets(lbuf, sizeof lbuf, f)||!strchr(lbuf, '\n')) {
			fclose (f);
			r=-ERR_RESPINVALID;
			break;
		}
		fclose(f);

		flags=0;
		digits=0;
		for (lpos=lbuf;*lpos;++lpos) {
			if(isspace((unsigned char)*lpos)) continue;
			if (!isxdigit((unsigned char)*lpos)) break;
			flags<<=4;
			flags|=(uint16_t)hexdig(*lpos);
			if (++digits==4) break;
		}
		if (digits!=4) {
			r=-ERR_RESPINVALID;
			break;
		}

		gigaset_to_be32(pos, val);
		gigaset_to_be16(pos+4, flags);
	}
	globfree (&globbuf);
	if (r<0)
		free (*buffer);
	else
		memset (pos, 0, 6);
	return r;
}

ssize_t gigaset_smsgettext  (struct gstatus * gs, char ** buffer)
{
	char name[80];
	int r;
	FILE *f;
	size_t s;
	char lbuf[256];
	char mbuf[140+4+4];
	char *dest,*src;
	int eol;
	char c,c2;

	if (handle<0) return -ERR_RESPINVALID;
	r=snprintf(name, sizeof name, "%d.sm", handle);
	if (r<0||(size_t)r>=sizeof name) return -ERR_RESPINVALID;

	f=fopen(name,"r");
	if (!f) return -ERR_RESPINVALID;
	if (!fgets(lbuf, sizeof lbuf, f)||!strchr(lbuf, '\n')) {
		fclose(f);
		return -ERR_RESPINVALID;
	}

	dest=mbuf;
	s=0;
	while (fgets(lbuf, sizeof lbuf, f)) {
		eol=0;
		src=lbuf;
		while (*src) {
			c=*src++;
			if (c=='\n') {
				eol=1;
				break;
			}
			if (isspace((unsigned char)c)) continue;
			c2=*src++;
			if (!isxdigit((unsigned char)c)||!isxdigit((unsigned char)c2)||s==sizeof mbuf) {
				fclose(f);
				return -ERR_RESPINVALID;
			}
			*dest++=hexdig(c)<<4|hexdig(c2);
			++s;
		}
		if (!eol) {
			fclose(f);
			return -ERR_RESPINVALID;
		}
	}
	fclose(f);

	*buffer=malloc(s);
	if (!*buffer) return -ERR_MEM;
	memcpy (*buffer, mbuf, s);
	return s;
}

static int opemmsg (FILE ** i, FILE ** o)
{
	char oname[100], iname[100];
	int r;
	int fd;

	if (handle<0) return -ERR_RESPINVALID;

	*o=NULL;
	*i=NULL;

	r=snprintf(iname, sizeof iname, "%d.sm", handle);
	if (r<0||(size_t)r>=sizeof iname) return -ERR_RESPINVALID;
	r=snprintf(oname, sizeof oname, "%d.tm", handle);
	if (r<0||(size_t)r>=sizeof oname) return -ERR_RESPINVALID;

	fd=open(oname,O_CREAT|O_EXCL|O_WRONLY,0666);
	if (fd<0) return -ERR_RESPINVALID;
	*o=fdopen(fd,"w");
	if (*o) *i=fopen(iname, "r");
	if (!*o||!*i) {
		close(fd);
		if (*o) fclose(*o);
		if (*i) fclose(*i);
		unlink(oname);
		return -ERR_RESPINVALID;
	}

	return 0;
}

static int copylines (FILE * i, FILE * o, int num)
{
	char lbuf [256];

	while (num && fgets(lbuf, sizeof lbuf, i)) {
		if (num>0) --num;
		if (!strchr(lbuf, '\n'))
			return -ERR_RESPINVALID;
		if (fputs(lbuf,o)==EOF)
			return -ERR_RESPINVALID;
	}
	if ((num<0 && ferror(i)) || num>0)
		return -ERR_RESPINVALID;
	return 0;
}

static int closemsg (FILE * i, FILE * o, int success)
{
	char oname[100], iname[100];
	int r;

	if (handle<0) return -ERR_RESPINVALID;

	r=0;
	fclose(i);
	if (fclose(o)==EOF) {
		success=0;
		r=-ERR_RESPINVALID;
	}

	r=snprintf(iname, sizeof iname, "%d.sm", handle);
	if (r<0||(size_t)r>=sizeof iname) return -ERR_RESPINVALID;
	r=snprintf(oname, sizeof oname, "%d.tm", handle);
	if (r<0||(size_t)r>=sizeof oname) return -ERR_RESPINVALID;

	if (success)
		if (rename(oname,iname)==-1) {
			success=0;
			r=-ERR_RESPINVALID;
		}

	if (!success)
		if (unlink(oname)==-1)
			r=-ERR_RESPINVALID;

	return r;
}

int gigaset_smsputtext  (struct gstatus * gs, const char * buffer, size_t len)
{
	int r;
	FILE *i,*o;
	size_t p;

	if (handle<0) return -ERR_RESPINVALID;
	r=opemmsg(&i,&o);
	if (r<0) return r;
	r=copylines(i,o,1);
	if (r<0) {
		closemsg (i, o, 0);
		return r;
	}

	r=0;
	for (p=0;p<len;++p) {
		r=fprintf (o, " %02x", (unsigned)(unsigned char)buffer[p]);
		if (r==EOF) break;
		if ((p&0xf)==0xf) {
			r=putc ('\n',o);
			if (r==EOF) break;
		}
	}
	if (r!=EOF&& p&0xf) r=putc ('\n',o);
	if (r==EOF) {
		closemsg (i, o, 0);
		return -ERR_RESPINVALID;
	}
	return closemsg (i, o, 1);
}

static char tohex[16]="0123456789abcdef";

static int setflag (uint16_t mask, uint16_t val)
{
	int r;
	FILE *i,*o;
	char lbuf [256];
	char *lpos;
	char * digpos[4];
	unsigned digits;
	uint16_t flags;

	if (handle<0) return -ERR_RESPINVALID;
	r=opemmsg(&i,&o);
	if (r<0) return r;

	if (!fgets(lbuf, sizeof lbuf, i)||!strchr(lbuf, '\n')) {
		closemsg (i, o, 0);
		return -ERR_RESPINVALID;
	}

	digits=0;
	flags=0;
	for (lpos=lbuf;*lpos;++lpos) {
		if (isspace((unsigned char)*lpos)) continue;
		if (!isxdigit((unsigned char)*lpos)) break;
		flags<<=4;
		flags|=(uint16_t)hexdig(*lpos);
		digpos[digits]=lpos;
		if (++digits==4) break;
	}
	if (digits!=4) {
		closemsg (i, o, 0);
		return -ERR_RESPINVALID;
	}
	flags=(flags&~mask)|val;
	*digpos[3]=tohex[flags & 0xf];
	flags >>=4;
	*digpos[2]=tohex[flags & 0xf];
	flags >>=4;
	*digpos[1]=tohex[flags & 0xf];
	flags >>=4;
	*digpos[0]=tohex[flags & 0xf];

	if (fputs(lbuf, o)==EOF) {
		closemsg (i, o, 0);
		return -ERR_RESPINVALID;
	}
	r=copylines(i,o,-1);
	if (r<0) {
		closemsg (i, o, 0);
		return r;
	}
	return closemsg (i, o, 1);
}

int gigaset_smsdelhandle(struct gstatus * gs, uint32_t num)
{
	int oldhandle = handle;
	int ret = gigaset_smsselect(gs, num);

	if (ret >= 0)
		ret = gigaset_smsdel(gs);
	
	gigaset_smsselect(gs, oldhandle);
	return ret;
}

int gigaset_smsoldhandle(struct gstatus * gs, uint32_t num)
{
	int oldhandle = handle;
	int ret = gigaset_smsselect(gs, num);

	if (ret >= 0)
		ret = gigaset_smsold(gs);
	
	gigaset_smsselect(gs, oldhandle);
	return ret;
}

int gigaset_smsdel (struct gstatus * gs)
{
	return setflag (SMS_DELETE, SMS_DELETE);
}

int gigaset_smsold (struct gstatus * gs)
{
	return setflag (SMS_NEW, 0);
}

int gigaset_smssend (struct gstatus * gs)
{
	return setflag (SMS_SEND, SMS_SEND);
}

int gigaset_smsstore (struct gstatus * gs)
{
	return setflag (SMS_STORE, SMS_STORE);
}

int gigaset_smsreset (struct gstatus * gs)
{
	return -ERR_RESPINVALID;
}

int gigaset_smsnew  (struct gstatus * gs, uint32_t * num)
{
	int newhandle;
	char namebuf[100];
	int r;
	int fd;
	FILE * f=NULL;

	if (handle==-2) return -ERR_RESPINVALID;

	for (newhandle=1;newhandle<=INT_MAX;++newhandle) {
		r=snprintf(namebuf, sizeof namebuf, "%d.sm", newhandle);
		if (r<0||(size_t)r>=sizeof namebuf) break;
		fd=open(namebuf,O_CREAT|O_EXCL|O_WRONLY,0666);
		if (fd<0) continue;
		f=fdopen(fd,"w");
		if (!f) {
			close(fd);
			unlink(namebuf);
		}
		break;
	}
	if (!f) return -ERR_RESPINVALID;

	r=fprintf (f, "%04x 04 14 11  07 04  F FFFF \n", SMS_STARTFLAGS); //FIXME
	if (r==EOF) newhandle=-ERR_RESPINVALID;
	r=fclose (f);
	if (r==EOF) newhandle=-ERR_RESPINVALID;
	if (newhandle<0)
		unlink(namebuf);
	else {
		*num=newhandle;
		handle=newhandle;
	}
	return newhandle;
}

int gigaset_smslevel  (struct gstatus * gs)
{
	return 0;
}

int gigaset_smsselect (struct gstatus * gs, uint32_t num)
{
	if (handle==-2) return -ERR_RESPINVALID;
	if (num>INT_MAX||!num) return -ERR_PARAMINVALID;
	handle=num;
	return 0;
}

int gigaset_smsnumber (struct gstatus * gs, const char * src)
{
	int r;
	FILE *i,*o;
	char lbuf [256];
	char *lpos;
	unsigned bytesleft;
	unsigned bytecount;
	size_t l;

	if (handle<0) return -ERR_RESPINVALID;
	l=strlen(src);
	if (l>19) return -ERR_RESPINVALID; /* FIXME or 20? */
	r=opemmsg(&i,&o);
	if (r<0) return r;

	if (!fgets(lbuf, sizeof lbuf, i)||!strchr(lbuf, '\n')) {
		closemsg (i, o, 0);
		return -ERR_RESPINVALID;
	}

	bytesleft=15+4;
	bytecount=0;
	for (lpos=lbuf;*lpos&&*lpos!='\n';++lpos,++bytecount) {
		if (isspace((unsigned char)*lpos)) continue;
		if (!bytesleft) break;
		--bytesleft;
	}
	if (bytesleft||bytecount+l+2>sizeof lbuf) {
		closemsg (i, o, 0);
		return -ERR_RESPINVALID;
	}
	memcpy(lpos,src,l);
	lpos+=l;
	*lpos++='\n';
	*lpos=0;

	if (fputs(lbuf, o)==EOF) {
		closemsg (i, o, 0);
		return -ERR_RESPINVALID;
	}
	r=copylines(i,o,-1);
	if (r<0) {
		closemsg (i, o, 0);
		return r;
	}
	return closemsg (i, o, 1);
}

ssize_t gigaset_smsinfo (struct gstatus * gs, char ** buf)
{
	char name[100];
	int r;
	FILE *f;
	size_t s;
	char lbuf[256];
	char *dest,*src;
	int eol;
	char c;
	int skip;

	if (handle<0) return -ERR_RESPINVALID;
	r=snprintf(name, sizeof name, "%d.sm", handle);
	if (r<0||(size_t)r>=sizeof name) return -ERR_RESPINVALID;

	f=fopen(name,"r");
	if (!f) return -ERR_RESPINVALID;
	if (!fgets(lbuf, sizeof lbuf, f)) {
		fclose(f);
		return -ERR_RESPINVALID;
	}
	fclose(f);
	s=0;
	dest=lbuf;
	src=lbuf;
	eol=0;
	skip=4;
	while (*src) {
		c=*src++;
		if (c=='\n') {
			eol=1;
			break;
		}
		if (isspace((unsigned char)c)) continue;
		if (skip)
			--skip;
		else {
			*dest++=c;
			++s;
		}
	}
	if (!eol||skip) return -ERR_RESPINVALID;
	*buf=malloc(s);
	if (!*buf) return -ERR_MEM;
	memcpy (*buf, lbuf, s);
	fprintf(stderr, "%d: (%d) <%.*s>\n", handle, (int) s, (int) s, *buf);
	return s;
}
#endif

int gigaset_smsputmessage(struct gstatus *gs, const struct gigaset_sm *msg)
{
	char buf[4 + 4 + 140];

	gigaset_to_be16(buf+0, msg->size + 4);
	gigaset_to_be16(buf+2, msg->characters);

	buf[4+0] = msg->flags;
	buf[4+1] = msg->unknown1;
	buf[4+2] = msg->dcs;
	buf[4+3] = msg->vp;

	memcpy(buf+8, msg->data, msg->size);

	return gigaset_smsputtext(gs, buf, msg->size+8);
}

/* frees the circular doubly linked lists *messages and *chains
 * (if not NULL) and sets the pointers to NULL */
void gigaset_smsfreemessages (struct gigaset_sm ** messages,
                              struct gigaset_sm_chain ** chains)
{
	struct gigaset_sm * m, * mt;
	struct gigaset_sm_chain * c, *ct;

	m=*messages;
	if (m) {
		do {
			mt=m;
			m=m->next;
			free (mt);
		} while (m!=*messages);
		*messages=NULL;
	}

	c=*chains;
	if (c) {
		do {
			ct=c;
			c=c->next;
			free (ct);
		} while (c!=*chains);
		*chains=NULL;
	}
}

/* reads UDH if present and sets udhl, udc, udv */
static void scanudh (struct gigaset_sm * msg)
{
	unsigned pos;
	unsigned len;

	if (!(msg->flags&SMS_TP_UDHI)||!msg->size) { /* FIXME warning if UDHI & !size */
		msg->udhl=0;
		msg->udc=0;
	} else {
		msg->udhl=msg->data[0]+1;
		if (msg->udhl > msg->size) msg->udhl=msg->size; /* FIXME warning? */

		pos=1;
		msg->udc=0;
		while (pos+1<msg->udhl) { /* warn if equal? */
			len=msg->data[pos+1];
			if (pos+2+len>msg->udhl) {
				/* warn? */
				break;
			}
			msg->udv[msg->udc]=msg->data+pos;
			++msg->udc;
			pos+=2+len;
		}
	}
}

/* if the msg is part of a concatenated message, add it to the appropriate chain */
static void concat (struct gigaset_sm * msg, struct gigaset_sm_chain ** chains)
{
	unsigned i;
	unsigned num;
	unsigned idx;
	unsigned id;
	struct gigaset_sm_chain * c, *cc;

	for (i=0;i<msg->udc;++i)
		if (msg->udv[i][0]==0&&msg->udv[i][1]>=3 /* warn if !=3 */
		    &&msg->udv[i][2+1]&&msg->udv[i][2+2]
		    &&msg->udv[i][2+2]<=msg->udv[i][2+1]) break;

	if (i==msg->udc) {
		msg->chain=NULL;
		return;
	}

	id=msg->udv[i][2+0];
	num=msg->udv[i][2+1];
	idx=msg->udv[i][2+2];

	c=*chains;
	cc=NULL;
	if (c) {
		do {
			if (id==c->chain_id
			    &&num==c->msgcount
			    &&!strcmp(msg->number,c->number)
			    &&!c->msg[idx-1]) { /* warn if msg[]!=NULL */
				cc=c;
				break;
			}
			c=c->next;
		} while (c!=*chains);
	}

	if (!cc) {
		cc=malloc(sizeof *cc);
		if (!cc) {
			/*warn*/
			msg->chain=NULL;
			return;
		}
		cc->chain_id=id;
		cc->msgcount=num;
		cc->missing=num;
		strcpy(cc->number,msg->number);
		for (i=0;i<num;++i)
			cc->msg[i]=NULL;

		if (!*chains) {
			cc->next=cc;
			cc->prev=cc;
			*chains=cc;
		} else {
			cc->next=*chains;
			cc->prev=(*chains)->prev;
			cc->next->prev=cc;
			cc->prev->next=cc;
		}
	}

	msg->chain=cc;
	msg->id=idx;

	--cc->missing;
	cc->msg[idx-1]=msg;
}

/* Read a message (including status information and "text"), append it to the
 * circular doubly linked list *messages, parse the UDHs, join concatenated
 * messages (returned in *chains).
 *
 * Parameters (in)
 *   list:     message list, as returned from gigaset_smsgetlist()
 *             (only the first entry is used)
 * Parameters (out)
 *   messages: circular doubly linked list of messages
 *   chains:   circular doubly linked list of concatenated messages
 * Returns:
 *   <0:       errorcode
 *   0:        no error
 */
int gigaset_smslist2msg (struct gstatus * gs, const unsigned char * list,
                         struct gigaset_sm ** messages,  struct gigaset_sm_chain ** chains)
{
	uint32_t handle;
	uint16_t status;
	unsigned val;
	struct gigaset_sm * msg=NULL;
	int r;
	ssize_t s;
	char * buf=NULL;
	static const char farray[4] = { 'F', 'F', 'F', 'F' }; // only works on systems with ASCII charset ;)

	handle = gigaset_from_be32(list);
	status = gigaset_from_be16(list+4);
	if (!handle&&!status) return -ERR_NOENT;

	msg=malloc(sizeof *msg);
	if (!msg) return -ERR_MEM;

	msg->handle=handle;
	msg->status=status;

	r=gigaset_smsselect(gs, handle);
	if (r<0) goto exit;

	s=gigaset_smsinfo (gs, &buf);
	if (s<0) {
		r=s;
		goto exit;
	}

	if (s<15||s>15+SMS_LEN_NUMBER||memchr(buf+15,0,s-15)) {
		gigaset_logbuf (gs->gl, GLOG_ERR, "invalid message information: ", buf, s);
		r=-ERR_DATA;
		goto exit;
	}
	memcpy (msg->number,buf+15,s-15);
	msg->number[s-15]=0;

	msg->minute = gigaset_fromdec2(buf + 0, &val) ? val : 255; //FIXME test for "FF"?
	msg->hour   = gigaset_fromdec2(buf + 2, &val) ? val : 255;
	msg->day    = gigaset_fromdec2(buf + 4, &val) ? val : 255;
	msg->month  = gigaset_fromdec2(buf + 6, &val) ? val : 255;
	msg->year   = gigaset_fromdec2(buf + 8, &val) ? val : 255;
	msg->MSN    = gigaset_fromdec1(buf[10], &val) ? val : 15;
	msg->error  = memcmp(buf+11, farray, 4) ? (gigaset_fromdec4(buf+11, &val) ? val : 0xfffe) : 0xffff; /*FIXME?*/;

	free (buf);
	buf=NULL;

	s=gigaset_smsgettext (gs, &buf);
	if (s<0) {
		r=s;
		goto exit;
	}
	if (s>140+4+4) s=140+4+4; /* FIXME warning? */
	if (s<4+4) {
		r=-ERR_DATA;
		goto exit;
	}
	msg->size=gigaset_from_be16(buf);
	msg->characters=gigaset_from_be16(buf+2);

	if (msg->size < 4 || (size_t)s-4 < msg->size) { /* FIXME warning if greater? */
		r=-ERR_DATA;
		goto exit;
	}
	msg->flags=buf[4];
	msg->unknown1=buf[5];
	msg->dcs=buf[6];
	msg->vp=buf[7];
	memcpy(msg->data, buf+4+4, msg->size-4);
	msg->size-=4;

	free(buf);
	buf=NULL;

	if (!*messages) {
		msg->next=msg;
		msg->prev=msg;
		*messages=msg;
	} else {
		msg->next=*messages;
		msg->prev=(*messages)->prev;
		msg->next->prev=msg;
		msg->prev->next=msg;
	}

	scanudh (msg);
	concat (msg, chains);

	return 0;

exit:	free (msg);
	free (buf);
	return r;
}

/* Read all messages (including status information and "text"), return them
 * as circular doubly linked list in *messages, parse the UDHs, join
 * concatenated messages (returned as circular doubly linked list in *chains).
 *
 * Parameters (out)
 *   messages: circular doubly linked list of messages
 *             previous content is not freed!
 *   chains:   circular doubly linked list of concatenated messages
 *             previous content is not freed!
 * Returns:
 *   <0:       errorcode
 *   0:        no error
 */
int gigaset_smsgetmessages (struct gstatus * gs, struct gigaset_sm ** messages,
                            struct gigaset_sm_chain ** chains)
{
	ssize_t s;
	char * list;
	size_t pos;
	int r;

	*messages=NULL;
	*chains=NULL;

	s=gigaset_smsgetlist (gs, &list);
	if (s<0) return s;

	r=-ERR_NOENT;
	for (pos=0;pos+5<(size_t)s;pos+=6) {
		r=gigaset_smslist2msg(gs, list+pos, messages, chains);
#ifdef SMLIST_WORKAROUND
		/* ZSMERR: invalid handle - expected with workaround */
		if (r == -FILTER_ZSMERR) r = 0;
#endif
		if (r<0) break;
	}

	free(list);
	if (r==-ERR_NOENT||r==0) return 0; // warn if 0?

	gigaset_smsfreemessages(messages, chains);
	return r;
}

static size_t decode_UCS2(unsigned char *dest, struct gigaset_sm *msg, unsigned *incomplete, unsigned char *prev, size_t destlen)
{
	unsigned octets;

	octets = msg->size;
	if (msg->characters < octets)
		octets = msg->characters;
	return gigaset_UCS2_to_utf8buf(msg->data + msg->udhl, octets - msg->udhl, dest, destlen, incomplete, prev);
}

static size_t decode_7bit(unsigned char * dest, struct gigaset_sm * msg, unsigned*esc, size_t destlen)
{
	char buf[160];
	unsigned septets;

	septets=msg->size+msg->size/7;
	if (msg->characters < septets) septets=msg->characters;

	septets=gigaset_unpackseptetbuf(buf,msg->data,septets,msg->udhl);
	return gigaset_septets_to_utf8buf (buf, septets, dest, destlen, esc);
}

static ssize_t encode_7bit(size_t * size, size_t * characters, const char * src, size_t srclen, char * dest,
                           size_t skip, unsigned * esc, char * escapedchar)
{
	char buf[160];
	ssize_t num;

	if (skip>=SMS_MAX_OCTETS) return -ERR_LENGTH;
	num=gigaset_utf8_to_septetbuf (src, srclen, buf, SMS_MAX_CHARS-skip-(skip+6)/7, esc, escapedchar, characters);
	if (num>=0) {
		*size=gigaset_packseptetbuf(dest, buf, *characters, skip);
		*characters+=skip+(skip+6)/7;
	}
	return num;
}

/* Decode SM text to UTF-8.
 * Parameters (in)
 *   msg:    SM to decode
 *   chain:  0:   decode this single message
 *           !=0: decode the whole chain (if a part is missing, the result
 *                is an empty string).
 * Parameters (out)
 *   buffer: pointer to a malloc()ed buffer (the initial value is ignored;
 *           only valid if no error occurs)
 *           The buffer holds the decoded message and one NUL-byte.
 * Returns:
 *   <0:     errorcode
 *   >=0:    size of the decoded message + 1
 */
ssize_t gigaset_sms2utf8 (struct gigaset_sm * msg, int chain, char ** buffer)
{
	unsigned encoding=0;
	size_t s,r,destlen;
	unsigned i;
	struct gigaset_sm_chain * c;
	char * pos;

	if (msg->dcs<SMS_DCS_TYPE_NONSTD)
		encoding=msg->dcs&SMS_DCS_EXTENCODING;
	else if ((msg->dcs&SMS_DCS_TYPE)==SMS_DCS_TYPE_F0)
		encoding=msg->dcs&SMS_DCS_ENCODING;
	else if ((msg->dcs&SMS_DCS_TYPE)==SMS_DCS_TYPE_MWI_UCS2)
		encoding=SMS_DCS_UCS2;

	if (chain)
		c=msg->chain;
	else
		c=NULL;

	if (!c) {
		s=msg->characters;
	} else {
		s=0;
		if (!c->missing)
			for (i=0;i<c->msgcount;++i)
				s+=c->msg[i]->characters;
	}
	s*=3; /* worst case for 16 bit unicode characters */
	destlen=s;
	++s; /* NUL */

	*buffer=malloc(s);
	if (!*buffer) return -ERR_MEM;

	if (encoding==SMS_DCS_UCS2) {
		unsigned char prev;
		unsigned incomplete = 0;
		if (!c) {
			s = decode_UCS2(*buffer, msg, &incomplete, &prev, destlen);
		} else {
			s = 0;
			pos = *buffer;
			if (!c->missing)
				for (i = 0; i<c->msgcount; ++i) {
					r = decode_UCS2(pos, c->msg[i], &incomplete, &prev, destlen);
					s += r;
					pos += r;
					destlen -= r;
				}
		}
		//FIXME warn if incomplete != 0
#if 0
	//FIXME
	} else if (encoding==SMS_DCS_8BIT) { /* we assume latin1... */
#endif
	} else { /* SMS_DCS_7BIT or unknown encoding */
		unsigned esc = 0;
		if (!c) {
			s=decode_7bit(*buffer,msg,&esc,destlen);
		} else {
			s=0;
			pos=*buffer;
			if (!c->missing)
				for (i=0;i<c->msgcount;++i) {
					r=decode_7bit(pos,c->msg[i],&esc,destlen);
					s+=r;
					pos+=r;
					destlen-=r;
				}
		}
		while (esc--)
			(*buffer)[s++]=0x20;
	}
	(*buffer)[s++]=0;
	//FIXME realloc
	return s;
}

/* Create SM from UTF-8 text.
 * Parameters (in)
 *   buffer:   text
 *   len:      length of the text
 *   chain:    0: no concatenated msgs
 *             1: concatenate if necessary (TP-UDHI will be set right)
 *             2: concatenate
 *   chain_id: chain id in the user data headers
 *   flags,
 *   unknown1,
 *   dcs,
 *   vp:       unknown header bytes
 *
 * Parameters (out)
 *   retmsg,
 *   retchain:
 *             pointers to message(s) and chain/NULL, malloc()ed  (the initial
 *             values are ignored; only valid if no error occurs)
 *
 * Returns:
 *   <0:       errorcode (e.g. invalid character, invalid length, out of memory)
 *   ==0:      no concatenated msgs
 *   >0:       number of messages
 */
int gigaset_smsfromutf8 (struct gigaset_sm ** retmsg, struct gigaset_sm_chain ** retchain,
                         uint8_t flags, uint8_t unknown1, uint8_t dcs, uint8_t vp,
                         int chain, uint8_t chain_id, const char * buffer, size_t len)
{
	struct gigaset_sm * msg, *prev;
	struct gigaset_sm_chain * c;
	unsigned encoding;
	ssize_t num;
	size_t characters, size;
	int r;
	unsigned i, msgcount;
	unsigned esc;
	char escapedchar;
	const char * pos;

	if (dcs<SMS_DCS_TYPE_NONSTD)
		encoding=dcs&SMS_DCS_EXTENCODING;
	else if ((dcs&SMS_DCS_TYPE)==SMS_DCS_TYPE_F0)
		encoding=dcs&SMS_DCS_ENCODING;
	else if ((dcs&SMS_DCS_TYPE)==SMS_DCS_TYPE_MWI_UCS2)
		encoding=SMS_DCS_UCS2;
	else if (  (dcs&SMS_DCS_TYPE)==SMS_DCS_TYPE_MWI
		 ||(dcs&SMS_DCS_TYPE)==SMS_DCS_TYPE_MWI_DISCARDMSG)
		encoding=SMS_DCS_7BIT;
	else
		return -ERR_DATA;

	if (chain<2) {
		msg=malloc (sizeof *msg);
		if (!msg) return -ERR_MEM;
		esc=0;
		switch (encoding) {
		case SMS_DCS_7BIT:	/* stuffed */
			num = encode_7bit(&size, &characters, buffer, len, msg->data, 0, &esc, &escapedchar);
			break;
		case SMS_DCS_8BIT:	/* we assume latin1... */
		case SMS_DCS_UCS2:	/* 2 bytes per char */
			//FIXME
		default:		/* unknown encoding */
			num = -ERR_DATA;
		}
		if (num<0||(size_t)num<len||esc) {
			free (msg);
			if (num < 0)
				return num;
			if (chain != 1)
				return -ERR_LENGTH;
			flags|=SMS_TP_UDHI;
		} else {
			if(chain==1) flags&=~SMS_TP_UDHI;
			msg->next=msg->prev=msg;
			msg->chain=0;
			msg->dcs=dcs;
			msg->flags=flags;
			msg->unknown1=unknown1;
			msg->vp=vp;
			msg->udhl=0;
			msg->udc=0;
			msg->size=size;
			msg->characters=characters;
			//FIXME rest?
			*retchain=NULL;
			*retmsg=msg;
			return 0;
		}
	}

	c=malloc (sizeof *c);
	if (!c) return -ERR_MEM;
	c->next=c->prev=c;
	c->chain_id=chain_id;
	c->number[0]=0;
	for (i=0;i<0xff;++i)
		c->msg[i]=NULL;
	c->missing=0;

	r=0;
	msgcount=0;
	prev=NULL;
	esc=0;
	pos=buffer;

	do {
		c->msg[msgcount]=msg=malloc(sizeof *msg);
		if (!msg) {
			r=-ERR_MEM;
			break;
		}

#if 0
		if (encoding==SMS_DCS_UCS2) {
		} else if (encoding==SMS_DCS_8BIT) { /* we assume latin1... */
		} else { /* SMS_DCS_7BIT or unknown encoding */
#endif
		if (encoding!=SMS_DCS_7BIT) {
			num=-ERR_DATA;//FIXME
		} else { //FIXME
			num=encode_7bit(&size, &characters, pos, len, msg->data, 6, &esc, &escapedchar);
		}
		if (num<0) {
			r=num;
			break;
		}
		len-=num;
		pos+=num;
		if(prev){
			prev->next=msg;
			msg->prev=prev;
		}
		msg->chain=c;
		msg->dcs=dcs;
		msg->flags=flags;
		msg->unknown1=unknown1;
		msg->vp=vp;
		msg->udhl=6;
		msg->udc=1;
		msg->udv[0]=msg->data+1;
		msg->size=size;
		msg->characters=characters;
		//FIXME rest?
		msg->id=msgcount+1;
		if(++msgcount==0xff) {
			r=-ERR_LENGTH;
			break;
		}
		prev=msg;
	} while (len||esc);

	if (r<0) {
		for (i=0;i<0xff;++i)
			free(c->msg[i]);
		free (c);
		return r;
	}

	msg=c->msg[0];
	msg->prev=prev;
	prev->next=msg;

	c->msgcount=msgcount;

	*retchain=c;
	*retmsg=msg;

	for (i=0;i<msgcount;++i) {
		msg=c->msg[i];
		msg->data[0]=5;
		msg->data[1]=0;
		msg->data[2]=3;
		msg->data[3]=chain_id;
		msg->data[4]=msgcount;
		msg->data[5]=i+1;
	}

	return msgcount;
}

