/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <QtGui>
#include "application.hh"
#include "smslist.hh"
#include "smscreator.hh"
#include "dial.hh"
#include "config.hh"
#include "logger.hh"

//#include <iostream>

ApplicationWindow::ApplicationWindow()
{
#if 0
	connect(&logger, SIGNAL(newMessage(unsigned int, QByteArray &)),
	        this, SIGNAL(logMessage(unsigned int, QByteArray &)),
	        Qt::QueuedConnection);
#endif
	connect(&logger, SIGNAL(newMessage(unsigned int, QByteArray)),
	        this, SIGNAL(logMessage(unsigned int, QByteArray)),
	        Qt::QueuedConnection);

	ws = new QWorkspace;
	setCentralWidget(ws);
	connect(ws, SIGNAL(windowActivated(QWidget *)), this, SLOT(updateMenus()));
	windowMapper = new QSignalMapper(this);
	connect(windowMapper, SIGNAL(mapped(QWidget *)), ws, SLOT(setActiveWindow(QWidget *)));

	createActions();
	createMenus();
	createToolBars();
	createStatusBar();
	updateMenus();

	readSettings();

	setWindowTitle(tr("QGigaset"));

	//loglist = new LogList(ws);
	loglist = new LogList();
	ws->addWindow(loglist);
	connect(this, SIGNAL(logMessage(unsigned int, QByteArray)),
		loglist, SLOT(newMessage(unsigned int, QByteArray)));
	//loglist->show();
	loglist->hide();
}

void ApplicationWindow::closeEvent(QCloseEvent *event)
{
	bool accept;

	ws->closeAllWindows();

	int num = ws->windowList().size();

	if (!num)
		accept = true;
	else if (num > 1)
		accept = false;
	else if (dynamic_cast<LogList*>(ws->activeWindow()))
		accept = true;
	else
		accept = false;

	if (!accept/*ws->activeWindow()*/) {
		event->ignore();
	} else {
	        writeSettings();
        	event->accept();
	}
}

void ApplicationWindow::smsNew()
{
	SmsCreator *w = new SmsCreator(gs);
	w->setAttribute(Qt::WA_DeleteOnClose);
	ws->addWindow(w);

	connect(w, SIGNAL(message(const QString&, int)), statusBar(), SLOT(message(const QString&, int)));
	w->show();
	//QCString QString::utf8 () const
}

void ApplicationWindow::devInfo()
{
	//TODO
}

void ApplicationWindow::showLogList()
{
	loglist->show();
}

void ApplicationWindow::dial()
{
	//FIXME allow only one dial window?
	DialWidget *w = new DialWidget(gs);
	w->setAttribute(Qt::WA_DeleteOnClose);
	ws->addWindow(w);

	connect(w, SIGNAL(message(const QString&, int)), statusBar(), SLOT(message(const QString&, int)));
	w->show();
}

void ApplicationWindow::config()
{
	//FIXME allow only one browser window?
	Config *w = new Config(gs);
	w->setAttribute(Qt::WA_DeleteOnClose);
	ws->addWindow(w);

	//w->setIcon( QPixmap("document.xpm") );
	connect(w, SIGNAL(message(const QString&, int)), statusBar(), SLOT(message(const QString&, int)));
	//w->load();
	w->show();
	//statusBar()->showMessage(tr("File loaded"), 2000);
}

void ApplicationWindow::smsBrowse()
{
	//FIXME allow only one browser window?
	SmsList *w = new SmsList(gs);
	w->setAttribute(Qt::WA_DeleteOnClose);
	ws->addWindow(w);

	//w->setIcon( QPixmap("document.xpm") );
	connect(w, SIGNAL(message(const QString&, int)), statusBar(), SLOT(message(const QString&, int)));
	w->load();
	w->show();
	//statusBar()->showMessage(tr("File loaded"), 2000);
}

void ApplicationWindow::closeDevice()
{
	int ret = gs.close();
	// TODO: ret
	qDebug("close: %d", ret);
}

DeviceDialog::DeviceDialog(QWidget *parent, GigasetQt &gsq)
	: QDialog(parent), gs(gsq)
{
	int ret;
	struct gigstrlist *list;

	cb = new QComboBox(this);
	cb->setEditable(true);

	ret = gs.enumerate_devices(&list);
	if (ret >= 0) {
		for (struct gigstrlist *cur = list; cur; cur = cur->next) {
			cb->addItem(cur->str);
		}
		gs.freestrlist(&list);
	}

	QPushButton *bOk = new QPushButton(tr("&Ok"), this);
	bOk->setDefault(true);
	connect(bOk, SIGNAL(clicked()), SLOT(accept()));

	QPushButton *bCancel = new QPushButton(tr("Cancel"), this);
	connect(bCancel, SIGNAL(clicked()), SLOT(reject()));

	QFrame *line = new QFrame(this);
	line->setFrameShape(QFrame::HLine);

	QHBoxLayout *buttonlayout = new QHBoxLayout;
	buttonlayout->addWidget(bOk);
	buttonlayout->addWidget(bCancel);
	buttonlayout->addStretch();

	QVBoxLayout *layout = new QVBoxLayout(this);
	layout->addWidget(cb);
	layout->addWidget(line);
	layout->addLayout(buttonlayout);

	setWindowTitle(tr("Choose device"));
	setLayout(layout);
}

QString DeviceDialog::chooseDevice(QWidget *parent, GigasetQt &gsq)
{
	DeviceDialog dd(parent, gsq);
	if (dd.exec() == QDialog::Accepted)
		return dd.cb->currentText();
	else
		return QString();
}

void ApplicationWindow::openDevice()
{
	int ret;

	if (gs.isopen())
		return; //FIXME msg

	QString dn = DeviceDialog::chooseDevice(this, gs);
	if (!dn.isEmpty()) {
		ret = gs.open(dn.toLocal8Bit(), logger.glogPtr()); //FIXME encoding?
		qDebug("open: %d/%d", ret, gs.get_sys_err());
		// TODO: ret
	}  else {
		statusBar()->showMessage(tr("Loading aborted"), 2000);
	}
}

ApplicationWindow::~ApplicationWindow()
{
	/*int ret =*/ gs.close();

	// TODO: ret
	// std::cerr << ret << std::endl;
}

#define stringify(x) str(x)
#define str(x) #x

void ApplicationWindow::about()
{
	QMessageBox::about(this, tr("About QGigaset"),
			   tr("<b>QGigaset</b>"
			      "<br>Gigaset graphical frontend tool"
			      "<br>from gigaset-frontend release ") +
			   stringify(VERSION) +
		tr("<br>Copyright &copy; 2006 Hansjoerg Lipp") +
		tr("<br>Copyright &copy; 2010 Tilman Schmidt"));
}

void ApplicationWindow::updateMenus()
{
	bool hasChild = (ws->activeWindow() != 0); //FIXME
	closeAct->setEnabled(hasChild);
	closeAllAct->setEnabled(hasChild);
	tileAct->setEnabled(hasChild);
	cascadeAct->setEnabled(hasChild);
	//arrangeAct->setEnabled(hasChild);
	nextAct->setEnabled(hasChild);
	previousAct->setEnabled(hasChild);
	separatorAct->setVisible(hasChild);//FIXME <-whut
}

void ApplicationWindow::updateWindowMenu()
{
	windowMenu->clear();
	windowMenu->addAction(closeAct);
	windowMenu->addAction(closeAllAct);
	windowMenu->addSeparator();
	windowMenu->addAction(tileAct);
	windowMenu->addAction(cascadeAct);
	//windowMenu->addAction(arrangeAct);
	windowMenu->addSeparator();
	windowMenu->addAction(nextAct);
	windowMenu->addAction(previousAct);
	windowMenu->addAction(separatorAct);

	QList<QWidget *> windows = ws->windowList();
	separatorAct->setVisible(!windows.isEmpty());

	for (int i = 0; i < windows.size(); ++i) {
		QWidget *child = windows.at(i);
		QString text;
		if (i < 9) {
			text = tr("&%1 %2").arg(i + 1)
				.arg(child->windowTitle());
		} else {
			text = tr("%1 %2").arg(i + 1)
				.arg(child->windowTitle());
		}
		QAction *action  = windowMenu->addAction(text);
		action->setCheckable(true);
		action->setChecked(child == ws->activeWindow());
		connect(action, SIGNAL(triggered()), windowMapper, SLOT(map()));
		windowMapper->setMapping(action, child);
	}
}

void ApplicationWindow::createActions()
{
	devOpenAct = new QAction(tr("&Open device..."), this);
	devOpenAct->setShortcut(tr("Ctrl+O"));
	devOpenAct->setStatusTip(tr("Open Gigaset device"));
	connect(devOpenAct, SIGNAL(triggered()), this, SLOT(openDevice()));

	devCloseAct = new QAction(tr("&Close device"), this);
	devCloseAct->setStatusTip(tr("Close Gigaset device"));
	connect(devCloseAct, SIGNAL(triggered()), this, SLOT(closeDevice()));

	//devInfoAct = new QAction(tr("Device &info..."), this); //FIXME disable if device not available
	//devInfoAct->setStatusTip(tr("Display information about the device"));
	//connect(devInfoAct, SIGNAL(triggered()), this, SLOT(devInfo()));

	configAct = new QAction(tr("Configure device..."), this);
	configAct->setStatusTip(tr("Configure device"));
	connect(configAct, SIGNAL(triggered()), this, SLOT(config()));

	dialAct = new QAction(tr("&Dial..."), this);
	dialAct->setStatusTip(tr("Dial"));
	connect(dialAct, SIGNAL(triggered()), this, SLOT(dial()));

	smsBrowseAct = new QAction(tr("&Browse"), this);
	smsBrowseAct->setStatusTip(tr("Open SMS browser window"));
	connect(smsBrowseAct, SIGNAL(triggered()), this, SLOT(smsBrowse()));

	smsNewAct = new QAction(tr("&New"), this);
	smsNewAct->setStatusTip(tr("Compose SM"));
	connect(smsNewAct, SIGNAL(triggered()), this, SLOT(smsNew()));

	logListAct = new QAction(tr("&Log..."), this);
	logListAct->setStatusTip(tr("show log messages"));
	connect(logListAct, SIGNAL(triggered()), this, SLOT(showLogList()));

	closeAct = new QAction(tr("Cl&ose"), this);
	closeAct->setShortcut(tr("Ctrl+F4"));
	closeAct->setStatusTip(tr("Close the active window"));
	connect(closeAct, SIGNAL(triggered()), ws, SLOT(closeActiveWindow()));

	closeAllAct = new QAction(tr("Close &All"), this);
	closeAllAct->setStatusTip(tr("Close all the windows"));
	connect(closeAllAct, SIGNAL(triggered()), ws, SLOT(closeAllWindows()));

#if 0
	cutAct = new QAction(tr("Cu&t"), this);
	cutAct->setShortcut(tr("Ctrl+X"));
	cutAct->setStatusTip(tr("Cut the current selection's contents to the clipboard"));
	connect(cutAct, SIGNAL(triggered()), this, SLOT(cut()));

	copyAct = new QAction(tr("&Copy"), this);
	copyAct->setShortcut(tr("Ctrl+C"));
	copyAct->setStatusTip(tr("Copy the current selection's contents to the clipboard"));
	connect(copyAct, SIGNAL(triggered()), this, SLOT(copy()));

	pasteAct = new QAction(tr("&Paste"), this);
	pasteAct->setShortcut(tr("Ctrl+V"));
	pasteAct->setStatusTip(tr("Paste the clipboard's contents into the current selection"));
	connect(pasteAct, SIGNAL(triggered()), this, SLOT(paste()));
#endif

	tileAct = new QAction(tr("&Tile"), this);
	tileAct->setStatusTip(tr("Tile the windows"));
	connect(tileAct, SIGNAL(triggered()), ws, SLOT(tile()));

	cascadeAct = new QAction(tr("&Cascade"), this);
	cascadeAct->setStatusTip(tr("Cascade the windows"));
	connect(cascadeAct, SIGNAL(triggered()), ws, SLOT(cascade()));

#if 0
	arrangeAct = new QAction(tr("Arrange &icons"), this);
	arrangeAct->setStatusTip(tr("Arrange the icons"));
	connect(arrangeAct, SIGNAL(triggered()), ws, SLOT(arrangeIcons()));
#endif

	nextAct = new QAction(tr("Ne&xt"), this);
	nextAct->setShortcut(tr("Ctrl+F6"));
	nextAct->setStatusTip(tr("Move the focus to the next window"));
	connect(nextAct, SIGNAL(triggered()), ws, SLOT(activateNextWindow()));

	previousAct = new QAction(tr("Pre&vious"), this);
	previousAct->setShortcut(tr("Ctrl+Shift+F6"));
	previousAct->setStatusTip(tr("Move the focus to the previous window"));
	connect(previousAct, SIGNAL(triggered()), ws, SLOT(activatePreviousWindow()));

	separatorAct = new QAction(this);
	separatorAct->setSeparator(true);

	aboutAct = new QAction(tr("&About"), this);
	aboutAct->setStatusTip(tr("Show the application's About box"));
	connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));

	aboutQtAct = new QAction(tr("About &Qt"), this);
	aboutQtAct->setStatusTip(tr("Show the Qt library's About box"));
	connect(aboutQtAct, SIGNAL(triggered()), qApp, SLOT(aboutQt()));

	exitAct = new QAction(tr("E&xit"), this);
	exitAct->setShortcut(tr("Ctrl+Q"));
	exitAct->setStatusTip(tr("Exit the application"));
	connect(exitAct, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));
}

void ApplicationWindow::createToolBars()
{
}

void ApplicationWindow::createMenus()
{
	deviceMenu = menuBar()->addMenu(tr("&Device"));
	deviceMenu->addAction(devOpenAct);
	deviceMenu->addAction(devCloseAct);
	//deviceMenu->addAction(devInfoAct);
	deviceMenu->addAction(dialAct);
	deviceMenu->addAction(configAct);
	deviceMenu->addSeparator();
	deviceMenu->addAction(logListAct);
	deviceMenu->addSeparator();
	deviceMenu->addAction(exitAct);

#if 0
	editMenu = menuBar()->addMenu(tr("&Edit"));
	editMenu->addAction(cutAct);
	editMenu->addAction(copyAct);
	editMenu->addAction(pasteAct);
#endif

	smsMenu = menuBar()->addMenu(tr("&SMS"));
	smsMenu->addAction(smsBrowseAct);
	smsMenu->addAction(smsNewAct);

	windowMenu = menuBar()->addMenu(tr("&Window"));
	connect(windowMenu, SIGNAL(aboutToShow()), this, SLOT(updateWindowMenu()));

	menuBar()->addSeparator();

	helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(aboutAct);
	helpMenu->addAction(aboutQtAct);
}

void ApplicationWindow::createStatusBar()
{
	statusBar()->showMessage(tr("Ready"));
}

void ApplicationWindow::readSettings()
{
	//QSettings settings(QSettings::IniFormat, QSettings::UserScope, "gigaset307x.sf.net", "QGigaset");
	QSettings settings("gigaset307x.sf.net", "QGigaset");
	QPoint pos = settings.value("pos", QPoint(200, 200)).toPoint();
	QSize size = settings.value("size", QSize(400, 400)).toSize();
	move(pos);
	resize(size);
}

void ApplicationWindow::writeSettings()
{
	//QSettings settings(QSettings::IniFormat, QSettings::UserScope, "gigaset307x.sf.net", "QGigaset");
	QSettings settings("gigaset307x.sf.net", "QGigaset");
	settings.setValue("pos", pos());
	settings.setValue("size", size());
}
