/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <QtGui>
#include "config.hh"
#include "gigasetqt.hh"
#include "main.hh"
#include <iostream>
#include <limits>
#include <algorithm>

// file://localhost/usr/local/lib/qt4/doc/html/itemviews-simpletreemodel.html
// file://localhost/usr/local/lib/qt4/doc/html/model-view-model-subclassing.html
// file://localhost/usr/local/lib/qt4/doc/html/qabstractitemmodel.html
// file://localhost/usr/local/lib/qt4/doc/html/model-view-delegate.html
// /usr/local/lib/qt4/examples/itemviews/spinboxdelegate/main.cpp

// file://localhost/usr/local/lib/qt4/doc/html/qtreeview.html#expand

bool Range::contains(int val) const
{
	//std::cerr << "testing " << val << std::endl;

	QList<struct Interval>::const_iterator it;

	for (it = ilist.constBegin(); it != ilist.constEnd(); ++it) {
		//std::cerr << "    " << it->min << "-" << it->max << std::endl;
		if (it->max < val)
			continue;
		if (val < it->min)
			return false;
		//std::cerr << "  => true" << std::endl;
		return true;
	}
	return false;
}

void Range::add(int min, int max)
{
	if (max < totmin || min > totmax)
		return;

	if (min < totmin)
		min = totmin;
	if (max > totmax)
		max = totmax;

	QList<struct Interval>::iterator it;

	for (it = ilist.begin(); it != ilist.end(); ++it)
		if (max > it->min)
			break;

	if (it == ilist.end() || min > it->max) {
		struct Interval iv = {min, max};
		ilist.insert(it, iv);
	} else { //FIXME overlapping intervals!
		if (min < it->min)
			it->min = min;
		if (max > it->max)
			it->max = max;
	}
}

bool Range::parsePart(const QString &str, int &val, int &pos, bool &done, int fac)
{
	QString cur;

	//std::cerr << "parsing " << pos << " in " << str.toLocal8Bit().constData() << std::endl;

	int end = str.indexOf('.', pos);
	if (end == -1) {
		cur = str.mid(pos);
		done = true;
	} else {
		cur = str.mid(pos, end - pos);
		pos = end + 1;
		if (pos == str.size())
			return false;
	}

	bool valid;
	int temp = cur.toInt(&valid);
	if (!valid)
		return false;

	val += temp * fac;

	//std::cerr << " ---> " << val << std::endl;

	return true;
}

bool Range::parseNum(const QString &str, int &val, int deflt, int fac1, int fac2, int fac3)
{
	int pos = 0;
	bool done = false;

	if (str.trimmed().isEmpty()) {
		val = deflt;
		return true;
	}

	val = 0;

	if (!parsePart(str, val, pos, done, fac1))
		return false;
	if (done)
		return true;

	if (!parsePart(str, val, pos, done, fac2))
		return false;
	if (done)
		return true;

	if (!parsePart(str, val, pos, done, fac3))
		return false;
	if (done)
		return true;

	return false;
}

bool Range::parse(const QString &str, int fac1, int fac2, int fac3)
{
	int pos = 0;
	bool done = false;
	do {
		QString cur;
		int end = str.indexOf(',', pos);
		if (end == -1) {
			cur = str.mid(pos);
			done = true;
		} else {
			cur = str.mid(pos, end - pos);
			pos = end + 1;
			if (pos == str.size())
				return false;
		}
		int sep = cur.indexOf('-');
		if (sep == -1)
			return false;
		int min, max;
		if (!parseNum(cur.left(sep), min, totmin, fac1, fac2, fac3))
			return false;
		if (!parseNum(cur.mid(sep+1), max, totmax, fac1, fac2, fac3)) // FIXME valid if sep+1==cur.size()?
			return false;
		add(min, max);
	} while (!done);
	return true;
}

bool Range::next(int &val) const
{
	QList<struct Interval>::const_iterator it;
	for (it = ilist.constBegin(); it != ilist.constEnd(); ++it) {
		if (it->max > val) {
			if (it->min > val)
				val = it->min;
			else
				++val;
			return true;
		}
	}
	return false;
}

bool Range::prev(int &val) const
{
	QList<struct Interval>::const_iterator it;
	for (it = ilist.constEnd(); it != ilist.constBegin(); ) {
		--it;
		if (it->min < val) {
			if (it->max < val)
				val = it->max;
			else
				--val;
			return true;
		}
	}
	return false;
}

ConfigItem::ConfigItem(ConfigItem *parent, const QString & name, bool node, bool templ)
	: key(name), parentItem(parent), is_node(node), is_template(templ),
	  dtype(DTInvalid), id(-1), tid(0), read_only(false),
	  tplt_delta(1), tplt_tiddelta(0), tplt_tid(0), expanded(false), val(0)
{
	buddyList[0] = 0;
	buddyList[1] = 0;
	fw_range.add(std::numeric_limits<int>::min(), std::numeric_limits<int>::max());
}

ConfigItem::~ConfigItem()
{
	qDeleteAll(childItems);
	//delete val; // gets too ugly
}

void ConfigItem::appendChild(ConfigItem *item)
{
	//std::cerr << "append " << item->key.toLocal8Bit().constData() << " to " << key.toLocal8Bit().constData() << std::endl;
	childItems.append(item);
}

void ConfigItem::removeChild(ConfigItem *item)
{
	childItems.removeAll(item);
}

ConfigItem *ConfigItem::findChild(const QString &key) const
{
	QList<ConfigItem*>::const_iterator it;
	for (it = childItems.constBegin(); it != childItems.constEnd(); ++it)
		if ((*it)->key == key)
			return *it;
	return 0;
}

ConfigItem *ConfigItem::child(int row)
{
	return childItems.value(row);
}

int ConfigItem::childCount() const
{
	return childItems.count();
}

int ConfigItem::row() const
{
	if (parentItem)
		return parentItem->childItems.indexOf(const_cast<ConfigItem*>(this));

	return 0;
}

bool ConfigItem::setData(const QVariant &value)
{
	user_data = value;
	return true;
}

QVariant ConfigItem::data(int column) const
{
	if (column < 0 || column > 1)
		return QVariant();

	if (!column)
		return key;
	if (user_data.isValid())
		return user_data;
	return device_data;
}

ConfigItem *ConfigItem::parent()
{
	return parentItem;
}

bool ConfigItem::isValid() const
{
	if (is_template)
		return !tplt_range.isEmpty();

	if (is_node)
		return true;

	return dtype != DTInvalid && id >= 0;
}

ConfigItem::State ConfigItem::state() const
{
	if (!user_data.isValid() || device_data == user_data)
		return SNone;
	if (!device_data.isValid())
		return SIgnore;
	if (val && !val->isValid(*this))
		return SInvalid;
	return SValid;
}

bool ConfigItem::fillGconf(struct gconf &gc, int version, bool data) const
{
	//std::cerr << "gconf " << key.toLocal8Bit().constData() << std::endl;

	if (!fw_range.contains(version))
		return false;

	if (data && state() != SValid)
		return false;

	gc.buf = 0;
	gc.len = 0;

	switch (dtype) {
	case DTByte:
		gc.datatype = 3;
		if (data) {
			gc.len = 1;
			gc.buf = new char[1];
			*gc.buf = user_data.toInt(); //FIXME range
		}
		break;
	case DTString:
		gc.datatype = 2;
		if (data) {
			const QByteArray str = user_data.toString().toLatin1(); //FIXME encoding //FIXME check value
			if (str.size() > 0xfffe)
				return false;
			gc.len = str.size() + 1;
			gc.buf = new char[gc.len];
			const char *strdata = str.constData();
			std::copy(strdata, strdata + gc.len, gc.buf);
		}
		break;
	default:
		return false;
	}

	gc.id = id;
	gc.tid = tid;

	//std::cerr << "  id " << id << std::endl;
	
	return true;
}

void ConfigItem::fillGconfList(struct gconf *&gc, QList<ConfigItem *> &empty, int version, bool data)
{
	//if (clear_data) {
	//	has_data = false;
	//	itemData[0] = QVariant();
	//}

	struct gconf tmp;
	if (fillGconf(tmp, version, data)) {
		struct gconf *cur = new struct gconf;
		*cur = tmp;
		cur->priv.ptr = this;
		cur->next = gc;
		gc = cur;
	} else {
		empty.append(this);
	}

	QList<ConfigItem*>::const_iterator it;
	for (it = childItems.constBegin(); it != childItems.constEnd(); ++it)
		(*it)->fillGconfList(gc, empty, version, data);
}

void ConfigItem::reset(QList<ConfigItem *> &itemlist)
{
	if (user_data.isValid()) {
		user_data = QVariant();
		itemlist.append(this);
	}

	QList<ConfigItem*>::const_iterator it;
	for (it = childItems.constBegin(); it != childItems.constEnd(); ++it)
		(*it)->reset(itemlist);
}

bool ConfigItem::clear()
{
	bool new_value = device_data.isValid();

	device_data = QVariant();

	return new_value;
}

bool ConfigItem::readGconf(const struct gconf &gc)
{
	QVariant data;

	switch (dtype) {
	case DTByte:
		data = QVariant(static_cast<unsigned>(gc.buf[0]));
		break;
	case DTString:
		data = QVariant(QString::fromLatin1(gc.buf)); //FIXME encoding // FIXME len
		break;
	default:
		return false;
	}

	bool new_value = !device_data.isValid() || device_data != data;
	device_data = data;

	return new_value;
}

void ConfigItem::debug(const QString &prefix) const
{
	std::cerr << prefix.toLocal8Bit().constData() << "[" << key.toLocal8Bit().constData() << "]";
	if (parentItem)
		std::cerr << " {" << parentItem->key.toLocal8Bit().constData() << "}";
	std::cerr << std::endl;

	QList<ConfigItem*>::const_iterator it;
	for (it = childItems.constBegin(); it != childItems.constEnd(); ++it)
		(*it)->debug(prefix + "  ");
}

void ConfigItem::cloneChildren(ConfigItem *parent, ConfigItem *tmpl, int index)
{
	/* copy children to new parent, using tmpl and index for calculating tid, etc. */
	QList<ConfigItem*>::const_iterator it;
	for (it = childItems.constBegin(); it != childItems.constEnd(); ++it) {
		ConfigItem *cur = new ConfigItem(parent, (*it)->key, (*it)->is_node, (*it)->is_template);

		if (!cur->is_node) {
			cur->read_only  = (*it)->read_only;
			cur->dtype      = (*it)->dtype;
			cur->id         = (*it)->id + index * tmpl->tplt_delta;
			cur->tid        = tmpl->tplt_tid + index * tmpl->tplt_tiddelta;
			cur->fw_range   = (*it)->fw_range;
			cur->val        = (*it)->val;
			//std::copy((*it)->buddyList, (*it)->buddyList+2, cur->buddyList);
		}

		if (cur->is_template) {
			cur->tplt_range    = (*it)->tplt_range;
			cur->tplt_delta    = (*it)->tplt_delta;
			cur->tplt_tiddelta = (*it)->tplt_tiddelta;
			cur->tplt_tid      = (*it)->tplt_tid;
			cur->expanded      = (*it)->expanded;
		}
		parent->appendChild(cur);
		(*it)->cloneChildren(cur, tmpl, index);
	}
}

bool ConfigItem::resolvePath(ConfigItem *root, QStringList &Path, ConfigItem *&buddy)
{
	//FIXME
	buddy = 0;
	return true;
}

bool ConfigItem::resolvePaths(ConfigItem *root)
{
	bool ret = true;
	for (int i = 0; i < 2; ++i) {
		if (!resolvePath(root, buddyPathList[i], buddyList[i]))
			ret = false;
		buddyPathList[i].clear();
	}
	QList<ConfigItem*>::const_iterator it;
	for (it = childItems.constBegin(); it != childItems.constEnd(); ++it) {
		if (!(*it)->resolvePaths(root))
			ret = false;
	}
	return ret;
}

void ConfigItem::expand() //FIXME buddyPaths
{
	//std::cerr << "expand" << std::endl;
	if (!is_template || expanded) {
		QList<ConfigItem*>::const_iterator it;
		for (it = childItems.constBegin(); it != childItems.constEnd(); ++it) {
			(*it)->expand();
		}
	} else {
		expanded = true;
		QList<ConfigItem*> newitems;
		int i = tplt_range.first();
		int index = 0;
		do {
			ConfigItem *cur = new ConfigItem(this, QString::number(i));
			cloneChildren(cur, this, index);
			++index;
			newitems.append(cur);
		} while (tplt_range.next(i));
		childItems.clear();
		QList<ConfigItem*>::const_iterator it;
		for (it = newitems.constBegin(); it != newitems.constEnd(); ++it) {
			appendChild(*it);
		}
	}
}

QVariant ConfigItemModel::headerData(int section, Qt::Orientation orientation,
                                     int role) const
{
	if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
		switch (section) {
		case 0:
			return tr("item");
		case 1:
			return tr("value");
		}
	}
	return QVariant();
}

Qt::ItemFlags ConfigItemModel::flags(const QModelIndex &index) const
{
	if (!index.isValid())
		return Qt::ItemIsEnabled;

	if (index.column() != 1)
		return Qt::ItemIsEnabled | Qt::ItemIsSelectable;

	ConfigItem *item = static_cast<ConfigItem*>(index.internalPointer());
	if (item->isNode() || item->isReadOnly())
		return Qt::ItemIsEnabled | Qt::ItemIsSelectable;

	return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable;
}

bool ConfigItemModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
	if (!index.isValid() || role != Qt::EditRole || index.column() != 1)
		return false;

	ConfigItem *item = static_cast<ConfigItem*>(index.internalPointer());
	if (item->isNode() || item->isReadOnly())
		return false;

	if (!item->setData(value))
		return false;

	emit dataChanged(index, index);
	return true;
}

QVariant ConfigItemModel::data(const QModelIndex &index, int role) const
{
	if (!index.isValid())
		return QVariant();

	ConfigItem *item = static_cast<ConfigItem*>(index.internalPointer());

	if (role == Qt::DisplayRole)
		return item->data(index.column());

	if (index.column() != 1 || item->isNode() || item->isReadOnly())
		return QVariant();

	if (role == Qt::BackgroundColorRole) {
		switch (item->state()) {
		case ConfigItem::SValid:
			return QColor(Qt::green);
		case ConfigItem::SInvalid:
			return QColor(Qt::red);
		case ConfigItem::SIgnore:
			return QColor(Qt::yellow);
		default:
			return QVariant();
		}
	}
#if 0
	if (role == Qt::TextColorRole) {
		switch (item->state()) {
		case ConfigItem::SValid:
			return QColor(Qt::black);
		case ConfigItem::SInvalid:
			return QColor(Qt::black);
		case ConfigItem::SIgnore:
			return QColor(Qt::black);
		default:
			return QVariant();
		}
	}
#endif
	if (role == Qt::ToolTipRole) {
		QString ret;
		QVariant devdata = item->deviceData();
		if (devdata.isValid())
			ret = tr("device: ") + devdata.toString();
		QVariant userdata = item->userData();
		if (userdata.isValid()) {
			QString userstr = tr("user: ") + userdata.toString();
			if (!devdata.isValid())
				ret = userstr;
			else {
				ret += "<br>";
				ret += userstr;
			}
		}
		return ret;
	}
	return QVariant();
}

int ConfigItemModel::columnCount(const QModelIndex &parent) const
{
	return 2;
}

int ConfigItemModel::rowCount(const QModelIndex &parent) const
{
	ConfigItem *parentItem;

	if (!parent.isValid())
		parentItem = rootItem;
	else
		parentItem = static_cast<ConfigItem*>(parent.internalPointer());

	return parentItem->childCount();
}

QModelIndex ConfigItemModel::parent(const QModelIndex &index) const
{
	if (!index.isValid())
		return QModelIndex();

	ConfigItem *childItem = static_cast<ConfigItem*>(index.internalPointer());
	ConfigItem *parentItem = childItem->parent();

	if (parentItem == rootItem)
		return QModelIndex();

	return createIndex(parentItem->row(), 0, parentItem);
}

QModelIndex ConfigItemModel::index(int row, int column, const QModelIndex &parent) const
{
	ConfigItem *parentItem;

	if (!parent.isValid())
		parentItem = rootItem;
	else
		parentItem = static_cast<ConfigItem*>(parent.internalPointer());

	ConfigItem *childItem = parentItem->child(row);
	if (childItem)
		return createIndex(row, column, childItem);
	else
		return QModelIndex();
}

QModelIndex ConfigItemModel::itemindex(ConfigItem* item, int column) const
{
	int row = item->row();
	return createIndex(row, column, item);
}

bool ConfigItemModel::readLine(QFile &file, QString &line)
{
	line = QString::fromUtf8(file.readLine());
	if (line.isEmpty()) {
		if (!file.atEnd())
			error = true;
		return false;
	}
	//std::cerr << "read: <" << line.toLocal8Bit().constData() << ">" << std::endl;
	return true;
}

bool ConfigItemModel::finishEntry()
{
	if (!curItem)
		return true;

	if (curItem->isValid()) {
		curItem = 0;
		return true;
	}

	curItem->parent()->removeChild(curItem);
	delete curItem;

	curItem = 0;
	return false;
}

bool ConfigItemModel::addKey(const QStringList &keys, bool mult)
{
	if (keys.isEmpty())
		return false;

	ConfigItem *parent = rootItem;

	QStringList::const_iterator it;
	QStringList::const_iterator last = keys.constEnd() - 1;
	for (it = keys.constBegin(); it != last; ++it) {
		ConfigItem *cur;
		cur = parent->findChild(*it);
		if (!cur) {
			cur = new ConfigItem(parent, *it);
			parent->appendChild(cur);
		} else if (!cur->isNode()) {
			return false;
		}
		parent = cur;
	}
	if (mult && !parent->isStatic())
		return false;
	ConfigItem *cur = parent->findChild(*it);
	if (!cur) {
		cur = new ConfigItem(parent, *it, mult, mult); //FIXME freeing after expand() not yet working
		parent->appendChild(cur);
	} else {//if (cur->isNode() != mult || cur->isTemplate() != mult) {
		return false;
	}
	curItem = cur;

	return true;
}

bool ConfigItemModel::setRangeType(const QStringList &args)
{
	if (args.size() != 1)
		return false;

	Range r;
	if (!r.parse(args[0]))
		return false;

	ConfigValidator *val = new CValRange(r);
	curItem->setValidator(val);
	validators.append(val);

	return true;
}

bool ConfigItemModel::setRegExpType(const QStringList &args)
{
	if (args.size() != 1)
		return false;

	QString str = args[0];
	int len = str.size();
	if (len < 2 || str[0] != '"' || str[len - 1] != '"')
		return false;

	--len;
	int i;
	QChar c;
	QString reg;
	for (i = 1; i < len; ++i) {
		c = str[i];
		if (c == '\\') {
			++i;
			if (i == len)
				return false;
			reg += str[i];
		} else {
			reg += c;
			if (c == '"')
				return false;
		}
	}

	QRegExp r(reg);
	if (!r.isValid())
		return false;

	ConfigValidator *val = new CValRegExp(r);
	curItem->setValidator(val);
	validators.append(val);

	return true;
}

bool ConfigItemModel::setDateType(ConfigItem::Type type, const QStringList &args)
{
	return true;
}

bool ConfigItemModel::setTimeType(ConfigItem::Type type, const QStringList &args)
{
	return true;
}


bool ConfigItemModel::parseType(const QString &str)
{
	if (curItem->validator())
		return false;

	int len = str.size();
	int i;
	QChar c;

	QString name;
	int spaces = 0;
	for (i = 0; i < len; ++i) {
		c = str[i];

		if (c == '(')
			break;

		name += c;
		if (c.isSpace())
			++spaces;
		else
			spaces = 0;
	}
	++i; // skip '('
	if (i >= len)
		return false;
	name.chop(spaces);

	bool done = false;
	QStringList arguments;

	do {
		bool instring = false;
		QString cur;
		spaces = 0;
		for (; i < len; ++i) {
			c = str[i];
			if (instring) {
				if (c == '\\') {
					++i;
					if (i == len)
						return false;
					cur += '\\';
					cur += str[i];
				} else {
					cur += c;
					if (c == '"')
						instring = false;
				}
			} else {
				if (cur.isEmpty() && c.isSpace())
					continue;
				if (c == ')') {
					done = true;
					break;
				} else if (c == ',') {
					++i;
					if (i == len)
						return false;
					break;
				}
				cur += c;
				if (c.isSpace())
					++spaces;
				else {
					spaces = 0;
					if (c == '"')
						instring = true;
				}
			}
		}
		if (i == len)
			return false;
		cur.chop(spaces);
		arguments << cur;
	} while (!done);
	++i; //skip ')'
	for (; i < len; ++i) {
		c = str[i];
		if (!c.isSpace())
			return false;
	}
	if (name == "range")
		return setRangeType(arguments);
	if (name == "regexp")
		return setRegExpType(arguments);
	if (name == "day")
		return setDateType(ConfigItem::TDay, arguments);
	if (name == "month")
		return setDateType(ConfigItem::TMonth, arguments);
	if (name == "year")
		return setDateType(ConfigItem::TYear, arguments);
	if (name == "minute")
		return setTimeType(ConfigItem::TMinute, arguments);
	if (name == "hour")
		return setTimeType(ConfigItem::THour, arguments);
	return false;
}

bool ConfigItemModel::addEntry(const QString &lhs, const QString &rhs)
{
	bool ret;
	int val;
	if (!curItem->isTemplate()) {
		if (lhs == "id") {
			val = rhs.toInt(&ret);
			if (!ret)
				return false;
			curItem->setId(val);
			return true;
		} else if (lhs == "tid") {
			val = rhs.toInt(&ret);
			if (!ret)
				return false;
			curItem->setTid(val);
			return true;
		} else if (lhs == "type") {
			return parseType(rhs);
		} else if (lhs == "fwrange") {
			Range r;
			if (!r.parse(rhs, 1000, 1))
				return false;
			curItem->setFWRange(r);
			return true;
		} else if (lhs == "readonly") {
			val = rhs.toInt(&ret);
			if (!ret || val < 0 || val > 1)
				return false;
			curItem->setReadOnly(val == 1);
			return true;
		} else if (lhs == "dtype") {
			if (rhs == "byte") {
				curItem->setDataType(ConfigItem::DTByte);
				return true;
			} else if (rhs == "string") {
				curItem->setDataType(ConfigItem::DTString);
				return true;
			} else
				return false;
		} else
			return false;
	} else {
		if (lhs == "delta") {
			val = rhs.toInt(&ret);
			if (!ret)
				return false;
			curItem->setDelta(val);
			return true;
		} else if (lhs == "tiddelta") {
			val = rhs.toInt(&ret);
			if (!ret)
				return false;
			curItem->setTidDelta(val);
			return true;
		} else if (lhs == "tid") {
			val = rhs.toInt(&ret);
			if (!ret)
				return false;
			curItem->setTemplateTid(val);
			return true;
		} else if (lhs == "range") {
			Range r;
			if (!r.parse(rhs))
				return false;
			curItem->setTemplateRange(r);
			return true;
		} else
			return false;
	}
	return false;
}

bool ConfigItemModel::parseKey(QFile &file, QString &line, int start, int &len, unsigned &lno)
{
	QStringList keys;
	bool mult = false;
	bool done = false;
	int i = start;

	do {
		QChar c;

		for (; i < len; ++i) {
			c = line[i];
			if (!c.isSpace())
				break;
		}

		if (i == len)
			return false;

		if (c == '@') {
			mult = true;
			++i;
			if (i == len)
				return false;
		}

		//int start = i;
		QString cur;
		int spaces = 0;
		for (; i < len; ++i) {
			c = line[i];
			if (c == '/') {
				if (mult)
					return false;
				//QString cur = line.mid(start, i - start).trimmed();
				cur.chop(spaces);
				keys << cur;
				++i;
				if (i == len)
					return false;
				break;
			}
			if (c == ']') {
				//QString cur = line.mid(start, i - start).trimmed();
				cur.chop(spaces);
				keys << cur;
				done = true;
				break;
			}
			if (c == '\\') {
				++i;
				if (i != len) {
					cur += line[i];
					spaces = 0;
				} else {
					QString next;
					if (!readLine(file, next))
						return false;
					++lno;
					line += next;
					len += next.size(); //FIXME overflow
				}
			} else {
				cur += c;
				if (c.isSpace())
					++spaces;
				else
					spaces = 0;
			}
		}
	} while (!done);

	++i; //skip ']'
	if (i != len) {
		for (;;) {
			QChar c;
			for (; i < len; ++i) {
				c = line[i];
				if (!c.isSpace())
					break;
			}
			if (i == len)
				break;
			if (c == '#')
				break;
			if (c != '\\')
				return false;
			++i;
			if (i != len)
				return false;
			QString next;
			if (!readLine(file, next))
				return false;
			++lno;
			line += next;
			len += next.size();
		}
	}

	return addKey(keys, mult);
}

bool ConfigItemModel::parseLine(QFile &file, QString &line, int start, int &len, unsigned &lno)
{
	QString lhs;
	QString rhs;
	int i = start;
	QChar c;

	int spaces = 0;
	for (; i < len; ++i) {
		c = line[i];
		if (lhs.isEmpty() && c.isSpace())
			continue;
		if (c == '=')
			break;
		if (c == '#')
			return lhs.isEmpty();
		if (c == '\\') {
			++i;
			if (i != len) {
				lhs += line[i];
				spaces = 0;
			} else {
				QString next;
				if (!readLine(file, next))
					return false;
				++lno;
				line += next;
				len += next.size(); //FIXME overflow
			}
		} else {
			lhs += c;
			if (c.isSpace())
				++spaces;
			else
				spaces = 0;
		}
	}
	if (i == len)
		return lhs.isEmpty();

	lhs.chop(spaces);
	++i; //skip '='

	spaces = 0;
	bool instring = false;
	for (; i < len; ++i) {
		c = line[i];
		if (instring) {
			if (c == '\\') {
				++i;
				if (i != len) {
					rhs += '\\';
					rhs += line[i];
				} else {
					QString next;
					if (!readLine(file, next))
						return false;
					++lno;
					line += next;
					len += next.size(); //FIXME overflow
				}
			} else {
				rhs += c;
				if (c == '"')
					instring = false;
			}
		} else {
			if (rhs.isEmpty() && c.isSpace())
				continue;
			if (c == '#')
				break;
			if (c == '\\') {
				++i;
				if (i != len) {
					rhs += line[i];
					spaces = 0;
				} else {
					QString next;
					if (!readLine(file, next))
						return false;
					++lno;
					line += next;
					len += next.size(); //FIXME overflow
				}
			} else {
				rhs += c;
				if (c.isSpace())
					++spaces;
				else {
					spaces = 0;
					if (c == '"')
						instring = true;
				}
			}
		}
	}
	if (instring)
		return false;
	rhs.chop(spaces);

	return addEntry(lhs, rhs);
}

ConfigItemModel::ConfigItemModel()
	: QAbstractItemModel(), error(false)
{
	curItem = 0;
	rootItem = new ConfigItem;
	
}

ConfigItemModel::~ConfigItemModel()
{
	//delete curItem;
	delete rootItem;
	qDeleteAll(validators);
}

bool ConfigItemModel::readConfigrc(QFile &file)
{
	unsigned lno = 0;
	for (;;) {
		QString line;
		if (!readLine(file, line)) {
			if (error) {
				std::cerr << "read: error" << std::endl;
				//FIXME: cleanup
				//FIXME: msg
				finishEntry();
				emit layoutChanged();
				return false;
			}
			break;
		}
		++lno;

		//QString cleanline = line.trimmed();
		int len = line.size();
		int i;

		for (i = 0; i < len; ++i)
			if (!line[i].isSpace())
				break;

		if (i == len)
			continue; //empty line

		int curline = lno;
		if (line[i] == '[') {
			if (!finishEntry()) {
				//FIXME msg
			}
			if (!parseKey(file, line, i + 1, len, lno)) {
				std::cerr << "error in line " << curline << std::endl;
				//FIXME: cleanup
				//FIXME: msg
				if (error) {
					finishEntry();
					emit layoutChanged();
					return false;
				}
				continue;
			}
		} else {
			if (!parseLine(file, line, i, len, lno)) {
				std::cerr << "error in line " << curline << std::endl;
				//FIXME: cleanup
				//FIXME: msg
				if (error) {
					finishEntry();
					emit layoutChanged();
					return false;
				}
				continue;
			}
		}
	}
	finishEntry();
	rootItem->expand();
	emit layoutChanged();
	if (!rootItem->resolvePaths(rootItem))
		return false;;
	rootItem->debug("");
	return true;
}

void ConfigItemModel::fillGconfList(struct gconf *&gc, QList<ConfigItem *> &empty, int version, bool data)
{
	empty.clear();
	rootItem->fillGconfList(gc, empty, version, data);
}

void ConfigItemModel::clear(const QList<ConfigItem *> &itemlist)
{
	QList<ConfigItem *>::const_iterator it;
	for (it = itemlist.constBegin(); it != itemlist.constEnd(); ++it) {
		if ((*it)->clear()) {
			QModelIndex index = itemindex(*it, 1);
			emit dataChanged(index, index);
		}
	}
}

void ConfigItemModel::readGconfList(struct gconf *&gc)
{
	for (struct gconf *cur = gc; cur; cur = cur->next) {
		ConfigItem *item = static_cast<ConfigItem*>(cur->priv.ptr);
		if (item->readGconf(*cur)) {
			/* FIXME msg on error? */;
			QModelIndex index = itemindex(item, 1);
			emit dataChanged(index, index);
		}
	}
}

Config::Config(GigasetQt &gigaset, QWidget *parent)
	: QWidget(parent), gs(gigaset)
{
	QFile configrc(App->dataDir() + "/configrc");
	if (!configrc.open(QIODevice::ReadOnly|QIODevice::Text))
		return; //FIXME: msg

	m.readConfigrc(configrc);

	QFrame *line;
	QHBoxLayout *buttonLayout;
	QVBoxLayout *layout;

	tv = new QTreeView(this);
	tv->setModel(&m);

	ConfigDelegate *delegate = new ConfigDelegate(this);
	tv->setItemDelegate(delegate);

	line = new QFrame(this);
	line->setFrameShape(QFrame::HLine);

	bLoad = new QPushButton(tr("&Load"), this);
	bLoad->setEnabled(gs.isavailable());

	bStore = new QPushButton(tr("&Store"), this);
	bStore->setEnabled(gs.isavailable());

	bReset = new QPushButton(tr("&Reset"), this);

	buttonLayout = new QHBoxLayout();
	buttonLayout->addWidget(bLoad);
	buttonLayout->addWidget(bStore);
	buttonLayout->addWidget(bReset);
	buttonLayout->addStretch();

	layout = new QVBoxLayout(this);
	layout->addWidget(tv);
	layout->addWidget(line);
	layout->addLayout(buttonLayout);

	connect(bLoad, SIGNAL(clicked()), SLOT(load()));
	connect(&gs, SIGNAL(deviceAvailable(bool)), bLoad, SLOT(setEnabled(bool))); //FIXME use action?

	connect(bStore, SIGNAL(clicked()), SLOT(store()));
	connect(&gs, SIGNAL(deviceAvailable(bool)), bStore, SLOT(setEnabled(bool))); //FIXME use action?

	connect(bReset, SIGNAL(clicked()), SLOT(resetData()));

	setWindowTitle(tr("Configuration Browser"));
	setLayout(layout);
}

void ConfigItemModel::resetData()
{
	QList<ConfigItem *> itemlist;
	rootItem->reset(itemlist);
	QList<ConfigItem *>::const_iterator it;
	for (it = itemlist.constBegin(); it != itemlist.constEnd(); ++it) {
		QModelIndex index = itemindex(*it, 1);
		emit dataChanged(index, index);
	}
}

void Config::resetData()
{
	m.resetData();  //FIXME Config::resetData not needed
}

void Config::store()
{
	int ver[4];
	int firmware;
	struct gconf *gc = 0;

	if (!gs.claim())
		return; //FIXME msg

	if (gs.unimodem() < 0 || gs.firmware(ver) < 0 || gs.cidmode() < 0) {
		gs.release();
		return; //FIXME msg
	}

	firmware = ver[0] * 1000 + ver[1];

	QList<ConfigItem *> empty;
	m.fillGconfList(gc, empty, firmware, true);
	empty.clear();

	if (!gc) {
		gs.release();
		//FIXME gc
		return; //FIXME msg
	}

	if (gs.config_start() < 0) {
		gs.release();
		//FIXME gc
		return; //FIXME msg
	}

	if (gs.config_prepare() < 0 ||
	    gs.config_write_list(gc) < 0) {
		//FIXME msg
		//FIXME gc
	} else {
		struct gconf *cur = gc;
		while (cur) {
			struct gconf *tmp = cur;
			cur = cur->next;
			delete[] tmp->buf;
			delete tmp;
		}
	}

	gs.config_end();
	gs.release();

	load();
}

void Config::load()
{
	int ver[4];
	int firmware;
	struct gconf *gc = 0;

	if (!gs.claim())
		return; //FIXME msg

	if (gs.unimodem() < 0 || gs.firmware(ver) < 0 || gs.cidmode() < 0) {
		gs.release();
		return; //FIXME msg
	}

	firmware = ver[0] * 1000 + ver[1];

	QList<ConfigItem *> empty;
	m.fillGconfList(gc, empty, firmware, false);

	if (!gc) {
		gs.release();
		//FIXME gc
		return; //FIXME msg
	}

	if (gs.config_start() < 0) {
		gs.release();
		//FIXME gc
		return; //FIXME msg
	}

	if (gs.config_prepare() < 0 ||
	    gs.config_read_list(gc) < 0) {
		//FIXME msg
	} else {
		m.clear(empty);
		m.readGconfList(gc);
		struct gconf *cur = gc;
		while (cur) {
			struct gconf *tmp = cur;
			cur = cur->next;
			Gigaset::free_buf(tmp->buf);
			delete tmp;
		}
	}

	gs.config_end();
	gs.release();
}

void Config::closeEvent(QCloseEvent *e)
{
	e->accept(); // e->ignore();
}

ConfigDelegate::ConfigDelegate(QObject *parent)
	: QItemDelegate(parent)
{
}

QWidget *ConfigDelegate::createEditor(QWidget *parent, const QStyleOptionViewItem &option,
                                      const QModelIndex &index) const
{
#if 0
	if (!index.isValid())
		return 0;

	if (index.column() != 1)
		return 0;
#endif

	ConfigItem *item = static_cast<ConfigItem*>(index.internalPointer());
#if 0
	if (item->isNode() || item->isReadOnly())
		return 0;
#endif

	switch (item->dataType()) {
	default:
		return 0;

	case ConfigItem::DTByte:
		{
			QSpinBox *sb = new QSpinBox(parent);
			sb->setMinimum(0);
			sb->setMaximum(255);
			sb->installEventFilter(const_cast<ConfigDelegate*>(this));
			return sb;
		}
	case ConfigItem::DTString:
		{
			QLineEdit *le = new QLineEdit(parent);
			le->installEventFilter(const_cast<ConfigDelegate*>(this));
			return le;
		}
	}
}

void ConfigDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const
{
	//FIXME make this object oriented?
	ConfigItem *item = static_cast<ConfigItem*>(index.internalPointer());
	switch (item->dataType()) {
	default:
		return;

	case ConfigItem::DTByte:
		{
			int byteval = index.model()->data(index, Qt::DisplayRole).toInt();
			QSpinBox *sb = static_cast<QSpinBox*>(editor);
			sb->setValue(byteval);
			return;
		}
	case ConfigItem::DTString:
		{
			QString strval = index.model()->data(index, Qt::DisplayRole).toString();
			QLineEdit *le = static_cast<QLineEdit*>(editor);
			le->setText(strval);
			return;
		}
	}
}

void ConfigDelegate::setModelData(QWidget *editor, QAbstractItemModel *model,
                                  const QModelIndex &index) const
{
	//FIXME make this object oriented?
	ConfigItem *item = static_cast<ConfigItem*>(index.internalPointer());
	switch (item->dataType()) {
	default:
		return;

	case ConfigItem::DTByte:
		{
			QSpinBox *sb = static_cast<QSpinBox*>(editor);
			sb->interpretText();
			int byteval = sb->value();
			model->setData(index, byteval);
			return;
		}
	case ConfigItem::DTString:
		{
			QLineEdit *le = static_cast<QLineEdit*>(editor);
			QString strval = le->text();
			model->setData(index, strval);
			return;
		}
	}
}

void ConfigDelegate::updateEditorGeometry(QWidget *editor, const QStyleOptionViewItem &option,
                                          const QModelIndex &index) const
{
	editor->setGeometry(option.rect);
}
