/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include "gigaset.hh"
//#include <iostream>
#include <cstdlib>
#include <cerrno>
#include <limits>
#include <iostream>

extern "C" {
#include <gigaset/sms.h>
#include <gigaset/interf.h>
}

int Gigaset::open(const char *device, struct glog *glog)
{
	if (gs)
		return -ERR_BUSY;

	int ret;
	gs = gigaset_open(&ret, &sys_err, device, glog, &sigint);
	if (!gs)
		return ret;

	ret = gigaset_lock(gs, 1); //FIXME okay here? move to claim()?

	if (ret < 0) {
		sys_err = gigaset_errno(gs);
		gigaset_close(gs);
		gs = 0;
		return ret;
	}

	available(true);
	return 0;
}


int Gigaset::close()
{
	if (!gs || dev_claimed)
		return -ERR_BUSY;

	int ret = gigaset_lock(gs, 0); //FIXME okay here? move to release()?
	if (ret < 0)
		sys_err = gigaset_errno(gs);
	else
		sys_err = 0;
	int ret2 = gigaset_close(gs);
	gs = 0;

	available(false);
	return ret < 0 ? ret : ret2;
}

int Gigaset::sms_start()
{
	//check gs
	//check if sms_start() called before
	int ret;
	ret = gigaset_smsstart(gs);
	return ret;
}

int Gigaset::sms_prepare()
{
	//check gs
	//check if sms_start() called before
	int ret;
	ret = gigaset_smscontinue(gs);
	return ret;
}

int Gigaset::sms_end()
{
	//check gs
	//check if sms_start() called before
	int ret;
	ret = gigaset_smsend(gs);
	return ret;
}

int Gigaset::sms_getmessages(struct gigaset_sm **messages, struct gigaset_sm_chain **chains)
{
	//check gs
	//check if sms_prepare() called before
	int ret;
	ret = gigaset_smsgetmessages(gs, messages, chains);
	return ret;
}

void Gigaset::sms_freemessages(struct gigaset_sm **messages, struct gigaset_sm_chain **chains)
{
	gigaset_smsfreemessages(messages, chains);
}

ssize_t Gigaset::sms_to_utf8(struct gigaset_sm *msg, int chain, char **buffer)
{
	return gigaset_sms2utf8(msg, chain, buffer);
}

bool parse_number(char *&pos, int &res, char delim)
{
	char *end;
	unsigned long val;

	errno = 0;
	val = strtoul(pos, &end, 10);
	if (errno || end == pos || val > (unsigned) std::numeric_limits<int>::max() || *end != delim)
		return false;
	res = val;
	pos = end + 1;
	return true;
}

int Gigaset::firmware(int ver[4])
{
	int ret;
	char *buf, *pos;

	ret = firmwarestr(&buf);
	if (ret < 0)
		return ret;

	pos = buf;
	if (parse_number(pos, ver[0], 0x2e) &&
	    parse_number(pos, ver[1], 0x2e) &&
	    parse_number(pos, ver[2], 0x2e) &&
	    parse_number(pos, ver[3], 0))
		ret = 0;
	else
		ret = -ERR_DATA;

	free_buf(buf);
	return ret;
}

int Gigaset::config_read_list(struct gconf *list)
{
	// FIXME gigaset_confreadarray didn't work with big item counts
	for (struct gconf *cur = list; cur; cur = cur->next) {
		struct gconf tmp = *cur;
		tmp.next = 0;
		int ret = gigaset_confreadarray(gs, &tmp);
		if (ret < 0)
			return ret;
		cur->len = tmp.len;
		cur->buf = tmp.buf;
	}
	return 0;
}

int Gigaset::config_write_list(struct gconf *list)
{
	int ret;
	for (struct gconf *cur = list; cur; cur = cur->next) {
		std::cerr << "writing " << cur->id << ", " << (unsigned)cur->tid << ", " << (unsigned)cur->datatype << ", " << cur->len << std::endl;
		ret = gigaset_confkeeplock(gs);
		if (ret < 0)
			return ret;
		ret = gigaset_confwrite(gs, cur->id, cur->tid, cur->datatype, cur->buf, cur->len);
		if (ret < 0)
			return ret;
	}
	return 0;
}
