/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#ifndef GSTATE_HH
#define GSTATE_HH

#include <csignal>
#include <cstdlib>
extern "C" {
#include <stdint.h>
#include <gigaset/interf.h>
#include <gigaset/sms.h>
#include <gigaset/gigdata.h>
#include <gigaset/basemisc.h>
#include <gigaset/conf.h>
}

#define ERR_BUSY  (ERR_LIBMAX + 1)   //FIXME
#define ERR_GSMAX (ERR_LIBMAX + 16)

class Gigaset
{
public:
	Gigaset(): glog(0), gs(0), sigint(0), dev_claimed(false) {}
	virtual ~Gigaset() { dev_claimed = false; close(); }
	int open(const char *device, struct glog *glog);
	int close();
	int cidmode() { return gigaset_cidmode(gs); }
	int unimodem() { return gigaset_unimodem(gs); }
	void setInterrupt(std::sig_atomic_t val) { sigint = val; }

	bool isopen() const { return gs; }
	bool isavailable() const { return gs && !dev_claimed; }
	int get_sys_err() const { return gs ? gigaset_errno(gs) : sys_err; }

	int sms_start();
	int sms_prepare();
	int sms_end();

	//int sms_del(uint32_t handle) { return gigaset_smsdelhandle(gs, handle); } /* does not work as expected */
	int sms_del() { return gigaset_smsdel(gs); }
	int sms_select(uint32_t handle) { return gigaset_smsselect(gs, handle); }
	bool sms_validchar(uint16_t c_ucs2) { return gigaset_valid_sms_char(c_ucs2); }
	int sms_getmessages(struct gigaset_sm **messages, struct gigaset_sm_chain **chains);

	static int sms_from_utf8(struct gigaset_sm **ret_messages, struct gigaset_sm_chain **ret_chain,
                                 uint8_t flags, uint8_t unknown1, uint8_t dcs, uint8_t vp,
                                 int chain, uint8_t chain_id, const char * buffer, size_t len)
	{
		return gigaset_smsfromutf8(ret_messages, ret_chain, flags, unknown1,
		                           dcs, vp, chain, chain_id, buffer, len);
	}
	static void sms_freemessages(struct gigaset_sm **messages, struct gigaset_sm_chain **chains);
	static ssize_t sms_to_utf8(struct gigaset_sm *msg, int chain, char **buffer) /*FIXME const*/;
	//static void sms_free_buf(char *buf) { std::free(buf); }

	int sms_new(uint32_t *handle) { return gigaset_smsnew(gs, handle); }
	int sms_put_text(const char *buf, size_t size) { return gigaset_smsputtext(gs, buf, size); }
	int sms_put_message(const struct gigaset_sm *msg) { return gigaset_smsputmessage(gs, msg); }
	int sms_number(const char *number) { return gigaset_smsnumber(gs, number); };
	int sms_store() { return gigaset_smsstore(gs); }
	int sms_send() { return gigaset_smssend(gs); }

	int_fast32_t config_read(uint16_t id, uint8_t tid, uint8_t datatype, char **buf)
	{
		return gigaset_confread(gs, id, tid, datatype, buf);
	}
	int config_start() { return gigaset_confinitlock(gs); }
	int config_prepare() { return gigaset_confkeeplock(gs); }
	int config_end() { return gigaset_confunlock(gs); }
	//int config_read_list(struct gconf *list) { return gigaset_confreadarray(gs, list); }
	int config_read_list(struct gconf *list);
	int config_write_list(struct gconf *list);

	int firmwarestr(char **buf) { return gigaset_version(gs, buf); }
	int firmware(int ver[4]);
	static void free_buf(char *buf) { std::free(buf); }

	int dial(unsigned type, const char *number, const char *id) { return gigaset_dial(gs, type, number, id); }

	static int enumerate_devices(struct gigstrlist **list) { return gigaset_enumerate_devices(list); }
	static void freestrlist(struct gigstrlist **list) { gigaset_freestrlist(list); }

	static void to_be16(char *buf, uint16_t val) { gigaset_to_be16(buf, val); }

	bool claim()
	{
		if (!gs || dev_claimed)
			return false;
		dev_claimed = true;
		
		available(false);
		return true;
	}
	void release()
	{
		if (!gs || !dev_claimed)
			return;
		dev_claimed = false;
		available(true);
	}
	virtual void available(bool) {}
protected:
private:
	struct glog *glog;
	struct gstatus *gs;
	int sys_err;
	volatile std::sig_atomic_t sigint;
	bool dev_claimed;
};

#endif
