/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#ifndef LOGGER_HH
#define LOGGER_HH

#include <QtGui>
//#include "gigasetqt.hh"
extern "C" {
#include <gigaset/interf.h>
}

#define GLOG_QTDBG     (0x0001U << GLOG_LIBBITS)
#define GLOG_QTWARN    (0x0002U << GLOG_LIBBITS)
#define GLOG_QTCRIT    (0x0004U << GLOG_LIBBITS)
#define GLOG_QTFATAL   (0x0008U << GLOG_LIBBITS)
#define GLOG_QTUNKNOWN (0x0010U << GLOG_LIBBITS)

extern "C" void loggercallback(void *data, unsigned level, const char *str);

class Logger: public QObject
{
	Q_OBJECT
public:
	Logger(bool qtHandler = true);
	~Logger();
	void callback(unsigned level, const char *str); // should not be public
#if 0
	void fillGlog(struct glog &g) {
		g.data = this;
		g.logfunc = loggercallback;
	}
#endif
	struct glog* glogPtr() { return &glog; }
	void installQtHandler();
	static Logger *getQtData() { return qtData; }
	static QString levelToString(unsigned level);
signals:
	void newMessage(unsigned int level, QByteArray msg);
private:
	Logger(const Logger &l); // not implemented
	Logger &operator=(const Logger &l); // not implemented
	struct glog glog;
	static Logger *qtData;
};

class LogListViewItem: public QTreeWidgetItem
{
public:
	LogListViewItem(QTreeWidget *parent, const QByteArray &msg, unsigned lev);
	const QByteArray &getMsg() const { return message; }
	unsigned getLevel() const { return level; }
private:
	QByteArray message;
	unsigned level;
};

class LogList: public QWidget
{
	Q_OBJECT

public:
	LogList(QWidget *parent = 0);
	//~LogList();
protected:
	//void closeEvent(QCloseEvent*);
public slots:
	void newMessage(unsigned int level, QByteArray msg);
	void save();
private:
	QTreeWidget *tw;
};

#endif
