/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <QTranslator>
#include <cstdlib>
#include "application.hh"
#include "main.hh"
//#include <iostream>

#define argtostr(x)  #x
#define stringify(x) argtostr(x)

Application::Application(int &argc, char **argv, const QString &name)
	: QApplication(argc, argv), appname(name)
{
	//argv/argc: "-datadir=foo"
#ifdef Q_OS_WIN32 // Q_WS_WIN
	const char *ddir = std::getenv("COMMON_APPDATA");
	if (ddir)
		datadir = QString::fromLocal8Bit(ddir).replace('\\', '/') + '/' + appname;
#elif defined QGIGASET_DATADIR
	datadir = QString::fromLocal8Bit(stringify(QGIGASET_DATADIR)) + '/' + appname;
#endif
	if (datadir.isEmpty() || !QDir(datadir).exists())
		datadir = QCoreApplication::applicationDirPath(); // or QDir::current()
}

int main(int argc, char *argv[])
{
	Application a(argc, argv, "qgigaset");

	//std::cerr << a.dataDir().toLocal8Bit().constData() << std::endl;

	QTranslator qtTranslator;
	qtTranslator.load("qt_" + QLocale::system().name(), a.dataDir());
	a.installTranslator(&qtTranslator);

	QTranslator qgTranslator;
	qgTranslator.load("qgigaset_" + QLocale::system().name(), a.dataDir());
	a.installTranslator(&qgTranslator);

	ApplicationWindow mw;
	mw.show();

	return a.exec();
}
