/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <QtGui>
#include "smscreator.hh"
#include "gigasetqt.hh"
#include <iostream>

VPSpinBox::VPSpinBox(QWidget *parent)
	: QSpinBox(parent)
{
	setRange(-1, 255);
	//setReadOnly(true);
	lineEdit()->setReadOnly(true);
	setValue(-1);
	setSpecialValueText(tr("none"));
}

QString VPSpinBox::textFromValue(int val) const
{
	//std::cerr << "textFromValue: " << val << std::endl;

	if (val < 0/*-1*/ || val > 255)
		return QString();

	/*if (val == -1)
		return tr("none");*/

	if (val < 144)
		return QString(tr("%1 minutes")).arg((val+1) * 5);

	if (val < 168)
		if (val % 2)
			return QString(tr("%1 hours")).arg(12 + (val-143) / 2);
		else
			return QString(tr("%1.5 hours")).arg(12 + (val-143) / 2);

	if (val < 197)
		return QString(tr("%1 days")).arg(val - 166);

	return
		QString(tr("%1 weeks")).arg(val - 192);
}

int VPSpinBox::valueFromText(const QString &text) const
{
	//std::cerr << "valueFromText: " << text.toLocal8Bit().constData() << std::endl;

	int suffixpos = text.indexOf(QRegExp("[^0-9]"));
	if (suffixpos == -1)
		return -2; //FIXME

	bool isvalid;
	int num = text.left(suffixpos).toInt(&isvalid, 10);
	if (!isvalid)
		return -2; //FIXME

	QString suffix = "%1" + text.mid(suffixpos);
	if (suffix == tr("%1 minutes")) {
		if (!(num % 5) && num >= 5 && num <= 720)
			return num / 5 - 1;
	} else if (suffix == tr("%1 hours")) {
		if (num >= 13 && num <= 24)
			return num * 2 + 119;
	} else if (suffix == tr("%1.5 hours")) {
		if (num >= 12 && num <= 23)
			return num * 2 + 120;
	} else if (suffix == tr("%1 days")) {
		if (num >= 2 && num <= 30)
			return num + 166;
	} else if (suffix == tr("%1 weeks")) {
		if (num >= 5 && num < 63)
			return num + 192;
	}
	return -2; //FIXME
}

QValidator::State VPSpinBox::validate(QString &input, int &pos) const
{
	//std::cerr << "validate: " << input.toLocal8Bit().constData() << "->" << valueFromText(input) << std::endl;
	if (valueFromText(input) == -2)
		//return QValidator::Intermediate;
		return QValidator::Invalid;
	else
		return QValidator::Acceptable;
}

SmsCreator::SmsCreator(GigasetQt &gigaset, QWidget *parent)
	: QWidget(parent), gs(gigaset)
{


	QLabel *lel, *cbl, *vpl, *sbl;
	QFrame *line1, *line2;

	QGridLayout *gridLayout;
	QHBoxLayout *buttonLayout, *cbLayout;
	QVBoxLayout *layout;

	teMessage = new QTextEdit(this);

	line1 = new QFrame(this);
	line1->setFrameShape(QFrame::HLine);

	leNumber = new QLineEdit(this);

	cbType = new QComboBox(this);
	cbType->addItem(tr("normal"));
	cbType->addItem(tr("voice message waiting: set"));
	cbType->addItem(tr("voice message waiting: clear"));

	sbVP = new VPSpinBox(this);

	sbChainID = new QSpinBox(this);
	sbChainID->setRange(0, 255);
	sbChainID->setEnabled(false);

	lel = new QLabel(tr("&Number:"), this);
	lel->setBuddy(leNumber);

	cbl = new QLabel(tr("&Type:"), this);
	cbl->setBuddy(cbType);

	vpl = new QLabel(tr("&Validity period:"), this);
	vpl->setBuddy(sbVP);

	sbl = new QLabel(tr("&Chain id:"), this);
	sbl->setBuddy(sbChainID);

	cbLarge = new QCheckBox(tr("&Large message"), this);
	cbStore = new QCheckBox(tr("&Keep message"), this);

	line2 = new QFrame(this);
	line2->setFrameShape(QFrame::HLine);

	bSend = new QPushButton(tr("&Send"), this);
	bSend->setEnabled(gs.isavailable());

	gridLayout = new QGridLayout();

	gridLayout->addWidget(leNumber,  0, 1, 1, 1);
	gridLayout->addWidget(sbVP,      1, 1, 1, 1);
	gridLayout->addWidget(sbChainID, 2, 1, 1, 1);
	gridLayout->addWidget(cbType,    3, 1, 1, 1);

	gridLayout->addWidget(lel,       0, 0, 1, 1);
	gridLayout->addWidget(vpl,       1, 0, 1, 1);
	gridLayout->addWidget(sbl,       2, 0, 1, 1);
	gridLayout->addWidget(cbl,       3, 0, 1, 1);

	cbLayout = new QHBoxLayout();
	cbLayout->addWidget(cbLarge);
	cbLayout->addWidget(cbStore);
	cbLayout->addStretch();

	buttonLayout = new QHBoxLayout();
	buttonLayout->addWidget(bSend);
	buttonLayout->addStretch();

	layout = new QVBoxLayout(this);
	layout->addWidget(teMessage);
	layout->addWidget(line1);
	layout->addLayout(gridLayout);
	layout->addLayout(cbLayout);
	layout->addWidget(line2);
	layout->addLayout(buttonLayout);

	connect(bSend, SIGNAL(clicked()), SLOT(send()));
	connect(&gs, SIGNAL(deviceAvailable(bool)), bSend, SLOT(setEnabled(bool))); //FIXME use action?
	connect(cbLarge, SIGNAL(toggled(bool)), sbChainID, SLOT(setEnabled(bool)));

	setWindowTitle(tr("Message Editor"));
	setLayout(layout);
}

int SmsCreator::sms_create(struct gigaset_sm *msg, const char *number, bool store)
{
	int r;
	uint32_t handle;

	r = gs.sms_new(&handle);
	if (r < 0)
		return r;

	r = gs.sms_put_message(msg); //FIXME delete if <0 ?
	if (r < 0)
		return r;

	r = gs.sms_number(number);
	if (r < 0)
		return r;

	if (store) {
		r = gs.sms_store();
		if (r < 0)
			return r;
	}

	return gs.sms_send();
}

void SmsCreator::send()
{
	int flags, vp, dcs;
	int chain_id;
	bool store;
	int chain_param;

	flags = 1;

	vp = sbVP->value();
	if (vp == -1)
		vp = 0;
	else
		flags |= 0x10;

	switch (cbType->currentIndex()) {
	case 0:
		dcs = 0xf1;
		break;
	case 1:
		dcs = 0xd8;
		break;
	case 2:
		dcs = 0xd0;
		break;
	default:
		emit message(tr("cannot determine DCS"), 0);
		return; //FIXME msg
	}

	chain_id = sbChainID->value();
	chain_param = cbLarge->checkState() == Qt::Unchecked ? 0 : 1;
	store = cbStore->checkState() == Qt::Checked;
	QByteArray number = leNumber->text().toLocal8Bit(); //FIXME encoding
	QByteArray msg = teMessage->toPlainText().toUtf8();

	emit message(tr("Sending ..."), 0);
	std::cerr << "---> " << std::endl;

	struct gigaset_sm *messages;
	struct gigaset_sm_chain *chain;
	int ret, ret2, num;

	num = Gigaset::sms_from_utf8(&messages, &chain, flags, 0, dcs, vp, chain_param,
	                             chain_id, msg.constData(), msg.size());

	if (num < 0) {
		emit message(tr("sms_from_utf8 failed"), 0);
		std::cerr << "sms_from_utf8 failed: " << num << std::endl;
		return; //FIXME msg
	}

	if (!gs.claim()) {
		emit message(tr("claim failed"), 0);
		std::cerr << "claim failed" << std::endl;
		gs.sms_freemessages(&messages, &chain);
		return; //-ERR_BUSY; //FIXME msg
	}

	ret = gs.cidmode();
	if (ret >= 0)
		ret = gs.sms_start();
	if (ret < 0) {
		emit message(tr("sms_start failed"), 0);
		std::cerr << "sms_start failed: " << ret << std::endl;
		gs.release();
		gs.sms_freemessages(&messages, &chain);
		return; //ret; //FIXME msg
	}
	ret = gs.sms_prepare();
	if (ret < 0)
		emit message(tr("sms_prepare failed"), 0);
	if (ret >= 0) {
		if (num == 0) {
			/* single message */
			ret = sms_create(messages, number, store);
		} else {
			for (unsigned i = 0; i < chain->msgcount; ++i) {
				ret = sms_create(chain->msg[i], number, store);
				if (ret < 0)
					break;
			}
		}
		if (ret < 0)
			emit message(tr("sms_create failed"), 0);
	}
	ret2 = gs.sms_end();
	if (ret2 < 0)
		emit message(tr("sms_end failed"), 0);
	std::cerr << "sms_create: " << ret << "/" << ret2 << std::endl;
	gs.release();
	gs.sms_freemessages(&messages, &chain);
	return; // ret < 0 ? ret : ret2;
}

SmsCreator::~SmsCreator()
{
}

void SmsCreator::closeEvent(QCloseEvent *e)
{
	e->accept(); // e->ignore();
}
