/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <QtGui>
#include "smslist.hh"
//#include "application.hh" //FIXME
#include "gigasetqt.hh"
#include <iostream>
#include <limits>

SmsList::SmsList(GigasetQt &gigaset, QWidget *parent)
    : QWidget(parent), gs(gigaset), messages(0), chains(0)
{

	QSplitter *s = new QSplitter(this);
	s->setOrientation(Qt::Vertical);

	tw = new QTreeWidget(s);
	tw->setColumnCount(3);
	QStringList labels;
	labels << tr("handle") << tr("date") << tr("number");
	tw->setHeaderLabels(labels);

	tw->setSelectionMode(QAbstractItemView::ExtendedSelection);
	tw->setSelectionBehavior(QAbstractItemView::SelectRows);

	//tw->setRootIsDecorated(TRUE);
	connect(tw, SIGNAL(currentItemChanged(QTreeWidgetItem *, QTreeWidgetItem *)), SLOT(updateText(QTreeWidgetItem *, QTreeWidgetItem *)));

	te = new QTextEdit(s);
	te->setReadOnly(TRUE);
	//te->setTextFormat(PlainText);

	s->addWidget(tw);
	s->addWidget(te);

	QPushButton *bReload = new QPushButton(tr("&Reload"), this);
	bReload->setEnabled(gs.isavailable());
	connect(bReload, SIGNAL(clicked()), SLOT(load()));
	connect(&gs, SIGNAL(deviceAvailable(bool)), bReload, SLOT(setEnabled(bool))); //FIXME use action?

	QPushButton *bDelete = new QPushButton(tr("&Delete"), this);
	bDelete->setEnabled(gs.isavailable());
	connect(bDelete, SIGNAL(clicked()), SLOT(del()));
	connect(&gs, SIGNAL(deviceAvailable(bool)), bDelete, SLOT(setEnabled(bool))); //FIXME use action?

	//QSpacerItem *spc = new QSpacerItem(40, 20, QSizePolicy::Expanding, QSizePolicy::Minimum);

	//FIXME this could be nicer:
	QHBoxLayout *buttonlayout = new QHBoxLayout;
	buttonlayout->addWidget(bReload);
	buttonlayout->addWidget(bDelete);
	//buttonlayout->addItem(spc);
	buttonlayout->addStretch();
	QVBoxLayout *layout = new QVBoxLayout;
	//layout->addWidget(tw);
	//layout->addWidget(te);
	layout->addWidget(s);
	layout->addLayout(buttonlayout);

	setWindowTitle(tr("SMS Browser"));

	setLayout(layout);
}

void SmsListViewItem::init()
{
	std::cerr << "\t\t" << msg->handle << '\t' << msg->number << std::endl;
	setText(0, QString::number(msg->handle));
	setText(1, QDate(msg->year < 80 ? msg->year + 2000 : msg->year, msg->month, msg->day).toString() + " " + QTime(msg->hour, msg->minute).toString()); //FIXME: error handling
	setText(2, msg->number);
}

void SmsList::updateText(QTreeWidgetItem *cur, QTreeWidgetItem *prev)
{
	te->clear();

	SmsListViewItem *item = dynamic_cast<SmsListViewItem*>(cur); // FIXME qobject_cast?

	if (!item)
		return;

	struct gigaset_sm *msg = item->getMsg();

	std::cerr << (item->parent() == 0) <<  '\t' << item->childCount() << std::endl;

	char *buf;
	ssize_t len = Gigaset::sms_to_utf8(msg, item->parent() == 0 && item->childCount() != 0, &buf);
	if (len <= 0) {
		std::cerr << "sms_to_utf8: " << len << std::endl;
		return;
	}

	--len; // trailing NUL

	if (len > std::numeric_limits<int>::max())
		len = std::numeric_limits<int>::max(); //FIXME

	QString str = QString::fromUtf8(buf, len); //FIXME error handling
	Gigaset::free_buf(buf);

	te->setPlainText(str);

}

int SmsList::delItem(SmsListViewItem * item)
{
	int ret;

	std::cerr << "delete " << item->getMsg()->handle << std::endl;

	ret = gs.sms_select(item->getMsg()->handle);
	if (ret < 0)
		return ret;

	return gs.sms_del();
}

int SmsList::del()
{
	int ret, ret2;
	if (!gs.claim())
		return -ERR_BUSY;

	ret = gs.cidmode();
	if (ret >= 0)
		ret = gs.sms_start();
	if (ret < 0) {
		gs.release();
		return ret;
	}
	ret = gs.sms_prepare();
	if (ret >= 0) {
		QList<QTreeWidgetItem *> sel = tw->selectedItems();
		QList<QTreeWidgetItem *>::const_iterator it;

		for (it = sel.constBegin(); it != sel.constEnd(); ++it) {
			if ((*it)->parent() != 0 && tw->isItemSelected((*it)->parent()))
				continue; // deleted with parent
			if ((*it)->childCount() != 0)
				for (int i = 0; i < (*it)->childCount(); ++i) // is this okay?
					delItem(dynamic_cast<SmsListViewItem*>((*it)->child(i))); //FIXME qobject_cast?
			else
				delItem(dynamic_cast<SmsListViewItem*>(*it)); //FIXME qobject_cast?
		}
#if 0
		for (QListViewItemIterator it(lv, QListViewItemIterator::Selected); it.current(); ++it) {
			if (it.current()->depth() != 0 && it.current()->parent()->isSelected())
				continue; // deleted with parent
			if (it.current()->childCount() != 0)
				for (QListViewItem *c = it.current()->firstChild(); c; c = c->nextSibling())
					delItem(dynamic_cast<SmsListViewItem*>(c));
					/*delItem(dynamic_cast<SmsListViewItem*>(c.current()))*/
			else
				delItem(dynamic_cast<SmsListViewItem*>(it.current()));
		}
#endif
	}

	ret2 = gs.sms_end();
	gs.release();
	load();
	return ret < 0 ? ret : ret2;
}

int SmsList::load()
{
	int ret, ret2;

	tw->clear();
	te->clear();

	if (!gs.claim())
		return -ERR_BUSY;

	ret = gs.cidmode();
	if (ret >= 0)
		ret = gs.sms_start();
	if (ret < 0) {
		gs.release();
		return ret;
	}
	ret = gs.sms_prepare();
	if (ret >= 0) {
		Gigaset::sms_freemessages(&messages, &chains);
		ret = gs.sms_getmessages(&messages, &chains);
		std::cerr << "getmessages: " << ret << '/' << gs.get_sys_err() << std::endl;
		if (ret >= 0) {
			struct gigaset_sm *msg;
			struct gigaset_sm_chain *c;
			QTreeWidgetItem *curtop = 0;
			if (msg = messages) do {
				std::cerr << (void*)msg << '\t' << msg->handle << std::endl;
				if ((c = msg->chain) && !c->missing) {
					if (msg->id != 1)
						continue;
					curtop = curtop ? new SmsListViewItem(tw, curtop, c->msg[0]): new SmsListViewItem(tw, c->msg[0]); //FIXME
					QTreeWidgetItem *cur = new SmsListViewItem(curtop, c->msg[0]);
					std::cerr << '\t' << 0 << '\t' << (void*)c->msg[0] << std::endl;
					for (unsigned i = 1; i < c->msgcount; ++i) {
						cur = new SmsListViewItem(curtop, cur, c->msg[i]);
						std::cerr << '\t' << i << '\t' << (void*)c->msg[i] << std::endl;
					}
				} else {
					std::cerr << '\t' << -1 << '\t' << msg << std::endl;
					curtop = curtop ? new SmsListViewItem(tw, curtop, msg): new SmsListViewItem(tw, msg);
				}
			} while ((msg = msg->next) != messages);
		}
		updateText(tw->currentItem(), NULL);
	}
	ret2 = gs.sms_end();
	gs.release();
	return ret < 0 ? ret : ret2;
}

SmsList::~SmsList()
{
	Gigaset::sms_freemessages(&messages, &chains);
}

void SmsList::closeEvent(QCloseEvent *e)
{
	e->accept(); // e->ignore();
}
